/*****************************************************************************
 * integrated.c: The main file of the userspace daemon implementation. 
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 * Modifications by: Rishi Bhardwaj, <rbhardw2 at uiuc dot edu>
 *
 *****************************************************************************/


/*************************** NOTES *******************************************

This file is the main file of the userspace daemon. The features 
in this file are:
1) Read in and process options from command line.
2) Read in and process a configuration file.
3) Initialize the routing and interface management components.
4) De-multiplex received protocol packets and send it on to appropriate
    component.
5) Support (unused now) for running and testing multiple userspace 
    daemons on a single machine - support is enabled by setting debug_mode flag

The structure of this implementation is inspired by the architecture used in
AODV-UU implementation from Uppsala university.

*********************************************************************************/

/* (Aug 1, 2007) Changes by Rishi Bhardwaj:
 * 1) Added an option to set the minimum size for hello packets
 * 2) Added priority to the management packets (hello messages, RREQs etc.).
 * 3) Added a mechanism to accept messages (commands) on COMMAND_LISTEN_PORT.
 *    Different system parameters may be changed through the commands while the program is still running.
 * 4) Added default handler for various signals.
 * 5) Added some debug message statements.      
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <netinet/ip.h>

#include "utilities.h"
#include "integrated.h"
#include "mcr_hello.h"
#include "timer.h"
#include "mcr_route.h"
#include "mcr_ioctl.h"
#include "mcr_gateway.h"
#include "command.h"

int daemon_proc = 0;
DEBUG_MODE_TYPE debug_mode = NO_DEBUG;

FILE *errMsgFile = NULL;
BOOL suppress_debug_msg = FALSE;

FILE *neighborFile = NULL;
int neighbor_file_updation_ctr = 0;

#ifdef ROUTE_EXPERIMENT
FILE *routeDebugMsgFile = NULL;
FILE *routeNotTakenMsgFile = NULL;
#endif

static void cleanup(void);
ServerData *serverData;     // global declaration of userspace data structure

/******************** Debug support  **********************************/

int
PacketNotFromNeighbor(ServerData *serv, struct in_addr *addr)
{
     DebugTable *list = serv->debugTable;

     while(list != NULL) {
	  if( !memcmp(&list->addr, addr, sizeof(struct in_addr)) )
	       return 0; // addr is that of a neighbor
	  list = list->next;
     }
     return 1;  // not found => not a neighbor
}

void
ReadDebugTable(ServerData *serv, char *filename) {
     FILE *fp;
     char line[MAX_MESSAGE_LENGTH];
     char myAddrStr[MAX_MESSAGE_LENGTH], neighborAddrStr[MAX_MESSAGE_LENGTH];
     struct in_addr myAddr, neighborAddr;
     int n;
     DebugTable *tmp;

     fp = fopen(filename,"r");
     if(fp == NULL)
	  err_quit("In DEBUG_MODE: Cannot open %s", filename);
    
     while(!feof(fp)) {
	  fgets(line, MAX_MESSAGE_LENGTH,fp);
	  n = sscanf(line, " %s %s",myAddrStr, neighborAddrStr);
	  if(n < 2)
	       continue;

	  n = inet_pton(AF_INET, myAddrStr, &myAddr);
	  if(n <= 0)
	       continue;
	  if(!memcmp(&serv->myIpAddr, &myAddr, sizeof(struct in_addr))) {
	       n = inet_pton(AF_INET, neighborAddrStr, &neighborAddr);
	       if( n <= 0)
		    continue;
	
	       tmp = (DebugTable *)malloc(sizeof(DebugTable));
	       if(tmp == NULL)
		    err_quit("Memory allocation failed\n");
	       memcpy(&tmp->addr, &neighborAddr, sizeof(struct in_addr));
	       tmp->next = serv->debugTable;
	       serv->debugTable = tmp;
	  }
     }
     fclose(fp);
}

/********************** Read in information from configuration file ************/
void
ReadChannelInfo(ServerData *serv, const char *filename) {
     HelloServerData *helloData = serv->helloData;
     FILE *fp;
     char line[MAX_MESSAGE_LENGTH];
     char key[MAX_MESSAGE_LENGTH], data[MAX_MESSAGE_LENGTH];
     int n, chan, totalChannels = 0, totalInterfaces = 0;
     SwitchPolicy policy = BALANCED_TWOHOP;
     int fixedChannel;

     fp = fopen(filename,"r");
     if(fp == NULL)
	  err_quit("In ReadChannelInfo: Cannot open %s", filename);
    
     while(!feof(fp)) {
	  if(fgets(line, MAX_MESSAGE_LENGTH,fp) == NULL)
	       break;

	  n = sscanf(line, " %s %s",key, data);
	  if(n < 2)
	       continue;
	  if(!strcasecmp("CHANNEL",key)) {
	       n = sscanf(data, "%d", &chan);
	       if( n < 1)
		    continue;

	       if(totalChannels >= TOTAL_CHANNEL)
		    err_quit("We only support %d channels in total\n", TOTAL_CHANNEL);
	       helloData->channelList[totalChannels] = chan;
	       totalChannels++;
	  } else if(!strcasecmp("INTERFACE", key)) {
	       char *name = (char *)malloc(strlen(data)+1);
	       strcpy(name, data);
	       if(totalInterfaces >= TOTAL_CHANNEL)
		    err_quit("We only support %d interfaces in total\n", TOTAL_CHANNEL);
	       helloData->inter[totalInterfaces].name = name;

	       // convention: 1st interface is called fixed interface
	       if(totalInterfaces == 0)
		    helloData->inter[totalInterfaces].isFixed = TRUE;
	       else
		    helloData->inter[totalInterfaces].isFixed = FALSE;
	       totalInterfaces ++;
	  } else if(!strcasecmp("FIXED-CHANNEL", key)) {
	       n = sscanf(data, "%d", &fixedChannel);
	       if( n < 1)
		    continue;
	       policy = FIXED;
	  } 
	  else if(!strcasecmp("GATEWAY", key)) {
	       if(serv->gatewayData->isGatewayNode == FALSE)
	       {
		    serv->gatewayData->isGatewayNode = TRUE;
		    Inet_pton(AF_INET, data, &serv->gatewayData->internetGateway);

	       } // else - GATEWAY entries after first ignored
	  } else if(!strcasecmp("METRIC", key)) {
	    
	       if(!strcasecmp("MCETT", data))
		    serv->routingData->routeMetricType = MCR_MCETT;
	       else if(!strcasecmp("WCETT", data))
		    serv->routingData->routeMetricType = MCR_WCETT;
	       else if(!strcasecmp("SHORTEST_DIVERSITY", data))
		    serv->routingData->routeMetricType = MCR_SHORTEST_HOP_AND_DIVERSITY;
	       else if(!strcasecmp("SHORTEST", data))
		    serv->routingData->routeMetricType = MCR_SHORTEST_HOP;	    
	       else
		    err_quit("Channel file read: Invalid metric %s\n",data);
	  }
     }
     fclose(fp);

     // Check that the values read in are legitimate

     if(totalInterfaces > totalChannels)
	  err_quit("Illegal: Interfaces (%d) is greater than channels (%d)\n", 
		   totalInterfaces, totalChannels);

     if(totalInterfaces < 1)
	  err_quit("Illegal: At least one interface is required\n"); 

     helloData->totalChannels = totalChannels;
     helloData->totalInterfaces = totalInterfaces;

     if(policy == FIXED) {
	  helloData->fixedChannel = fixedChannel;
	  helloData->switchPolicy = FIXED;
     }
     else
	  helloData->fixedChannel = helloData->channelList[0];

}

/******************** Handle periodic timeout for printing out status information  *****/
void  
SetStatusTimeout(ServerData *serv)
{
     struct timeval delay;
     delay.tv_sec = STATUS_TIMEOUT_VALUE;
     delay.tv_usec = 0;
     AddTimeout(serv, &delay, STATUS_TIMEOUT, 0, 0);
}


void 
StatusTimeout(ServerData *serv)
{
     char localAddr[INET_ADDRSTRLEN];
     err_msg("(%s): I am alive\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr))); 
    
     SetStatusTimeout(serv);
}

/********************  De-Multiplex and execute command *************/

/* This function executes control commands given by another process.
 */
void
ExecuteCommand(ServerData *serv, const char *pkt, int size)
{
     struct command cmd;
     if(size < sizeof(struct command)) {
	  err_msg("Invalid command packet received\n");
	  return;
     }
     copyStructFromPkt(cmd,pkt);
     switch(cmd.option)
     {
	  case CHANGE_METRIC:
	  {
	       serv->routingData->routeMetricType = cmd.field.metric;
#ifdef DEBUG_MSG_ON
	       err_msg("Metric:Changed metric type to %d , CHAN_REUSE=%d\n",
			       serv->routingData->routeMetricType, MCR_CHAN_REUSE_CETT);
#endif
	       break;
	  }
	  case CHANGE_BETA:
	  {
	       serv->routingData->beta = cmd.field.beta;
#ifdef DEBUG_MSG_ON
		err_msg("Metric:Changed metric beta to %f\n",
			       serv->routingData->beta);
#endif

	       break;
	  }
	  case CHANGE_ALPHA:
	  {
	       serv->routingData->alpha = cmd.field.alpha;
#ifdef DEBUG_MSG_ON
	       err_msg("Metric:Changed metric alpha to %f\n",
			       serv->routingData->alpha);
#endif

	       break;
	  }
	  case FREEZE_METRICS:
	  {
	       /* Freezing the link quality value of neighbors.
		* Used while conducting throughput experiments.
		* Hello Messages get dropped when high data rate
		* flows are established. This results in perceived 
		* link quality degradation, though the link is still good.
		* Thus we freeze link quality value when throughput 
		* experiment is in progress
		*/
	       serv->freeze_metrics = TRUE;
#ifdef DEBUG_MSG_ON
	       err_msg("Metric:Metrics frozen\n");
#endif
	
	       break;	       
	  }
	  case UNFREEZE_METRICS:
	  {
		  serv->freeze_metrics = FALSE;
#ifdef DEBUG_MSG_ON
	       err_msg("Metric:Metrics unfrozen\n");
#endif
	       break;
		  
	  }
	  case SUPPRESS_DEBUG_MSG:
	  {
		  suppress_debug_msg = TRUE;
		  break;
	  }
	  case START_DEBUG_MSG:
	  {
		  suppress_debug_msg = FALSE;
		  break;
	  }
	  default: err_msg("Invalid command option received");
     }
   
} 


/********************  De-multiplex  protocol messages **************/
void
ProcessPacket(ServerData *serv, const char *pkt, int size)
{
     McrPacketHeader hdr;

     char localAddr[INET_ADDRSTRLEN];

     if(size < sizeof(McrPacketHeader)) {
	  err_msg("Invalid packet received\n");
	  return;
     }
     copyStructFromPkt(hdr,pkt);
     size-=sizeof(hdr);
     if(size != hdr.len) {
	  err_msg("ERROR:integrated.c:Payload of packet corrupted size = %d, hdr_len = %d\n", size, hdr.len);
	  return;
     }
    
     // Do not process my own packets
     if( !memcmp(&hdr.fromAddr, &serv->myIpAddr, sizeof(struct in_addr)) )
	  return;

     if(debug_mode) 
	  if(PacketNotFromNeighbor(serv, &hdr.fromAddr))
	       return;
     // Pass on packet to appropriate module    
     switch(hdr.type) {
     case LINK_HELLO_PKT:
	  ProcessHello(serv, &hdr, pkt + sizeof(hdr), hdr.len);
	  break;
    
     case ROUTE_PKT:
	  ProcessRoutingPacket(serv, &hdr, pkt + sizeof(hdr), hdr.len);
	  break;

     case GATEWAY_PKT:
	  ProcessGatewayPacket(serv, &hdr, pkt + sizeof(hdr), hdr.len);
	  break;

     default:
	  err_msg("(%s): Received an invalid message of type %d\n",
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)), hdr.type);
     }
}


/************* Main loop to read protocol packets and invoke timeouts *********************/

void
ProcessMessages(ServerData *serv)
{
     struct sockaddr_in cliaddr;
     socklen_t len = sizeof(cliaddr);
     int n, numReady;
     char pkt[MAX_MESSAGE_LENGTH];
     char cmnd[sizeof(struct command)+1];
     struct timeval *timeout;
     fd_set rset;
     int receivefd;

     receivefd = (serv->receivefd > serv->netlinkfd) ? serv->receivefd: serv->netlinkfd;    
     receivefd = (serv->cmnd_recv_fd > receivefd) ? serv->cmnd_recv_fd : receivefd ;
     
     while(1) {
	  FD_ZERO(&rset); // Can this be moved outside the loop?
	  FD_SET(serv->receivefd, &rset);
	  FD_SET(serv->netlinkfd, &rset);
	  FD_SET(serv->cmnd_recv_fd, &rset);

// For simplicity, the userspace daemon is single-threaded. Therefore, in a main loop
// we have to handle timeouts while waiting for protocol packets

	  timeout = GetNextTimeout(serv);
	  if( (numReady = select(receivefd+1, &rset, NULL, NULL, timeout)) < 0) {
	       if(errno == EINTR) {
		    err_msg("Received an interrupt\n");
		    continue;
	       }
	       else
		    err_sys("Select returned an error\n");
	  }

	  if(numReady == 0) {
	       ExecuteTimeout(serv);
	       continue;
	  } 

	  if(FD_ISSET(serv->receivefd, &rset))
	  {
	       n = Recvfrom(serv->receivefd, pkt, MAX_MESSAGE_LENGTH, 0, (SA *)&cliaddr, &len);
	       ProcessPacket(serv, pkt, n);
	       numReady --;
	  }

	  if(FD_ISSET(serv->netlinkfd, &rset))
	  {
	       ReadNetlinkSocket(serv);
	       numReady--;
	  }

	  if(FD_ISSET(serv->cmnd_recv_fd, &rset))
	  { 
	       n = Recvfrom(serv->cmnd_recv_fd, cmnd, (sizeof(struct command) + 1), 0, (SA *)&cliaddr, &len);
	       ExecuteCommand(serv, cmnd, n);
	       numReady--;
	  }


	  if(numReady != 0) 
	       err_msg("Surprise! numReady not zero\n");
     }
}


/****************  Exit functions to convert process into a daemon ************/

void 
sigsegv_handler(int type)
{
     int fd_err = open("/root/sigsegv.txt",O_WRONLY | O_CREAT);
     char errMsg[45];
     sprintf(errMsg, "Received Signal of type %d \n", type);
     if(fd_err >= 0)
     {     
	   write(fd_err, errMsg ,30);
	   write(fd_err, "Going in an infinite loop, so u can use gdb\n",48);
	   close(fd_err);
     }
     err_msg("Received a Signal of type %d\n", type);
     
     // infinite loop
     // One can now use gdb to figure out where and why sigsegv was received.
     while(1)
     {
     }
     exit(0);
}


void 
signal_handler(int type)
{
     int fd_err = open("/root/error.txt",O_WRONLY | O_CREAT);
     char errMsg[45], signalNo[15];
     sprintf(errMsg, "Received Signal of type %d ,Dying\n", type);
     if(fd_err >= 0)
     {     
           sprintf(signalNo, "SIGHUP:%d\n", SIGHUP);
	   write(fd_err,signalNo, 13);
    	   sprintf(signalNo, "SIGINT:%d\n", SIGINT);
	   write(fd_err,signalNo, 13);
 	   sprintf(signalNo, "SIGTERM:%d\n", SIGTERM);
	   write(fd_err,signalNo, 13);
           sprintf(signalNo, "SIGSEGV:%d\n", SIGSEGV);
	   write(fd_err,signalNo, 13);
           sprintf(signalNo, "SIGILL:%d\n", SIGILL);
	   write(fd_err,signalNo, 13);
           sprintf(signalNo, "SIGABRT:%d\n", SIGABRT);
	   write(fd_err,signalNo, 13);
           sprintf(signalNo, "SIGKILL:%d\n", SIGKILL);
	   write(fd_err,signalNo, 13);
	   write(fd_err, errMsg ,45);
	   close(fd_err);
     }
     err_msg("Received a Signal of type %d Dying\n", type);
     exit(1);
}

static void
cleanup(void) 
{
     NeighborEntry *list, *tmp;
     list = serverData->helloData->neighborTable;
     while(list != NULL) {
	  tmp = list;
	  free(tmp);
	  list = list->next;
     }
     close(serverData->netlinkfd);
     free(serverData);
}


/**************** Options that can be read in from command line *************/
struct option longopts[] = {
     {"broadcastIp", required_argument, NULL, 'b'},
     {"myIp", required_argument, NULL, 'm'},
     {"bondInterface", required_argument, NULL, 'i'},
     {"helloInterval", required_argument, NULL, 'h'},
     {"neighborTtl", required_argument, NULL, 't'},
     {"switchProb", required_argument, NULL, 's'},
     {"channelFile", required_argument, NULL, 'c'},
     {"daemonize", no_argument, NULL, 'd'},
     {"debugfile", required_argument, NULL, 'D'},
     {0}
};

void usage(void)
{
     printf
	  ("\nUsage: helloServer -b ipaddr -m ipaddr [-h #num] [-t #num]" 
	   "[-s #num] [-c filename] [-p #num] [-d] [-D filename]\n\n"
	   "-b, --broadcastIp   ipaddr   Ipaddress used for sending data\n"
	   "-m, --myIp          ipaddr   Local node ip address\n"
	   "-h, --helloInterval #int     Time interval between hello packets (in seconds)\n"
	   "-t, --neighborTtl   #int     Number of hello attempts that are missed\n"
	   "-s, --switchProb    #float   Probability used for switching fixed channel\n"
	   "-c, --channelFile   #name    File with information on channels and interfaces\n"
	   "-d, --daemonize              Run as a daemon\n"
	   "-D, --debugfile     #name    Debug file used to specify topology\n"
	   "-p, --min_pack_size #int     Minimum packet size for the hello packets, (default = 100)\n"
	   "\n\n"
	       );

     exit(1);
}


/************** Read in options and initialize all protocol components **********/

int
main(int argc, char **argv)
{
     int receivefd, sendfd, cmnd_recv_fd;
     const int on = 1;
     ServerData *serv;
     char *filename = NULL, *debugFilename = NULL, *broadcastIp = NULL, *myIp = NULL, *bondInterface = NULL;
     
     int n;
     double d;     
     unsigned int tos;
     errMsgFile=fopen("/root/helloServerErrMsg.txt","w");
     if(errMsgFile == NULL)
     {
          err_msg("Could not open file /root/helloServerErrMsg.txt");
     }

     neighborFile=fopen("/root/Neighbors.txt","w");
     if(neighborFile == NULL)
     {
          err_msg("Could not open file /root/Neighbors.txt");
     }

     
#ifdef ROUTE_EXPERIMENT
     routeDebugMsgFile = fopen("/root/routeExperiment.txt","w");
     routeNotTakenMsgFile = fopen("/root/routesNotTaken.txt", "w");
     if(routeDebugMsgFile == NULL)
     {
          err_msg("Could not open file /root/routeExperiment.txt");
     }
     if(routeNotTakenMsgFile == NULL)
     {
          err_msg("Could not open file /root/routeNotTaken.txt");
     }


#endif

     
     serv = (ServerData *) malloc(sizeof(ServerData));
     if(serv == NULL)
	  err_quit("Could not allocate memory for the server\n");
     serverData = serv; // assign to a global variable for cleanup
     serv->timerQueue = NULL;
     serv->freeze_metrics = FALSE;
     srandom(time(NULL)+getpid());    // initialize seed for random number generator
     
     HelloInit(serv);
     RoutingInit(serv);
     GatewayInit(serv);

     /* Catch SIGHUP, SIGINT and SIGTERM type signals */
     signal(SIGHUP, signal_handler);
     signal(SIGINT, signal_handler);
     signal(SIGTERM, signal_handler);
     signal(SIGILL, signal_handler);
     signal(SIGABRT, signal_handler);
     signal(SIGKILL, signal_handler);
#ifdef DEBUG_MSG_ON
     signal(SIGSEGV, sigsegv_handler);
#else
     signal(SIGSEGV, signal_handler);
#endif

     // Parse options
     while (1) {
	  int opt;
	
	  opt = getopt_long(argc, argv, "b:m:i:h:t:s:p:c:dD:", longopts, 0);
	
	  if (opt == EOF)
	       break;

	  switch (opt) {
	  case 0:
	       break;

	  case 'p':
	       /* Minimum size of hello messages.
		* Should be similar to actual data packet size for link quality estimation to be reliable
		*/
	       n = atoi(optarg);
	       if (n <= 0) {
		    fprintf(stderr, "-p should be at least 1!\n");
		    exit(1);
	       }
	       else if(n > MAX_MESSAGE_LENGTH)
	       {
		   fprintf(stderr, "-p should be less than MAX_MESSAGE_LENGTH = %d\n", MAX_MESSAGE_LENGTH);
		   exit(1); 
	       }
	       serv->helloData->min_hello_packet_size = n;
	       break;
	  case 'b':
	       broadcastIp = optarg;
	       break;

	  case 'm':
	       myIp = optarg;
	       break;

	  case 'i':
	       bondInterface = optarg;
	       break;

	  case 'h':
	       	n = atoi(optarg);
	       if (n == 0) {
		    fprintf(stderr, "-n should be at least 1!\n");
		    exit(1);
	       }
       	       serv->helloData->helloInterval = n;
	       break;

	  case 't':
	       n = atoi(optarg);
	       if (n == 0) {
		    fprintf(stderr, "-t should be at least 1!\n");
		    exit(1);
	       }
	       serv->helloData->neighborTtl = n;
	       break;

	  case 's':
	       d = atof(optarg);
	       if (d > 1 || d <= 0) {
		    fprintf(stderr, "-d should lie between 0 and 1!\n");
		    exit(1);
	       }
	       serv->helloData->switchProb = d;
	       break;

	  case 'c':
	       filename = optarg;
	       break;
	    
	  case 'd':
	       daemon_proc = 1;
	       break;

	  case 'D':
	       debugFilename = optarg;
	       break;

	  case ':':
	  case '?':
	       usage();
	  default:
	       usage();

	  }
     }

// Verify options that had to be specified have been indeed specified
     if(broadcastIp == NULL || myIp == NULL)
	  usage();

     Inet_pton(AF_INET, myIp, &serv->myIpAddr);
    
     serv->debugTable = NULL;
     if(debugFilename) {
	  debug_mode = LOCAL_ONLY;
	  printf("In debug mode\n");
	  ReadDebugTable(serv, debugFilename);
     }

     if(filename == NULL)
	  ReadChannelInfo(serv, "channel.txt");
     else
	  ReadChannelInfo(serv, filename);

     serv->seq = 1;
     
     
// Add self to neighbor table
     AddToNeighborTable(serv, &serv->myIpAddr, serv->helloData->fixedChannel, serv->helloData->totalInterfaces, INDIRECT, SYMMETRIC, 0);

     if(bondInterface == NULL)
	  err_quit("Bond interface name not given\n");


     if(serv->helloData->totalInterfaces != 1)
	  strncpy(serv->ifr.ifr_name, bondInterface, IFNAMSIZ + 1);
     else
	  strncpy(serv->ifr.ifr_name, serv->helloData->inter[0].name, IFNAMSIZ + 1);
     serv->ioctlSock = Socket(AF_INET, SOCK_DGRAM, 0);
    
     err_msg("Initialization (%s): Channels = %d, Interfaces = %d bond = %s", 
	     myIp, serv->helloData->totalChannels, serv->helloData->totalInterfaces, serv->ifr.ifr_name);

     InitHelloKernelTable(serv);

     
     err_msg("Just finished with InitHelloKernelTable\n");
     
     if(daemon_proc) {
	  if (fork() != 0)
	       exit(0);
	  /* Close stdin, stdout and stderr... */
	  close(0);
	  close(1);
	  close(2);
	  setsid();
     }

     atexit(&cleanup);
     
// Socket for inter process communication
// Receving commands to change variables from another process
     cmnd_recv_fd = Socket(AF_INET, SOCK_DGRAM, 0);
     Setsockopt(cmnd_recv_fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
     bzero( &serv->cmnd_recv_sock, sizeof(struct sockaddr_in));
     serv->cmnd_recv_sock.sin_family = AF_INET;
     serv->cmnd_recv_sock.sin_addr.s_addr = htonl(INADDR_ANY);
     serv->cmnd_recv_sock.sin_port = htons(COMMAND_LISTEN_PORT);
     Bind(cmnd_recv_fd, (SA *) &serv->cmnd_recv_sock, sizeof(struct sockaddr_in));
     serv->cmnd_recv_fd = cmnd_recv_fd;

     
// Socket for receiving data
     receivefd = Socket(AF_INET, SOCK_DGRAM, 0);
     Setsockopt(receivefd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
     bzero( &serv->recvSocket, sizeof(struct sockaddr_in));
     serv->recvSocket.sin_family = AF_INET;
     serv->recvSocket.sin_addr.s_addr = htonl(INADDR_ANY);
     serv->recvSocket.sin_port = htons(HELLO_PORT);
     Bind(receivefd, (SA *) &serv->recvSocket, sizeof(struct sockaddr_in));

// Socket for sending out data
     sendfd = Socket(AF_INET, SOCK_DGRAM, 0);
     Setsockopt(sendfd, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on));
     
     /* Setting the TOS field for the socket for broadcast packets.
      * All broadcast packets send from this socket will be management packets.
      * These packets should get higher priority than the data packets.
      */
     tos = 0xc0;
     Setsockopt(sendfd, IPPROTO_IP, IP_TOS, &tos, sizeof(tos));
     
     bzero( &serv->sendBroadcastSocket, sizeof(struct sockaddr_in));
     serv->sendBroadcastSocket.sin_family = AF_INET;
     Inet_pton(AF_INET, broadcastIp, &serv->sendBroadcastSocket.sin_addr);
     serv->sendBroadcastSocket.sin_port = htons(HELLO_PORT);
 
     serv->sendfd = sendfd;
     serv->receivefd = receivefd;

     bzero(&serv->sendUnicastSocket, sizeof(struct sockaddr_in));
     serv->sendUnicastSocket.sin_family = AF_INET;
     serv->sendUnicastSocket.sin_port = htons(HELLO_PORT);

     err_msg("Setting all timers\n");
     
     SetHelloTimeout(serv);
     SetStatusTimeout(serv);
     SetRouteTimeout(serv);
     SetGatewayTimeout(serv);

// A single interface node needs to choose a channel having a multi-interface node on it
     if(serv->helloData->totalInterfaces == 1)
	  StartScanChannelIfRequired(serv, -1); // Initial channel scan if single interface

     err_msg("Helloserver started successfully\n");

     ProcessMessages(serv); // main loop

     return 1;   // Program never reaches here.
}
