/*****************************************************************************
 * mcr_gateway.c: Support for gateways
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 *
 *****************************************************************************/


/*************************** NOTES *******************************************

This file implements userspace support for gateways. The key features
in this file are:
1) Support for sending out gateway beacon (advertisement) messages.
2) Support for choosing a gateway when multiple gateways are available
3) Interaction with kernel multichannel routing module to setup access to 
    gateway

*********************************************************************************/

#include "utilities.h"
#include "integrated.h"
#include "timer.h"
#include "mcr_gateway.h"
#include "mcr_ioctl.h"
#include "kmcr_netlink.h"

/************* Interaction with kernel multichannel routing module **************/

void
ChangeDefaultGateway(ServerData *serv, struct in_addr *oldAddr,
		     struct in_addr *newAddr)
{
     struct in_addr defaultAddr;

     // default route
     Inet_pton(AF_INET, "0.0.0.0", &defaultAddr);

     // Update kernel with new gateway information
     if(oldAddr != NULL && newAddr == NULL)
	  Netlink_create_and_send_msg(serv, KMCR_GATEWAY, &defaultAddr, &defaultAddr, 0, 0);
     else if(newAddr != NULL)
	  Netlink_create_and_send_msg(serv, KMCR_GATEWAY, newAddr, newAddr, 0, 0);

     if(oldAddr != NULL)
	  Ioctl_DeleteRoute(serv, &defaultAddr, oldAddr);
     if(newAddr != NULL)
	  Ioctl_AddRoute(serv, &defaultAddr, newAddr);
}

/**************** Gateway table management *************************/
void
DeleteOldGatewayEntries(ServerData *serv)
{
     GatewayData *data = serv->gatewayData;
     GatewayEntry *cur = data->gatewayList, *prev = NULL;
     struct in_addr oldAddr;
     BOOL gatewayChanged = FALSE;

     while(cur != NULL)
     {
	  if(cur->ticks <= 0)
	  {
	       if(prev == NULL)
		    data->gatewayList = cur->next;
	       else
		    prev->next = cur->next;
	
	       if(cur == data->currentGateway) {
		    memcpy(&oldAddr, &data->currentGateway->addr, 
			   sizeof(struct in_addr));
		    gatewayChanged = TRUE;
		    data->currentGateway = NULL;
	       }
	       free(cur);

	       if(prev == NULL)
		    cur = data->gatewayList;
	       else
		    cur = prev->next;
	  } 
	  else
	  {
	       cur->ticks --;
	       prev = cur;
	       cur = cur->next;
	  }
     }

     // Update default gateway if required
     if(gatewayChanged == TRUE && data->gatewayList != NULL)
     {
	  int min = 100000; // some large number of hops
	  cur = data->gatewayList;

	  // Right now, we choose the closest gateway as the default gateway. It should be 
	  // trivial to implement other metrics for gateway selection
	  while(cur != NULL)
	  {
	       if(cur->hops < min) {
		    min = cur->hops;
		    prev = cur;   // prev is being reused here
	       }
	       cur = cur->next;
	  }
	
	  data->currentGateway = prev;
	  ChangeDefaultGateway(serv, &oldAddr, &data->currentGateway->addr);

     } else if(gatewayChanged == TRUE && data->gatewayList == NULL)
	  ChangeDefaultGateway(serv, &oldAddr, NULL);
}

void
UpdateGatewayInfo(ServerData *serv, const GatewayPacketHeader *hdr)
{
     GatewayData *data = serv->gatewayData;
     GatewayEntry *list = data->gatewayList;

     // If I am the gateway, ignore
     if(!memcmp(&serv->myIpAddr, &hdr->addr, sizeof(struct in_addr) ))
	  return;

     while(list != NULL)
     {
	  // Update entry
	  if(!memcmp(&list->addr, &hdr->addr, sizeof(struct in_addr)))
	  {
	       list->seq = hdr->seq;
	       list->ticks = GATEWAY_ENTRY_TICKS;
	       list->hops = hdr->hops;
	       return;
	  }
	  list = list->next;
     }

     list = (GatewayEntry *) malloc( sizeof(GatewayEntry));
     list->next = data->gatewayList;
     data->gatewayList = list;

     list->seq = hdr->seq;
     list->ticks = GATEWAY_ENTRY_TICKS;
     list->hops = hdr->hops;
     memcpy(&list->addr, &hdr->addr, sizeof(struct in_addr));
    
// Check if default gateway has to be changed
// Current metric is to use shortest hop count    
     if(data->currentGateway == NULL || data->currentGateway->hops > list->hops)
     {
	  if(data->currentGateway == NULL)
	       ChangeDefaultGateway(serv, NULL, &list->addr);
	  else
	       ChangeDefaultGateway(serv, &data->currentGateway->addr, &list->addr);

	  data->currentGateway = list;
     }
	
}

/*************** Handle gateway beacons *****************************/

BOOL
IsNewBeacon(ServerData *serv, const GatewayPacketHeader *hdr)
{
     GatewayData *data = serv->gatewayData;
     GatewayEntry *list = data->gatewayList;

     while(list != NULL)
     {
	  if(!memcmp(&list->addr, &hdr->addr, sizeof(struct in_addr)))
	  {
	       if(hdr->seq > list->seq)
		    return TRUE;
	       else
		    return FALSE;
	  }
	  list = list->next;
     }
     return TRUE;

}

void
SendGatewayBeacon(ServerData *serv, const struct in_addr *src, unsigned long seq, int ttl, int hops)
{
     McrPacketHeader mcrHdr;
     GatewayPacketHeader gHdr;
     char packet[MAX_MESSAGE_LENGTH];
     socklen_t len = sizeof(struct sockaddr_in);

     gHdr.seq = seq;
     gHdr.ttl = ttl;
     gHdr.hops = hops;
     memcpy(&gHdr.addr, src, sizeof(struct in_addr));
    
     mcrHdr.len = sizeof(GatewayPacketHeader);
     mcrHdr.type = GATEWAY_PKT;
     memcpy(&mcrHdr.fromAddr, &serv->myIpAddr, sizeof(struct in_addr));

     copyStructIntoPkt(mcrHdr, packet);
     copyStructIntoPkt(gHdr,packet+sizeof(mcrHdr));
    
     Sendto(serv->sendfd, packet, sizeof(mcrHdr)+mcrHdr.len, 0, (SA *) &serv->sendBroadcastSocket, len);
}

// Receive and forward gateway messages, if necessary
void
ProcessGatewayPacket(ServerData *serv, const McrPacketHeader *hdr, const char *dataPkt, int dataSize)
{
     GatewayPacketHeader gHdr;

     copyStructFromPkt(gHdr, dataPkt);

     if( gHdr.ttl > 0 && IsNewBeacon(serv, &gHdr) )
     {
	  UpdateGatewayInfo(serv, &gHdr);
	  if(gHdr.ttl > 1)
	       SendGatewayBeacon(serv, &gHdr.addr, gHdr.seq, gHdr.ttl - 1, gHdr.hops++);
     }

}

/******************* Timer management ***************************/

void
SetGatewayTimeout(ServerData *serv)
{
     struct timeval delay;
    
     delay.tv_sec = GATEWAY_BEACON_INTERVAL;
     delay.tv_usec = 0;
     AddTimeout(serv, &delay, GATEWAY_TIMEOUT, 0, 0);
}

void
GatewayTimeout(ServerData *serv)
{
     GatewayData *data = serv->gatewayData;
     DeleteOldGatewayEntries(serv);
    
     if(data->isGatewayNode == TRUE)
     {
	  SendGatewayBeacon(serv, &serv->myIpAddr, data->seq++, GATEWAY_MAX_TTL, 0);
	
	  if(data->gatewaySet == FALSE) {
	       struct in_addr defaultAddr;
	    
	       data->gatewaySet = TRUE;
	       Inet_pton(AF_INET, "0.0.0.0", &defaultAddr);

	       // Commented out code below that would automatically setup the wired interface
	       // at the gateway node
	       // FIX ADDROUTE to use "eth0" as the interface address
	       //    Ioctl_AddRoute(serv, &defaultAddr, &data->internetGateway);
	  }
     }
     SetGatewayTimeout(serv);

}

/********************* Initialization **************************************/

void
GatewayInit(ServerData *serv)
{
     GatewayData *gatewayData;

     gatewayData = (GatewayData *) malloc( sizeof(GatewayData));
     gatewayData->isGatewayNode = FALSE;
     gatewayData->seq = 1;
     gatewayData->gatewayList = NULL;
     gatewayData->currentGateway = NULL;
     gatewayData->gatewaySet = FALSE;

     serv->gatewayData = gatewayData;
}
