/*****************************************************************************
 * mcr_hello.c: Interface management protocol
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 * Modifications by: Rishi Bhardwaj, <rbhardw2 at uiuc dot edu>
 * 
 *****************************************************************************/


/*************************** NOTES *******************************************

This file implements the interface management protocol  that is responsible
for channel assignment. The main functionality in this file are:

1) Support for exchanging "hello" messages with neighbors to learn channels
    being used by neighbors.
2) Implement different policies for channel assignment.
3) Maintain a neighbor table to store information about neighbors. Keep track
    of neighbors, and identify unreachable neighbors.
4) Initalize and control kernel channel abstraction module.

*********************************************************************************/

#include "utilities.h"
#include "integrated.h"
#include "timer.h"
#include "mcr_hello.h"
#include "mcr_route.h"
#include "mcr_ioctl.h"
#include "kmcr_netlink.h"


extern FILE* neighborFile;
extern int neighbor_file_updation_ctr;

/************** Interaction with kernel channel abstraction module ****************/

void
InitHelloKernelTable(ServerData *serv)
{
     int i,j, nonFixed;
     HelloServerData *helloData = serv->helloData;

     // Bonding module is not loaded if there is a single interface
     if(helloData->totalInterfaces == 1)
	  return;

     // Step 1: Add valid channels
     for(i = 0; i < helloData->totalInterfaces; i++) 
	  for(j = 0; j < helloData->totalChannels; j++)
	       Ioctl_AddValidChannel(serv, helloData->inter[i].name, helloData->channelList[j]);

     // Step 2: Setup broadcast table
     // currently we assign all non-fixed channels to one non-fixed interface
     // that policy can be changed
    
     for(i = 0; i < helloData->totalInterfaces; i++)
	  if(helloData->inter[i].isFixed == FALSE)
	       break;
     nonFixed = i;

     for(i = 0; i < helloData->totalChannels; i++) {
	  if(helloData->channelList[i] == helloData->fixedChannel)
	       Ioctl_AddBroadcastEntry(serv, helloData->inter[helloData->fixedInterface].name, helloData->fixedChannel);
	  else
	       Ioctl_AddBroadcastEntry(serv, helloData->inter[nonFixed].name, helloData->channelList[i]);
		
     }

     for(i = 0; i < helloData->totalChannels; i++)
	  if(helloData->channelList[i] != helloData->fixedChannel)
	       break;
     nonFixed = i;

     // Step 3: Switch each interface to a valid channel
     for(i = 0; i < helloData->totalInterfaces; i++) {
	  if(helloData->inter[i].isFixed == TRUE)
	  {
	       struct in_addr defaultAddr;
	       Ioctl_SwitchChannel(serv, helloData->inter[i].name, helloData->fixedChannel);

	       // Add a default entry, for sending out unicast packets that have no other entry
	       inet_pton(AF_INET, "0.0.0.0", &defaultAddr); // Default address is represented by IP 0.0.0.0
	       Ioctl_AddUnicastEntry(serv, &defaultAddr,  helloData->inter[i].name, helloData->fixedChannel);
	  }
	  else
	       Ioctl_SwitchChannel(serv, helloData->inter[i].name, helloData->channelList[nonFixed]);
     }
}

// Add an unicast entry into channel abstraction module
void
UpdateKernelTableForUnicast(ServerData *serv, NeighborEntry *entry)
{
     int interface;
     HelloServerData *helloData = serv->helloData;
     RemoteNode *remoteList = entry->remoteList;
    
     if(helloData->totalInterfaces == 1)
     {
	  Ioctl_AddUnicastEntry(serv, &entry->addr, NULL, entry->channel);
	  return;
     }
     if(entry->channel == helloData->fixedChannel)
	  interface = helloData->fixedInterface;
     else {
	  for(interface = 0; interface < helloData->totalInterfaces; interface++)
	       if(helloData->inter[interface].isFixed == FALSE)
		    break;
		    
	  if(interface == helloData->totalInterfaces)
	       err_quit("Could not find a non-fixed interface\n");
     }
     entry->interface = interface;
  
     //Update all remote nodes that depended on this node
     while(remoteList != NULL)
     {

	  Ioctl_AddUnicastEntry(serv, &remoteList->addr, helloData->inter[interface].name, entry->channel);
	  remoteList = remoteList->next;
     }
}

void
SwitchFixedChannel(ServerData *serv, int oldChannel, int newChannel)
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable;
     int nonFixed;
     struct in_addr defaultAddr;
      
     for(nonFixed = 0; nonFixed < helloData->totalInterfaces; nonFixed++)
	  if(helloData->inter[nonFixed].isFixed == FALSE)
	       break;
     if(nonFixed == helloData->totalInterfaces)
	  err_quit("SwitchFixedChannel: Could not find a non-fixed interface");

     while(list != NULL) {
	  if( (list->type & DIRECT) == DIRECT && ( list->channel == oldChannel || list->channel == newChannel) )
	       UpdateKernelTableForUnicast(serv, list); 
	  list = list->next;
     }

     Ioctl_AddBroadcastEntry(serv, helloData->inter[nonFixed].name, oldChannel);
     Ioctl_AddBroadcastEntry(serv, helloData->inter[helloData->fixedInterface].name, newChannel);

     // Add a default entry, for sending out unicast packets that have no other entry
     inet_pton(AF_INET, "0.0.0.0", &defaultAddr); // Default address is represented by IP 0.0.0.0
     Ioctl_AddUnicastEntry(serv, &defaultAddr,  helloData->inter[helloData->fixedInterface].name, newChannel);
     Ioctl_SwitchChannel(serv, helloData->inter[helloData->fixedInterface].name, newChannel);
}

/**************** Functions that operate on neighbor table *********************************/

/* This function returns the number of ones in the different bit positions in x
 * Author: Rishi Bhardwaj (rbhardw2 at uiuc dot edu) , June 19, 2007. 
 */

unsigned short numberOfOnes(LINK_QUALITY_INT x)
{
     unsigned short num =0;
     while(x)
     {
	  num++;
	  x ^= (x & ((~x) + 1));
     }
     return num;
}


// function to estimate quality of link to neighbor, used by routing metric
float
GetNeighborDeliveryRatio(ServerData *serv, const struct in_addr *addr)
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable;
     float successProb, fwdProb, revProb;
     while(list != NULL) {
	  if(!memcmp(&list->addr, addr, sizeof(struct in_addr)))
	  { 
	       fwdProb = numberOfOnes(list->fwdLinkQuality)/((float) sizeof(list->fwdLinkQuality)*BITS_IN_BYTE);
	       revProb = numberOfOnes(list->revLinkQuality)/((float) sizeof(list->revLinkQuality)*BITS_IN_BYTE);
#ifdef DEBUG_MSG_ON
		    err_msg("mcr_hello.c:debug_msg:ETX:fwdProb = %f , revProb = %f , neighb %s \n", fwdProb, revProb, print_ip(addr->s_addr));
#endif
	       successProb = fwdProb * revProb;
	       return (1/successProb);
	  }
	
	  list = list->next;
     }
     return 0;
}

/*Function to check if a node is one hop symmetrical neighbor
 * Author: Rishi Bhardwaj
 */
BOOL IsOneHopNeighbor(ServerData *serv, const struct in_addr *addr)
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable;
     
     while(list != NULL) {
	  if(!memcmp(&list->addr, addr, sizeof(struct in_addr)))
	  {
	       if((list->type & DIRECT) && (list->link == SYMMETRIC)) 
		    return TRUE;
	       else
	       {
		    return FALSE;
	       }	    
	  }
	    	
	  list = list->next;
     }
     return FALSE;
	
}


BOOL
IsDirectNeighbor(ServerData *serv, const struct in_addr *addr)
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable;
     // if node is already present as a DIRECT neighbor, do nothing
     while(list != NULL) {
	  if(!memcmp(&list->addr, addr, sizeof(struct in_addr))
	     && (list->type & DIRECT) ) 
	       return TRUE;
	
	  list = list->next;
     }
     return FALSE;
}

BOOL
IsSymmetricNeighbor(ServerData *serv, const struct in_addr *addr)
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable;
     // if node is already present as a SYMMETRIC neighbor, do nothing
     while(list != NULL) {
	  if(!memcmp(&list->addr, addr, sizeof(struct in_addr))
	     && (list->link == SYMMETRIC) ) 
	       return TRUE;
	
	  list = list->next;
     }
     return FALSE;
}

// A remote node is some node X that is not directly reachable. We need to inform the
// channel abstraction layer the channel to use while sending out packets for X. This channel
// to use is the channel used on the first hop of the route to X.
void
AddRemoteEntry(ServerData *serv, const struct in_addr *remoteNode, const struct in_addr *nextHop)
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable;
     RemoteNode *node;
     
     // If nextHop is present as a DIRECT neighbor, then add node to remoteList of nextHop
     list = helloData->neighborTable;
    
     while(list != NULL) {
	  if(!memcmp(&list->addr, nextHop, sizeof(struct in_addr))
	     && (list->type & DIRECT)) 
	  {
	       node = (RemoteNode *) malloc(sizeof(RemoteNode));
	       memcpy(&node->addr, remoteNode, sizeof(struct in_addr));
	       node->next = list->remoteList;
	       list->remoteList = node;

	       Ioctl_AddUnicastEntry(serv, remoteNode,  helloData->inter[list->interface].name, list->channel);

	       return;
	  }
	
	  list = list->next;
     }
}

void 
DeleteRemoteEntry(ServerData *serv, const struct in_addr *remoteNode, const struct in_addr *nextHop)
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable;
     RemoteNode *node, *prev=NULL;
     
     // If nextHop is present as a DIRECT neighbor, then delete node from remoteList of nextHop
     list = helloData->neighborTable;
    
     while(list != NULL) 
     {
	  if(!memcmp(&list->addr, nextHop, sizeof(struct in_addr))
	     && (list->type & DIRECT)) 
	  {
	       node = list->remoteList;
	    
	       // check in remoteList
	       while(node != NULL) 
	       {
		    // if remoteNode is present, delete it
		    if(!memcmp(&node->addr, remoteNode, sizeof(struct in_addr))) 
		    {
			 if(prev == NULL)
			      list->remoteList = node->next;
			 else
			      prev->next = node->next;
		    
			 Ioctl_DeleteUnicastEntry(serv, remoteNode,  helloData->inter[list->interface].name, list->channel);
		    
			 free(node);
			 return;
		    }
		    prev = node;
		    node = node->next;
	       }
	       return;
	  } // outer if
	  list = list->next;
     } // outer while
}

void
DeleteAsymmetricEntry(ServerData *serv, NeighborEntry *cur)
{
     HelloServerData *helloData = serv->helloData;
     RemoteNode *node, *prevNode;
          if(cur != NULL){
               node = cur->remoteList;
               cur->remoteList = NULL;
               while(node != NULL)
               {
                    Ioctl_DeleteUnicastEntry(serv, &node->addr,
                    			helloData->inter[cur->interface].name,
                                        cur->channel);
                    prevNode = node;
                    node = node->next;
                    free(prevNode);
               }
               NeighborNotReachable(serv, &cur->addr);
          }
          else
          err_msg("WARNING: Reached an invalid position in DeleteAsymmetricEntry\n"); // we should not come here
}

/* Checks to see if the node addr is present in erraticNeighbor List. If it is then it updates the
 * node's hellosToGo and then returns true. If the node is not on the list, the funct. returns false.
 */
BOOL checkAndUpdateUnreliableNeighbors(ServerData* serv, const struct in_addr *addr, int channel)
{
	
#ifdef DEBUG_MSG_ON
     err_msg("mcr_hello.c:debug_msg:checking and updating unreliable list for %s\n", print_ip(addr->s_addr));
#endif
     UnreliableNeighborEntry *erratic = (serv->helloData)->erraticNeighbors, *prev = NULL;
     while(erratic !=NULL)
     {
	  if(!memcmp(&erratic->addr, addr, sizeof(struct in_addr)))
	  {
	       erratic->hellosToGo--;
#ifdef DEBUG_MSG_ON
	       err_msg("mcr_hello.c:debug_msg:neighbor %s hellosTogo %d\n",print_ip(addr->s_addr),erratic->hellosToGo);
#endif
	       if(erratic->hellosToGo <=0)
	       {

#ifdef DEBUG_MSG_ON
		    err_msg("mcr_hello.c:debug_msg:neighbor %s unreliable, but now better, removing from erratic list\n",
			      print_ip(addr->s_addr));
#endif
		    // The link to the neighbor has improved
		    // Remove the neighbor from the Unreliable list
		    if(prev == NULL)
		    {
			 (serv->helloData)->erraticNeighbors = erratic->next;
		    }
		    else
		    {
			 prev->next = erratic->next;
		    }
		    free(erratic);		 
	       }
	       return TRUE;
	  }
	  prev = erratic;
	  erratic = erratic->next;
     }
     return FALSE;
}


/*
 * New AddToNeighborTable, changes by Rishi
 * Called when we receive a hello message from a neighbor. 
 * The neighbor table has entires for two hop (INDIRECT) and one hop (DIRECT) neighbors.
 * One hop neighbors may be SYMMETRIC (who see the node as DIRECT neighbor too) or  
 * ASYMMETRIC (who do not see the node as a DIRECT neighbor).
 */
void
AddToNeighborTable(ServerData *serv, const struct in_addr *addr, 
		   int channel, int numInterfaces, NeighborEntryType type, NeighborLinkType link, LINK_QUALITY_INT fwdLinkQuality) 
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable;

     while(list != NULL) 
     {
          if(!memcmp(&list->addr, addr, sizeof(struct in_addr))) 
	  {
	      
	       if(type == DIRECT)
	       {
		    if(list->type & DIRECT)
		    {
#ifdef DEBUG_MSG_ON
			 err_msg("mcr_hello.c:debug_msg:Hello:hello message from %s is direct\n",print_ip(list->addr.s_addr));
#endif
			 
			 list->directTtl = helloData->neighborTtl;		    
			 if(!serv->freeze_metrics){
			      list->revLinkQuality|=1;  
			      list->fwdLinkQuality = fwdLinkQuality;
			 }
			 
			 if(link == SYMMETRIC)
			 {
			      list->symmTtl = helloData->neighborTtl;
			      if(list->channel != channel || (list->type & DIRECT) == 0 || (list->link == ASYMMETRIC)) 
			      {
				   list->channel = channel;
				   list->numInterfaces = numInterfaces;
				   UpdateKernelTableForUnicast(serv, list);
			      }
			      
			 }
			 else
			 {
			      // link is asymmetric		         			 
			      list->asymmTtl = helloData->neighborTtl;
			      if(list->channel != channel || (list->type & DIRECT) == 0 || (list->link == SYMMETRIC) || 
					(list->inKernelTable == 1))
			      {
				   // update and remove entry from routing table if present as one hop neighbor.
				   list->channel = channel;
				   list->numInterfaces = numInterfaces;
				   DeleteAsymmetricEntry(serv, list);  
				   list->inKernelTable = 0;
			      }
			 }	    
			 
			 list->link = link;
		    }
		    else
		    {
			 // list->type is INDIRECT and type is DIRECT
			 // Check and see if list is present in erraticNeighborList
			 if(checkAndUpdateUnreliableNeighbors(serv,addr,channel))
			 {
			      // Check Successful, the neighbor is in erratic list
			      // The entry in erratic list has been updated, job done, go home.
			      return;
			 }
			 //Neighbor not on erratic list make the link direct.

			 // If freeze_metric is set. No new neighbors are added. For experiment purposes
			 if(!serv->freeze_metrics)
			 {
			      list->directTtl = helloData->neighborTtl;
			      list->revLinkQuality = ~ 0 ;
			      list->fwdLinkQuality = fwdLinkQuality ;
			      list->channel = channel;
			      list->numInterfaces = numInterfaces;
		    	      list->type |=type;
			      list->link = link;
			      if(link == SYMMETRIC)
			      {
				   list->symmTtl = helloData->neighborTtl;
				   UpdateKernelTableForUnicast(serv, list);     
				   list->inKernelTable = 1;
			      }
			 }  
		    }
	       } 
	       else 
	       {
		    //if received "link" was INDIRECT
		    list->indirectTtl = helloData->neighborTtl;
		    if( (list->type & DIRECT) == 0) 
		    {
			 // only an indirect entry
			 if(list->channel != channel )
			 {
			      list->channel = channel;
			      list->numInterfaces = numInterfaces;
			 }
		    }
	       }
	      list->type |= type;
	      return;
	  } // outer if
	  list = list->next;
     }

     if(type == DIRECT)
     {
	  // Check and see if list is present in erraticNeighborList
	  if(checkAndUpdateUnreliableNeighbors(serv,addr,channel))
	  {
	       // Check Successful, the neighbor is in erratic list
	       // The entry in erratic list has been updated, job done, go home.
	       return;
	  }
	  else if(serv->freeze_metrics)
	  {
	       // If Metrics are frozen, no new neighbors are added
	       // Used for experimentation.
	       return;
	  }
	  //Neighbor not on erratic list make the link direct.

     }
     
     list = (NeighborEntry *) malloc( sizeof(NeighborEntry));
     list->next = helloData->neighborTable;
     list->channel = channel;
     list->numInterfaces = numInterfaces;
     list->remoteList = NULL;
     list->interface = 0;
     memcpy(&list->addr, addr, sizeof(struct in_addr));
     list->type = type;
     list->link = link;
     list->revLinkQuality = 0;
     list->fwdLinkQuality = 0;
     if(type == DIRECT)
     {
	  list->directTtl = helloData->neighborTtl;
	  list->revLinkQuality = ~ 0;
	  list->indirectTtl = 0;

	  if(link == SYMMETRIC)
	  {
	       list->fwdLinkQuality = fwdLinkQuality;
	       UpdateKernelTableForUnicast(serv, list);
	       list->inKernelTable = 1;
	  }
     } 
     else 
     {
	  list->directTtl =  0;
	  list->indirectTtl = helloData->neighborTtl;
     }

     helloData->neighborTable = list;
}


/* This function decreases the cushionWindow of each erratic neighbor
 * If cushionWindow becomes 0, the hellosToGo and window both are reset.
 * Author: Rishi Bhardwaj
 */
void updateUnreliableNeighborTable(ServerData *serv)
{
     UnreliableNeighborEntry *erratic = (serv->helloData)->erraticNeighbors;
#ifdef DEBUG_MSG_ON
     err_msg("mcr_hello.c:debug_msg:Updating the Unreliable list\n");
#endif
     while(erratic!=NULL)
     {
	  erratic->helloCushionWindow--;
	  if(erratic->helloCushionWindow <=0)
	  {
	       // The erratic neighbor missed the window of oppurtunity
	       // Reset the window
#ifdef DEBUG_MSG_ON
	       err_msg("mcr_hello.c:debug_msg:The link to neighbor %s is still bad\n", print_ip((erratic->addr).s_addr));
#endif
	       erratic->hellosToGo = NEIGHBOR_MIN_TTL;
	       erratic->helloCushionWindow = NEIGHBOR_MAX_TTL;
	  }
	  erratic = erratic->next;
     }

}


void addNeighborToUnreliableList(ServerData *serv, struct in_addr *addr, int channel)
{

#ifdef DEBUG_MSG_ON
     err_msg("mcr_hello.c:debug_msg:Adding %s to Unreliable List\n", print_ip(addr->s_addr));
#endif
     
     UnreliableNeighborEntry *erratic = (serv->helloData)->erraticNeighbors;
     UnreliableNeighborEntry *new =(UnreliableNeighborEntry*)malloc(sizeof(UnreliableNeighborEntry));
     memcpy(&new->addr,addr,sizeof(struct in_addr));
     new->channel =channel;
     new->hellosToGo = NEIGHBOR_MIN_TTL;
     new->helloCushionWindow = NEIGHBOR_MAX_TTL;
     new->next = erratic;
     (serv->helloData)->erraticNeighbors = new;

} 

void DeleteOldEntries(ServerData *serv)
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *cur = helloData->neighborTable, *prev = NULL;
     RemoteNode *node, *prevNode;
     float successProb= 1, fwdProb, revProb;
     char buf[NEIGHBOR_FILE_BUF_LENGTH+50], addr[16];
     int buf_len = 0;

     updateUnreliableNeighborTable(serv);
     
#ifndef DEBUG_ROUTE_DELAY
     neighbor_file_updation_ctr = (neighbor_file_updation_ctr + 1) % NEIGHBOR_FILE_UPDATE_INTERVAL ;
     if(!neighbor_file_updation_ctr)
     {	  
	  fclose(neighborFile);
	  neighborFile=fopen("/root/Neighbors.txt","w");
	  if(neighborFile == NULL)
	  {
	       err_msg("Could not open file /root/Neighbors.txt");
	  }
	  buf_len += sprintf(buf, "# %-20s %-15s %-15s %-15s %-15s\n", "Addr", "TTL", "Rev. Link",
		    "Fwd. Link ", "Link Success Prob.");
     }	  
#endif
     
     while(cur != NULL) {

	  fwdProb = numberOfOnes(cur->fwdLinkQuality)/((float) sizeof(cur->fwdLinkQuality)*BITS_IN_BYTE);
	  revProb = numberOfOnes(cur->revLinkQuality)/((float) sizeof(cur->revLinkQuality)*BITS_IN_BYTE);	       
	  successProb = fwdProb * revProb ;
	  if(cur->directTtl > 0)
	  {
	       cur->directTtl--;	       
	       if(!serv->freeze_metrics)
		    cur->revLinkQuality<<=1; 
	       
#ifndef DEBUG_ROUTE_DELAY
	       err_msg("debug_route_delay not defined");
	       if(!neighbor_file_updation_ctr)
	       {
		    /*if(buf_len > NEIGHBOR_FILE_BUF_LENGTH)
		    {
			 write_to_file();
			 buf_len = 0;
		    }*/
		    sprintf(addr, "%d.%d.%d.%d",
			      0x0ff & cur->addr.s_addr,
			      0x0ff & (cur->addr.s_addr >> 8),
			      0x0ff & (cur->addr.s_addr >> 16), 0x0ff & (cur->addr.s_addr >> 24));
		    buf_len +=sprintf(buf + buf_len, "  %-20s %-15d %10f %15f %15f \n", addr, cur->directTtl, revProb, fwdProb, successProb);
		    buf[buf_len] = '\0';
		    if(neighborFile != NULL)
		    {
			 //fprintf(neighborFile, buf);
			 fputs(buf, neighborFile);
			 fflush(neighborFile);
		    }
		    buf_len = 0 ;		    
		       
	       }
	       
#endif
	       
#ifdef DEBUG_MSG_ON
	       err_msg("mcr_hello.c:debug_msg:Hello:Dir Entry:%s ttl = %d revLink %llx fwdLink %llx successProb %f \n",
			 print_ip(cur->addr.s_addr), cur->directTtl, cur->revLinkQuality, cur->fwdLinkQuality, successProb);
#endif
	  }
	  else if( (cur->type & DIRECT) == DIRECT) 
	  { 

#ifdef DEBUG_MSG_ON
	       err_msg("mcr_hello.c:debug_msg:Dir Entry:%s timed out server neighborTable, look for RemoteNodes",
			       print_ip(cur->addr.s_addr));
#endif
	       cur->type = cur->type & INDIRECT; // remove direct entry
	       cur->link = ASYMMETRIC;

	       // Watch for potential race condition here
	       node = cur->remoteList;
	       cur->remoteList = NULL;
	       while(node != NULL)
	       {

#ifdef DEBUG_MSG_ON
	   	       
		    err_msg("mcr_hello.c:debug_msg:node:%s is remoteNode\n",print_ip(node->addr.s_addr));
#endif
		    Ioctl_DeleteUnicastEntry(serv, &node->addr, helloData->inter[cur->interface].name, cur->channel);

#ifdef DEBUG_MSG_ON
		    err_msg("mcr_hello.c:debug_msg:DeletedUnicastEntry for node:%s\n", print_ip(node->addr.s_addr)); 
#endif
		    prevNode = node;
		    node = node->next;
		    free(prevNode);
	       }
	       NeighborNotReachable(serv, &cur->addr);
	       addNeighborToUnreliableList(serv,&cur->addr,cur->channel);
	  }
	
	  if(cur->indirectTtl > 0)
	       cur->indirectTtl--;
	  else if( (cur->type && INDIRECT) == INDIRECT)
	       cur->type = cur->type & DIRECT; // remove indirect entry

	  if(cur->type == 0) { // entry can be deleted
		  if(prev == NULL){ // first entry
		    prev = cur;
		    cur = cur->next;
		    helloData->neighborTable = cur;
		    free(prev);
		    prev = NULL;
		  } 
		  else {
		    prev->next = cur->next;
		    free(cur);
		    cur = prev->next;
		  }
	  } 
	  else {
	       prev = cur;
	       cur = cur->next;
	  }
     }
}


/******************** Functions to implement different fixed channel selection policies *********/

int
GetIndex(ServerData *serv, int channel)
{
     HelloServerData *helloData = serv->helloData;
     int i;
     for(i = 0; i < helloData->totalChannels; i++)
	  if(helloData->channelList[i] == channel)
	       return i;
     return -1;  // invalid value, which will cause a crash
}

// Balanced policy is what is used in our simulations and papers
BOOL 
BalancedPolicyCheckForChannelSwitch(ServerData *serv) 
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable;
     float channelCount[TOTAL_CHANNEL];
     int i, minCount, fixedChannel = GetIndex(serv, helloData->fixedChannel);;
     float min, max, avg;
  
     // Step 1: Count channel usage count of each channel

     for(i = 0; i < helloData->totalChannels; i++)
	  channelCount[i] = 0;

     while(list != NULL) {
	  if(list->numInterfaces > 1) // ignore single interface nodes while balancing channels
	       channelCount[GetIndex(serv,list->channel)]++;
	  list = list->next;
     }
  
     min = 100000; // some large number
     max = -1;
     avg = 0;
     for( i = 0 ; i < helloData->totalChannels; i++) {
	  min = min > channelCount[i] ? channelCount[i]: min;
	  max = max < channelCount[i] ? channelCount[i]: max;
	  avg += channelCount[i];

     }
     avg = avg / helloData->totalChannels;
    
     // Step 2: If our current fixed channel is not the best, then we consider switching to min channel
     if(channelCount[fixedChannel] >= (avg + 1) && (min + 1) < channelCount[fixedChannel]) {
	  double switchProb = helloData->switchProb * (2.0/ channelCount[fixedChannel]) ; // dampen number of nodes that can switch
	  double perChannelProb, randomNum;
	  int  temp;
	  minCount = 0;
	  for( i = 0 ; i < helloData->totalChannels; i++) {
	       if(channelCount[i] == min)
		    minCount ++;
	  }
    
	  // There are minCount possible channels to switch to. We want to switch at all with prob SWITCH_PROB. 
	  // So with SWITCH_PROB/ minCount, we want to switch to any particular channel
	  perChannelProb = switchProb / minCount;

	  randomNum = (double)random()/RAND_MAX;
	  if(randomNum >= switchProb)
	       return FALSE;
	  
	  temp = (int)(randomNum / perChannelProb);
    
	  for( i = 0 ; i < helloData->totalChannels; i++) {
	       if(channelCount[i] == min && temp == 0) // this is the chosen channel
	       {
		    helloData->fixedChannel = helloData->channelList[i];
		    return TRUE;
	       } else if(channelCount[i] == min)
		    temp --;
	  } // for
	  err_msg("WARNING: Reached an invalid position\n"); // we should not come here
     }
     return FALSE;
}

BOOL 
RandomPolicyCheckForChannelSwitch(ServerData *serv)
{
     HelloServerData *helloData = serv->helloData;
     double switchProb = helloData->switchProb;
     double randomNum;
     int newChannel;

     randomNum = (double) random()/RAND_MAX;
     if(randomNum < switchProb) {
	  newChannel = (int)( ((double)random() / RAND_MAX) 
			      * helloData->totalChannels);
	  if(helloData->channelList[newChannel] == helloData->fixedChannel)
	       return FALSE;
	  helloData->fixedChannel = helloData->channelList[newChannel];
	  return TRUE;
     }
     return FALSE;
}

void
ChangeFixedChannelIfRequired(ServerData *serv)
{
     HelloServerData *helloData = serv->helloData;
     BOOL switchRequired = FALSE;
     int oldFixedChannel = helloData->fixedChannel;

#ifdef SINGLE_CHANNEL_EXPERIMENT
     return ;     
#endif
     
     if(serv->freeze_metrics)
     {
	  return ;
     }
     
     switch(helloData->switchPolicy) {
     case RANDOM:
     {
	  switchRequired = RandomPolicyCheckForChannelSwitch(serv);
	  break;
     }
    
     case BALANCED:
     case BALANCED_TWOHOP:
     {
	  switchRequired = BalancedPolicyCheckForChannelSwitch(serv);
	  break;
     }
     default:
	  break;
     }

     if(!switchRequired)
	  return;
  
     SwitchFixedChannel(serv, oldFixedChannel, helloData->fixedChannel);
     return;
}


/********************* Create and send out "hello" messages ********************************/

void
SendHelloPacket(ServerData *serv, int sendfd, HelloType type, const struct in_addr *addr) 
{
     HelloServerData *helloData = serv->helloData;
     char packet[MAX_MESSAGE_LENGTH];
     McrPacketHeader mcrHdr;
     HelloPacketHeader helloHdr;
     HelloPacketData data;
     socklen_t len = sizeof(struct sockaddr_in);
     NeighborEntry *list = helloData->neighborTable;
     int numEntries = 0, bytePointer=0;
     helloHdr.channel = helloData->fixedChannel;
     helloHdr.numInterfaces = helloData->totalInterfaces;
     helloHdr.seq = serv->seq++;
     helloHdr.type = type;
     memcpy(&mcrHdr.fromAddr, &serv->myIpAddr, sizeof(struct in_addr));
     bytePointer = sizeof(mcrHdr) + sizeof(helloHdr);
     
     // "type" is used to figure out if we are responding to a received hello message
     // this is true when a multi-interface node is responding to a scan message of a single interface node
     //Nistha - changing the comment - NOACK is fine for normal BALANCED TWO-HOP packets
     while(list != NULL && type == NOACK) {
	  if(list->type & DIRECT) {
	       memcpy(&data.addr, &list->addr, sizeof(struct in_addr));
	       data.channel = list->channel;
	       data.numInterfaces = list->numInterfaces;
	       data.revLinkQuality = list->revLinkQuality;
	       memcpy(packet + bytePointer, &data, sizeof(HelloPacketData));
	       bytePointer+=sizeof(HelloPacketData);
	       numEntries++;
	       if(bytePointer + sizeof(HelloPacketData) > MAX_MESSAGE_LENGTH)
	       {
		    err_msg("mcr_hello.c:SendHelloPacket():debug_msg:MAX_MESSAGE_LENGTH=%d is smaller than the length required for hello packet --- ?\n", 
			      MAX_MESSAGE_LENGTH);
		   break; 
	       }
	  }
	  list = list->next;
     }

     helloHdr.numEntries = numEntries;
          
     
     // Length of Hello packet after McrPacketHeader
     mcrHdr.len = helloData->min_hello_packet_size > bytePointer ? 
	  (helloData->min_hello_packet_size - sizeof(mcrHdr)) : bytePointer - sizeof(mcrHdr);
    
     mcrHdr.type = LINK_HELLO_PKT;
     copyStructIntoPkt(mcrHdr, packet);
     copyStructIntoPkt(helloHdr,packet+sizeof(mcrHdr));


// We can send a large size hello packet depending on the specified command line argument

// Hello messages used to announce node's fixed channel are broadcast.  Hello messages that
// are acknowledging a scan request are unicast back to the desired node
     if(type != HELLOACK)
	  Sendto(sendfd, packet, sizeof(mcrHdr)+mcrHdr.len, 0, (SA *) &serv->sendBroadcastSocket, len);
     else {
	  struct sockaddr_in socket; 	  // socket for unicasting hello ACK
	  bzero( &socket, sizeof(struct sockaddr_in));
	  socket.sin_family = AF_INET;
	  socket.sin_port = htons(HELLO_PORT);
	  memcpy(&socket.sin_addr, addr, sizeof(struct in_addr));

	  Sendto(sendfd, packet, sizeof(mcrHdr)+mcrHdr.len, 0, (SA *) &socket, len);
     }
}


/*********** Channel scan facility for supporting single interface nodes ***************/

int
GetBestScanChannelIndex(ServerData *serv, int channel)
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable, *choice = NULL;
     int maxTtl = -1;

     // See if we know of a DIRECT neighbor, not on channel "channel"
     // if so, pick neighbor with largest ttl
     // XXX - we could also prioritize on picking a neighbor with at least
     // two interfaces

     while(list != NULL)
     {
	  if( (list->type & DIRECT) && list->channel != channel && list->directTtl > maxTtl)
	  {
	       choice = list;
	       maxTtl = list->directTtl;
	  }
	  list = list->next;
     }

     if(choice == NULL)
	  return 0;
     else {
	  // XXX: We could use GetIndex function here
	  int i;
	  for(i = 0; i < helloData->totalChannels; i++)
	       if(helloData->channelList[i] == choice->channel)
		    return i;
     }

     err_sys("MCR_HELLO: GetBestScanChannelIndex() - reached an unexpected return point , choice->channel = %d\n", choice->channel);
     return 0;

}

void
SendNextScanPacket(ServerData *serv)
{
     HelloServerData *helloData = serv->helloData;
     struct timeval delay;
     int curChannel;

// a single interface node is not scanning if NOSCAN flag is set  (this should be the common case)
     if(helloData->scanStatus == NOSCAN)
	  return;
    
     curChannel = helloData->fixedChannel;
     helloData->fixedChannel = helloData->channelList[helloData->scanAttemptIndex];

     if(curChannel != helloData->fixedChannel)
	  Ioctl_CardSetChannel(serv, helloData->inter[helloData->fixedInterface].name, helloData->fixedChannel);

     // Update for next attempt
     if(++helloData->scanAttemptNumber >=  MAX_SCAN_ATTEMPT_PER_CHANNEL) 
     {
	  ++helloData->scanAttemptIndex;
	  helloData->scanAttemptNumber = 0;
	  if( helloData->scanAttemptIndex == helloData->totalChannels)
	       helloData->scanAttemptIndex = 0;
     }
 
     // sendout scan message and set timeout
     SendHelloPacket(serv, serv->sendfd, NEEDACK, NULL);

     delay.tv_sec = SCAN_ATTEMPT_TIMEOUT / 1000;
     delay.tv_usec = (SCAN_ATTEMPT_TIMEOUT % 1000)*1000; 
     AddTimeout(serv, &delay, SCAN_TIMEOUT, 0, 0);
}

BOOL
StillValidChannel(ServerData *serv, int channel)
{
     HelloServerData *helloData = serv->helloData;
     NeighborEntry *list = helloData->neighborTable;

     while(list != NULL)
     {
	  if( (list->type & DIRECT) && list->channel == channel && list->directTtl > 0)
	       return TRUE;

	  list = list->next;
     }
     return FALSE;
}

void
StartScanChannelIfRequired(ServerData *serv, int channel)
{
     HelloServerData *helloData = serv->helloData;
    
     if(helloData->scanStatus == INSCAN)
	  return;
    
     if(StillValidChannel(serv, channel))
	  return;
	
     helloData->scanStatus = INSCAN;
     helloData->scanAttemptNumber = 0;
    
     helloData->scanAttemptIndex = GetBestScanChannelIndex(serv, channel);
     SendNextScanPacket(serv);
}

/********************** Handling expiration of timers ***********************/

void
ScanTimeout(ServerData *serv)
{
     if(serv->helloData->scanStatus == INSCAN)
	  SendNextScanPacket(serv);
}

void
SetHelloTimeout(ServerData *serv)
{
     HelloServerData *helloData = serv->helloData;
     struct timeval delay;
    
     delay.tv_sec = helloData->helloInterval;
     delay.tv_usec = (int)(((double) random()/RAND_MAX)*HELLO_JITTER);   // randomize when hellos are sent out
     AddTimeout(serv, &delay, HELLO_TIMEOUT, 0, 0);
}


void
HelloTimeout(ServerData *serv) 
{
     HelloServerData *helloData = serv->helloData;
     int sendfd = serv->sendfd;


     DeleteOldEntries(serv);

     if(helloData->totalInterfaces > 1)
	  ChangeFixedChannelIfRequired(serv);
     else
	  StartScanChannelIfRequired(serv, helloData->fixedChannel);

     AddToNeighborTable(serv, &serv->myIpAddr, helloData->fixedChannel, helloData->totalInterfaces, INDIRECT, SYMMETRIC,0); 
     SendHelloPacket(serv, sendfd, NOACK, NULL);
     SetHelloTimeout(serv);
}

/*********************** Receive and process hello messages *****************/


void
ProcessHello(ServerData *serv, const McrPacketHeader *hdr, const char *dataPkt, int dataSize)
{
     HelloServerData *helloData = serv->helloData;
     int numEntries ; 
     int i;
     HelloPacketHeader helloHdr;
     HelloPacketData data;

     int symmflag=0; 
     LINK_QUALITY_INT fwdLinkQuality = 0 ;
     copyStructFromPkt(helloHdr, dataPkt);

#ifdef DEBUG_MSG_ON
     err_msg("mcr_hello.c:debug_msg:Hello:Received a hello message from %s\n",print_ip(hdr->fromAddr.s_addr));
#endif

     numEntries = helloHdr.numEntries;
     for( i = 0; i < numEntries; i++) {
	  memcpy(&data, dataPkt + sizeof(HelloPacketHeader) + i*sizeof(HelloPacketData), 
		 sizeof(HelloPacketData));
	  
	  // Making the entries symmetric by default, to avoid overwriting of Symmetric status on 
	  // receiving hello from mutual neighbor
	  AddToNeighborTable(serv, &data.addr, data.channel, data.numInterfaces, INDIRECT, SYMMETRIC, data.revLinkQuality);
	  
	  //check if my ip; if yes set a boolean flag
 	  if (htonl(serv->myIpAddr.s_addr) == htonl(data.addr.s_addr)){ 
		  symmflag = 1;
		  fwdLinkQuality = data.revLinkQuality;

	  }
	  
     }
	  	    
     // A node is a DIRECT neighbor only if we can directly reach it.
     // For a single interface node, this requires the fixed channel of the
     // other node to be same as ourselves
     if(helloData->totalInterfaces > 1 || helloHdr.channel == helloData->fixedChannel)
     {
	if(symmflag)     	
		AddToNeighborTable(serv, &hdr->fromAddr, helloHdr.channel, helloHdr.numInterfaces, DIRECT, SYMMETRIC, fwdLinkQuality);
	else
		AddToNeighborTable(serv, &hdr->fromAddr, helloHdr.channel, helloHdr.numInterfaces, DIRECT, ASYMMETRIC, fwdLinkQuality);
     }
     else
	  AddToNeighborTable(serv, &hdr->fromAddr, helloHdr.channel, helloHdr.numInterfaces, INDIRECT, SYMMETRIC, fwdLinkQuality);
     symmflag = 0;

     // Question: Should we allow a single interface node to ACK - this might
     // cause two single interface nodes to partition from rest of
     // the network by staying on a common channel. 
     // For now, we prevent a single interface node from acking by checking if totalInterfaces is greater than 1

     if(helloHdr.type == NEEDACK && helloHdr.channel == helloData->fixedChannel && helloData->totalInterfaces > 1)
	  SendHelloPacket(serv, serv->sendfd, HELLOACK, &hdr->fromAddr);

     // No need to scan further, there is a neighbor present
     if(helloHdr.type == HELLOACK) 
	  helloData->scanStatus = NOSCAN;

     // An optimization to change channels quickly if original neighbor has
     // moved on to a different channel
     if(helloData->totalInterfaces == 1 && helloHdr.type == NOACK)
	  StartScanChannelIfRequired(serv, helloData->fixedChannel);
}



/**************************** Initialization ********************************8***/

void
HelloInit(ServerData *serv)
{
     HelloServerData *helloData;
     
     helloData = (HelloServerData *) malloc(sizeof(HelloServerData));
     if(helloData == NULL)
	  err_quit("Could not allocate memory for the helloData\n");

     helloData->neighborTable = NULL;
     helloData->helloInterval = HELLO_INTERVAL;
     helloData->switchProb = SWITCH_PROB;
     helloData->neighborTtl = NEIGHBOR_MAX_TTL;
     helloData->fixedInterface = 0;
     helloData->switchPolicy = BALANCED_TWOHOP;
     helloData->scanStatus = NOSCAN;
     helloData->erraticNeighbors = NULL;
     helloData->min_hello_packet_size = DEFAULT_MIN_HELLO_PACKET_SIZE;
     serv->helloData = helloData;
}
