/*****************************************************************************
 * mcr_hello.h: Interface management protocol
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 * Modifications by: Rishi Bhardwaj, <rbhardw2 at uiuc dot edu>
 * 		     Nistha Tripathi, <ntripa2 at uiuc dot edu>
 * 
 *****************************************************************************/

#ifndef _MCR_HELLO_H
#define _MCR_HELLO_H

/*********** Default constants *******************/

#define TOTAL_CHANNEL 15                // maximum number of supported chanenls
#define NEIGHBOR_MAX_TTL  15         	// maximum number of missed hellos before a neighbor is deemed to be unreachable

#define NEIGHBOR_MIN_TTL  11		// minimum no. of hellos a node must receive from an "unreachable" neighbor in 
					// NEIGHBOR_MAX_TTL * HELLO_INTERVAL time before the "unreachable" neighbor
					// is deemed fit to be considered a DIRECT one hop neighbor again.
					// Note  NEIGHBOR_MIN_TTL <= NEIGHBOR_MAX_TTL
 
#define DEFAULT_MIN_HELLO_PACKET_SIZE 100  // Defualt min. size of hello query packets, overridden by -p command line option


#define SWITCH_PROB  0.4                 // constant used in balanced policy channel assignment
#define HELLO_INTERVAL 3                // period of time, in seconds, for sending out hello messages, overridden by command line argument -h
#define HELLO_JITTER  100000           // in microseconds (i.e. 100 ms)

// Single interface canning parameters
#define MAX_SCAN_ATTEMPT_PER_CHANNEL 1
#define SCAN_ATTEMPT_TIMEOUT    200  // in milli seconds

#define INVALID_CHANNEL -1

#define MIN_CONS_HELLO 1

#define BITS_IN_BYTE 8

#define NEIGHBOR_FILE_UPDATE_INTERVAL	3
#define NEIGHBOR_FILE_BUF_LENGTH	200

#define LINK_QUALITY_INT unsigned long long      // Using 64 bit long long for hello packet, link quality bitmap

/************** Protocol specific lags ****************/

typedef enum {
     NOSCAN = 1,
     INSCAN
} ScanStatusType;

typedef enum {
     NOACK = 1,
     NEEDACK,
     HELLOACK
} HelloType;

// Different Hello policies that are supported
typedef enum {
     RANDOM = 1,
     HASH,
     BALANCED,
     BALANCED_TWOHOP,
     FIXED,
     TEST
} SwitchPolicy;

// Type of neighbor
typedef enum {
     DIRECT = 0x1, // 1-hop neighbor
     INDIRECT = 0x2
} NeighborEntryType;

typedef enum {
     SYMMETRIC = 0x1, // symmetric neighbor
     ASYMMETRIC = 0x2
} NeighborLinkType;

/************* Hello packet structures *****************/

typedef struct struct_hello_packet_header {
     int channel;
     unsigned int seq;
     HelloType type;
     int numInterfaces;
     int numEntries;       // Number of hello_packet_data links
} HelloPacketHeader;

typedef struct struct_hello_packet_data {
     struct in_addr addr;
     int channel;
     int numInterfaces;
     LINK_QUALITY_INT revLinkQuality;
} HelloPacketData;

/************ Neighbor table structures ***************/

typedef struct struct_remote_node {
     struct in_addr addr;
     struct struct_remote_node *next;    
} RemoteNode;

// Neighbor table entry
typedef struct struct_neighbor_entry {
     struct in_addr addr;
     int interface; // name of the interface to use for sending data
    
     int channel;
     int numInterfaces; // number of interfaces
     int directTtl;
     int indirectTtl;
     NeighborEntryType type;
     NeighborLinkType link; 
     int symmTtl;
     int asymmTtl;
     LINK_QUALITY_INT revLinkQuality;   // bit map of last 64(long long) hello messages received from neighbor. 
     LINK_QUALITY_INT fwdLinkQuality;   // bit map of last 64 hello messages sent to neighbor.	     
     BOOL inKernelTable ;    		// indicates if the one hop entry exists in the kernel routing table
     
     RemoteNode *remoteList;
     struct struct_neighbor_entry *next;
} NeighborEntry;


// Interface information
typedef struct struct_interface_data {
     char *name;
     BOOL isFixed;
} InterfaceData;


// Struct to store information about unreliable neighbors.
// If the node does not receive any hello messages from a neighbor
// for MAX_NEIGHBOR_TTL hello message intervals the one hop link from
// neighbor to the node is deemed unusable. The neighbor is removed from
// the neighbor table and added in the unreliable neighbor table (erraticNeighbors).
// The node must receive hellosToGo hello messages inside helloCushionWindow
// from the unrelliable neighbor to once again deem the one hop link to 
// the neighbor usable.
typedef struct struct_unreliable_neighbor_entry{
	struct in_addr addr;			// Neighbor's IP
	unsigned int channel;			// Fixed channel of the neighbor
	unsigned short hellosToGo;			
	unsigned short helloCushionWindow;
	struct struct_unreliable_neighbor_entry *next;	// Next Unreliable Neighbor
} UnreliableNeighborEntry;	


/************** Main hello protocol data structure *****************/

typedef struct struct_hello_server_data {

     int fixedChannel;
     int fixedInterface;

     int totalChannels;
     int totalInterfaces;

     InterfaceData inter[TOTAL_CHANNEL];
     int channelList[TOTAL_CHANNEL];

     NeighborEntry *neighborTable;

     UnreliableNeighborEntry *erraticNeighbors;
     
     SwitchPolicy switchPolicy;

     // Configuration parameters
     int helloInterval; // in seconds
     int min_hello_packet_size;  // minimum hello packet size in bytes 
     int neighborTtl; // num of neighbor message losses that are tolerated
     double switchProb; // probability of switching
    
     ScanStatusType scanStatus;
     int scanAttemptIndex; // channel we are currently scanning - we scan in a round robin fashion
     int scanAttemptNumber;
} HelloServerData;


/*************** Function declarations for external usage ******************/

BOOL IsDirectNeighbor(ServerData *serv, const struct in_addr *addr);
BOOL IsSymmetricNeighbor(ServerData *serv, const struct in_addr *addr);

BOOL IsOneHopNeighbor(ServerData* serv, const struct in_addr *addr);

float GetNeighborDeliveryRatio(ServerData *serv, const struct in_addr *addr);

void AddRemoteEntry(ServerData *serv, const struct in_addr *node, const struct in_addr *nextHop);

void DeleteRemoteEntry(ServerData *serv, const struct in_addr *node, const struct in_addr *nextHop);

void HelloInit(ServerData *serv);

void ProcessHello(ServerData *serv, const McrPacketHeader *hdr, const char *dataPkt, int dataSize);

void ScanTimeout(ServerData *serv);

void HelloTimeout(ServerData *serv);

void SetHelloTimeout(ServerData *serv);

void SendNextScanPacket(ServerData *serv);

void AddToNeighborTable(ServerData *serv, const struct in_addr *addr, int channel, int numInterfaces, NeighborEntryType type, NeighborLinkType link, LINK_QUALITY_INT fwdLinkQuality);

void addNeighborToUnreliableList(ServerData *serv, struct in_addr *addr, int channel);

void updateUnreliableNeighborTable(ServerData *serv);

BOOL checkAndUpdateUnreliableNeighbors(ServerData* serv, const struct in_addr *addr, int channel);

void StartScanChannelIfRequired(ServerData *serv, int channel);

void InitHelloKernelTable(ServerData *serv);

#endif /* _MCR_HELLO_H */
