/*****************************************************************************
 * mcr_ioctl.c: Support for communication with the kernel
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 *
 *****************************************************************************/


/*************************** NOTES *******************************************

This file provides an interface for the different protocol components
to communicate with the kernel multichannel routing module and 
channel abstraction layer.

*********************************************************************************/

#include "utilities.h"
#include "integrated.h"
#include "mcr_ioctl.h"
#include "mcr_route.h"
#include "mcr_hello.h"
#include "kmcr_netlink.h"

extern DEBUG_MODE_TYPE debug_mode;

/*************** Netlink socket code to send message to kernel multichannel routing module *** ********/

int Netlink_create_and_send_msg(ServerData *serv, int type, const struct in_addr *dest,
				const struct in_addr *nhop, u_int32_t lifetime,
				int flags)
{
     int status;
     struct {
	  struct nlmsghdr nlh;
	  kmcr_rt_msg_t rtm;
     } req;

     memset(&req, 0, sizeof(req));
     req.nlh.nlmsg_len = NLMSG_LENGTH(sizeof(kmcr_rt_msg_t));
     req.nlh.nlmsg_flags = NLM_F_REQUEST;
     req.nlh.nlmsg_type = type;
     req.nlh.nlmsg_pid = serv->netlink_local.nl_pid;;
     req.rtm.dest = dest->s_addr;
     req.rtm.nhop = nhop->s_addr;
     req.rtm.flags = flags;
     req.rtm.route_expire = lifetime;

     status = sendto(serv->netlinkfd, &req, req.nlh.nlmsg_len, 0,
		     (struct sockaddr *) &serv->netlink_peer, sizeof(struct sockaddr_nl));

     return status;
}


/********  Interaction with (standard) linux kernel routing table + kernel multichannel routing module *********/
void
Ioctl_AddRoute(ServerData *serv, struct in_addr *dst, struct in_addr *nextHop)
{
     struct sockaddr_in destination, mask, gateway;
     int res;
     struct rtentry entry;

     char localAddr[INET_ADDRSTRLEN];
     char destAddr[INET_ADDRSTRLEN];
     char nextHopAddr[INET_ADDRSTRLEN];
 
#ifdef DEBUG_IOCTL   
     err_msg("IOCTL: AddRoute (%s): Destination = %s Nexthop = %s", 
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, dst, destAddr, sizeof(destAddr)),
	     Inet_ntop(AF_INET, nextHop, nextHopAddr, sizeof(nextHopAddr)));
#endif

     if(debug_mode)
	  return;

     memset(&entry, 0, sizeof(struct rtentry));

     destination.sin_family = AF_INET;
     memcpy( &destination.sin_addr.s_addr, dst, sizeof(struct in_addr));
     gateway.sin_family = AF_INET;

     // If nexthop is same as destination, we can set gateway to 0
     if ( !memcmp(dst, nextHop, sizeof(struct in_addr)) ) 
     {
	  memset( &gateway.sin_addr.s_addr, 0, sizeof(struct in_addr));
	  entry.rt_flags = RTF_HOST | RTF_UP | RTF_DYNAMIC;
     }
     else
     {
	  memcpy( &gateway.sin_addr.s_addr, nextHop, sizeof(struct in_addr));
	  if(dst->s_addr == 0) // default gateway
	       entry.rt_flags = RTF_GATEWAY | RTF_UP | RTF_DYNAMIC;
	  else
	       entry.rt_flags = RTF_HOST | RTF_GATEWAY | RTF_UP | RTF_DYNAMIC;
     }

     // Net mask can be 0
     mask.sin_family = AF_INET;
     memset( &mask.sin_addr.s_addr, 0, sizeof(struct in_addr));

     memcpy(&entry.rt_dst, &destination, sizeof(struct sockaddr_in));
     memcpy(&entry.rt_gateway, &gateway, sizeof(struct sockaddr_in));
     memcpy(&entry.rt_genmask, &mask, sizeof(struct sockaddr_in));

//    if(dst->s_addr != 0)
//	entry.rt_dev = serv->ifr.ifr_name; // name of exposed bonding interface
//    else
//	entry.rt_dev = "eth0"; // default route for gateway
//    entry.rt_dev = NULL;
     entry.rt_dev = serv->ifr.ifr_name; // name of exposed bonding interface

     entry.rt_metric = 1; // some dummy number

     res = ioctl(serv->ioctlSock, SIOCADDRT, &entry);

     if(res != 0)
	  err_msg("IOCTL: AddRoute (%s): Ioctl Failed for Destination = %s Nexthop = %s, error = %s", 
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  Inet_ntop(AF_INET, dst, destAddr, sizeof(destAddr)),
		  Inet_ntop(AF_INET, nextHop, nextHopAddr, sizeof(nextHopAddr)), strerror(errno));

     if(dst->s_addr != 0) // ignore default gateway
	  Netlink_create_and_send_msg(serv, KMCR_ADDROUTE, dst, nextHop, KERNEL_ROUTE_TIMEOUT, 0);
    
}

void
Ioctl_DeleteRoute(ServerData *serv, struct in_addr *dst, struct in_addr *nextHop) 
{
     struct sockaddr_in destination, mask, gateway;
     int res;
     struct rtentry entry;

     char localAddr[INET_ADDRSTRLEN];
     char destAddr[INET_ADDRSTRLEN];
     char nextHopAddr[INET_ADDRSTRLEN];

#ifdef DEBUG_IOCTL       
     err_msg("IOCTL: DeleteRoute (%s): Destination = %s Nexthop = %s", 
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, dst, destAddr, sizeof(destAddr)),
	     Inet_ntop(AF_INET, nextHop, nextHopAddr, sizeof(nextHopAddr)));
#endif

     if(debug_mode)
	  return;

     memset(&entry, 0, sizeof(struct rtentry));

     destination.sin_family = AF_INET;
     memcpy( &destination.sin_addr.s_addr, dst, sizeof(struct in_addr));
     /* We don't really use nextHop as we delete any route
      * to destination at the moment, but this can be changed 
      * if needed
      */
     gateway.sin_family = AF_INET;
     memset( &gateway.sin_addr.s_addr, 0, sizeof(struct in_addr));
     if(dst->s_addr != 0) // ignore default gateway
	  entry.rt_flags = RTF_HOST;
     else
	  entry.rt_flags = RTF_GATEWAY;

     // Net mask can be 0
     mask.sin_family = AF_INET;
     memset( &mask.sin_addr.s_addr, 0, sizeof(struct in_addr));

     memcpy(&entry.rt_dst, &destination, sizeof(struct sockaddr_in));
     memcpy(&entry.rt_gateway, &gateway, sizeof(struct sockaddr_in));
     memcpy(&entry.rt_genmask, &mask, sizeof(struct sockaddr_in));

// We don't care about the interface either
//    entry.rt_dev = serv->ifr.ifr_name; // name of exposed bonding interface
 
     res = ioctl(serv->ioctlSock, SIOCDELRT, &entry);

     if(res != 0)
	  err_msg("IOCTL: DeleteRoute (%s): Ioctl Failed for Destination = %s Nexthop = %s", 
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  Inet_ntop(AF_INET, dst, destAddr, sizeof(destAddr)),
		  Inet_ntop(AF_INET, nextHop, nextHopAddr, sizeof(nextHopAddr)));

     if(dst->s_addr != 0) // ignore default gateway
	  Netlink_create_and_send_msg(serv, KMCR_DELROUTE, dst, nextHop, 0, 0);
}

/************** Interaction with wireless card through "wireless extensions" interface ************/

void
Ioctl_CardSetChannel(ServerData *serv, const char *interface, int channel)
{
     char localAddr[INET_ADDRSTRLEN];
     struct iwreq card_req;
     int res;
    
#ifdef DEBUG_IOCTL   
     err_msg("IOCTL: CardSetChannel (%s): Interface = %s, channel = %d", 
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     interface, channel);    
#endif
    
     if(debug_mode) // In debug mode, we do nothing
	  return;
    
     strcpy(card_req.ifr_ifrn.ifrn_name, interface);
     card_req.u.freq.m = channel;
     card_req.u.freq.e = 0;
     card_req.u.freq.i = 0;

     res = ioctl(serv->ioctlSock, SIOCSIWFREQ, &card_req);
     if(res != 0)
	  err_msg("IOCTL: CardSetChannel (%s): Ioctl request failed for %s, %d", 
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  interface, channel);
    


}


/******************* Interaction with the channel abstraction module  ************************/

/*TODO:Function should return a boolean specifying if the IOCTL call was successful or not */
void
Ioctl_GetSwitchingStats(ServerData *serv, const char *interface, struct muic_stats_u_response *resp)
{
     struct muic_stats_u_req muic_req;

     char localAddr[INET_ADDRSTRLEN];
     int res;
    
#ifdef DEBUG_IOCTL   
     err_msg("IOCTL: GetSwitchingStats (%s): Bond = %s Interface = %s", 
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)), serv->ifr.ifr_name,
	     interface);
#endif

     if(debug_mode)
	  return;

     strcpy(muic_req.slave, interface);
     muic_req.response = resp;
     serv->ifr.ifr_data = (char *) &muic_req;
     res = ioctl(serv->ioctlSock, BOND_MUIC_STATS, &serv->ifr);
     if(res != 0)
	  err_msg("IOCTL: GetSwitchingStats (%s): Ioctl request failed for %s", 
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  interface);

}

void
Ioctl_AddValidChannel(ServerData *serv, const char *interface, int channel)
{
     struct muic_channel_req muic_req;
     char localAddr[INET_ADDRSTRLEN];
     int res;

#ifdef DEBUG_IOCTL   
     err_msg("IOCTL: AddValidChannel (%s): Bond = %s Interface = %s, channel = %d", 
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)), serv->ifr.ifr_name,
	     interface, channel);
#endif

     if(debug_mode || serv->helloData->totalInterfaces == 1)
	  return;
  
     strcpy(muic_req.slave, interface);
     muic_req.channel = channel;
     muic_req.add = 1;
     serv->ifr.ifr_data = (char *) &muic_req;
     res = ioctl(serv->ioctlSock, BOND_MUIC_CHANNEL, &serv->ifr);
     if(res != 0)
	  err_msg("IOCTL: AddValidChannel (%s): Ioctl request failed for %s, %d", 
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  interface, channel);
}

void
Ioctl_SwitchChannel(ServerData *serv, const char *interface, int channel)
{
     char localAddr[INET_ADDRSTRLEN];
     struct muic_channel_switch_req muic_req;
     int res;

#ifdef DEBUG_IOCTL    
     err_msg("IOCTL: SwitchChannel (%s): Interface = %s, channel = %d", 
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     interface, channel);    
#endif

     if(debug_mode || serv->helloData->totalInterfaces == 1)
	  return;
  
     strcpy(muic_req.slave, interface);
     muic_req.channel = channel;
     serv->ifr.ifr_data = (char *) &muic_req;
     res = ioctl(serv->ioctlSock, BOND_MUIC_SWITCH_CHANNEL, &serv->ifr);
     if(res != 0)
	  err_msg("IOCTL: SwitchChannel (%s): Ioctl request failed for %s, %d", 
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  interface, channel);
}
void Ioctl_AddUnicastEntry(ServerData *serv, const struct in_addr *addr, const char *interface, int channel)
{
     char localAddr[INET_ADDRSTRLEN];
     struct muic_unicast_req muic_req;
     int res;

#ifdef DEBUG_IOCTL   
     char addrString[INET_ADDRSTRLEN];
     err_msg("IOCTL: AddUnicastEntry (%s): Ip = %s Interface = %s, channel = %d", 
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, addr, addrString, sizeof(addrString)),
	     interface, channel); 
#endif
          
     if(debug_mode){
	  err_msg("nt-debug: we are returning from Ioctl_AddUnicastEntry due to debug mode\n");
	  return;
     }

     Netlink_create_and_send_msg(serv, KMCR_ADDROUTE, addr, addr, KERNEL_ROUTE_TIMEOUT, 0);
    
     if(serv->helloData->totalInterfaces == 1)
	  return;
    
     strcpy(muic_req.slave, interface);
     muic_req.channel = channel;
     muic_req.addr = addr->s_addr;
     serv->ifr.ifr_data = (char *) &muic_req;
     res = ioctl(serv->ioctlSock, BOND_MUIC_ROUTE, &serv->ifr);
     if(res != 0)
	  err_msg("IOCTL: AddUnicastEntry (%s): Ioctl request failed for %s, %d", 
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  interface, channel); 
}

void
Ioctl_DeleteUnicastEntry(ServerData *serv, const struct in_addr *addr, const char *interface, int channel)
{
     char localAddr[INET_ADDRSTRLEN];
     struct muic_unicast_req muic_req;
     int res;

#ifdef DEBUG_IOCTL   
     char addrString[INET_ADDRSTRLEN];
     err_msg("IOCTL: DeleteUnicastEntry (%s): Ip = %s Interface = %s, channel = %d", 
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, addr, addrString, sizeof(addrString)),
	     interface, channel);
#endif

     if(debug_mode)
	  return;

     Netlink_create_and_send_msg(serv, KMCR_DELROUTE, addr, addr, 0, 0);
   
     if(serv->helloData->totalInterfaces == 1)
	  return;
  
     strcpy(muic_req.slave, "");
     muic_req.channel = channel;
     muic_req.addr = addr->s_addr;
     serv->ifr.ifr_data = (char *) &muic_req;
     res = ioctl(serv->ioctlSock, BOND_MUIC_ROUTE, &serv->ifr);
     if(res != 0)
	  err_msg("IOCTL: DeleteUnicastEntry (%s): Ioctl request failed for %s, %d", 
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  interface, channel);    
}


void
Ioctl_AddBroadcastEntry(ServerData *serv, const char *interface, int channel)
{
     char localAddr[INET_ADDRSTRLEN];
     struct muic_broadcast_req muic_req;
     int res;

#ifdef DEBUG_IOCTL   
     err_msg("IOCTL: AddBroadcastEntry (%s): Interface = %s, channel = %d", 
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     interface, channel);    
#endif

     if(debug_mode || serv->helloData->totalInterfaces == 1)
	  return;
  
     strcpy(muic_req.slave, interface);
     muic_req.channel = channel;
     serv->ifr.ifr_data = (char *) &muic_req;
     res = ioctl(serv->ioctlSock, BOND_MUIC_BROADCAST_CHANNEL, &serv->ifr);
     if(res != 0)
	  err_msg("IOCTL: AddBroadcastEntry (%s): Ioctl request failed for %s, %d", 
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  interface, channel); 
}
