/*****************************************************************************
 * mcr_route.c: Multichannel routing protocol
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 * Modifications by: Rishi Bhardwaj, <rbhardw2 at uiuc dot edu>
 * 
 *****************************************************************************/


/*************************** NOTES *******************************************

This file implements the multichannel routing protocol. The key features
in this file are:
1) Support for route discovery and route maintenance.
2) Support for maintaining routing table, and other tables to aid route discovery.
3) Support for controlling the kernel multichannel routing module.


Changes by Rishi Bhardwaj:

1) Neighbors do not get default entry into routing table. Quality of one hop link to a 
   neighbor may be very bad, in which case multi hop path to neighbor may be a better option.
   As dictated by the routing metrics ETT, WCETT, MCR ...   
1) Destination node may issue multiple RREP for different routes discovered by a RREQ.
   Earlier RREQ with the same sequence number as the RREQ last "accepted" (associated route accepted) 
   were discarded.
1) forwarded_rrep table maintained at all the nodes to keep tab of route discovered
   from a source to destination and the latest RREP seq. no. seen from a destination.
   This helps to suppress a RREP from destination to source if already a more recent, thus better,
   RREP has been forwarded to the source.
2) We use replied_rreq table at the destination to do book keeping of the last route discovered from a source.
   When a RREQ or RFRESH message is received from a source, the newly discovered route is compared to the
   old route. If the new route discovered is the same as the route in replied_rreq table, 
   we update the routing metric associated with the route. Otherwise if the routing metric for the new route 
   is better than the last route, the destination choses the new route and sends RREP to the source.
3) Introduce a new routing metric: MCR_CHAN_REUSE_CETT and its associated routing parameters. 
   A Modification of MCR metric.
4) Properly handled the situation where the forward and reverse routes between the source and destination differ.
5) Accept forwarding RERR netlink message from Kmcr module, send a RERR to the source of the packet dropped. 

5) Added the option of either sending a unicast or broadcast of RERR, depending whether the node knows the last hop
   of the failed route or not.

7) Added the check for routing management packets to be processed only when received from a symmetric one hop neighbor.
6) Added debug messages

Bugs Fixed:   
1) Fixed the computation of MCR, WCETT routing metric.
2) Deletion of routes upon receiving a RERR now properly handled.


TODO: Remove isBroadcast from forwarding RREQ function.
      Handle multiple sources for a common destination properly. 
      RERR when next hop for the destination fails needs to be sent to all sources.
      

*********************************************************************************/
#include "utilities.h"
#include "integrated.h"
#include "timer.h"
#include "mcr_route.h"
#include "mcr_hello.h"
#include "mcr_ioctl.h"
#include "kmcr_netlink.h"

extern DEBUG_MODE_TYPE debug_mode;


static void SetTimeout(ServerData *serv, struct timeval *delay, ROUTE_TIMEOUT_TYPE type, int data);


#ifdef DEBUG_ROUTE_DELAY
char debugStr[MAX_DEBUG_STR][MAX_DEBUG_STR_SIZE];
unsigned short debugStrCtr = 0;
unsigned int debugDispCtr = 1;
#endif


#ifdef ROUTE_EXPERIMENT

char route_debug_msg[MAX_ROUTE_DEBUG_MESSAGE_LENGTH];
extern FILE *routeDebugMsgFile;
extern FILE *routeNotTakenMsgFile;
void routeDebugMsg(FILE* file, const char *fmt, ...);
#endif


/*********************** Route metric computation ************************/

/* Computes the cost of the path discovered by RREQ so far.
 * lastHopCost is the ETT (Switching cost from last hop + ETX * Packet_Size/Trx_Rate)
 * from last hop on the RREQ path
 */

float
ComputeMetricCost(ServerData *serv, const char *pkt, int numEntries, float lastHopCost, int lastChannel, struct in_addr *route)
{
     HopEntry lastHop;
     float metricValue = 0;
     char packet[MAX_MESSAGE_LENGTH];
#ifdef ROUTE_EXPERIMENT
	  int route_msg_ctr=0;
#endif     
     
     memcpy(packet, pkt, sizeof(HopEntry)*numEntries);
     lastHop.metric = lastHopCost;
     lastHop.channel = lastChannel;
     lastHop.addr.s_addr = 0;
     copyStructIntoPkt(lastHop, packet+sizeof(HopEntry)*numEntries);
     numEntries++;
   
     switch(serv->routingData->routeMetricType) 
     {
     case MCR_SHORTEST_HOP: 
     {
	  int i;
	  HopEntry metric1;
	  metricValue = numEntries;
	  for(i = 0; i < numEntries; i++) {
	       copyStructFromPkt(metric1, packet + i*sizeof(HopEntry));

	       route[i] = metric1.addr;
	  }

	  break;
     }
     case MCR_SHORTEST_HOP_AND_DIVERSITY: 
     {
	  HopEntry metric1, metric2;
	  float diversityCount=0, nodeCount;
	  int i,j;

	  for(i =0; i < numEntries; i++) {
	       copyStructFromPkt(metric1, packet + i*sizeof(HopEntry));
	       route[i] = metric1.addr;
	       nodeCount = 0;
	       for( j = 1; j < MCR_DIVERSITY_COUNT_WINDOW && (i+j) < numEntries; j++) {
		    copyStructFromPkt(metric2,  packet + (i+j)*sizeof(HopEntry));

		    if(metric1.channel == metric2.channel)
			 nodeCount ++;
	       }
	       diversityCount += nodeCount;
	  }
	  metricValue = numEntries + (diversityCount / MCR_DIVERSITY_COUNT_WINDOW);
	  break;
     }

     case MCR_MCETT:   
     case MCR_WCETT:
     {
	  HopEntry metric1, metric2;
	  float sumCost=0, localCost=0, maxCost = 0;
	  int i,j,maxChannel=-1, bottleNeckEntry = -1;
	  
  
	  
	  
#ifdef DEBUG_ROUTE
	  err_msg("ROUTEDISC:debug_msg:Metric:WCETT/MCETT:Comput Metric Cost for RREQ so far\n");
#endif
	  
	  // compute aggregate cost
	  for(i = 0; i < numEntries; i++) {
	       copyStructFromPkt(metric1, packet + i*sizeof(HopEntry));

	       route[i] = metric1.addr;
	       sumCost += metric1.metric;

#ifdef DEBUG_ROUTE
	  err_msg("ROUTEDISC:debug_msg:Metric:%d %s  lnk metric %f, channel %u\n",i, print_ip(metric1.addr.s_addr),metric1.metric,metric1.channel);
#endif

#ifdef ROUTE_EXPERIMENT
	  route_msg_ctr += sprintf(route_debug_msg + route_msg_ctr,"%d)%s   %f   %u\n",
		    (i+1), print_ip(metric1.addr.s_addr), metric1.metric, metric1.channel);
#endif
	  
	       
	       // compute max on channel of current node - this is sub-optimal computation
	       localCost = metric1.metric;
	       for( j = 0; j < numEntries; j++) {
		    copyStructFromPkt(metric2,  packet + j*sizeof(HopEntry));

		    if(metric1.channel == metric2.channel && i != j)
			 localCost += metric2.metric;
	       }
	       if(localCost > maxCost)
	       {
		    maxCost = localCost;
		    maxChannel = metric1.channel;
		    bottleNeckEntry = i;
	       }
	  }
 	  metricValue = (1.0 - serv->routingData->beta)*sumCost + serv->routingData->beta*maxCost ;
 
#ifdef DEBUG_ROUTE
	  err_msg("ROUTEDISC:debug_msg:Metric:total metric %f, max channel %u max cost %f\n", metricValue, maxChannel, maxCost);
#endif

#ifdef ROUTE_EXPERIMENT
	  route_msg_ctr += sprintf(route_debug_msg + route_msg_ctr,"Metric: %f , max channel %u  max cost %f\n", metricValue, maxChannel, maxCost);
#endif
	  
	  break;
     }
     case MCR_CHAN_REUSE_CETT:
     {
	  HopEntry metric1, metric2;
	  float sumCost=0, localCost=0, maxCost = 0, tx_prob =1, new_txprob = 1;
	  int i,j,maxChannel=-1, k, bottleNeckEntry = -1;
#ifdef DEBUG_ROUTE
	  err_msg("ROUTEDISC:debug_msg:Metric:CHAN_REUSE:Comput Metric Cost for RREQ so far\n");
#endif
	  
	  // compute aggregate cost
	  for(i = 0; i < numEntries; i++) {
	       copyStructFromPkt(metric1, packet + i*sizeof(HopEntry));

	       route[i] = metric1.addr;
	       sumCost += metric1.metric;

#ifdef DEBUG_ROUTE
	       err_msg("ROUTEDISC:debug_msg:Metric:%d %s  lnk metric %f, channel %u\n",i, print_ip(metric1.addr.s_addr),metric1.metric,metric1.channel);
#endif

	       
#ifdef ROUTE_EXPERIMENT

	       route_msg_ctr += sprintf(route_debug_msg + route_msg_ctr,"%d)%s   %f   %u\n",
		    (i+1), print_ip(metric1.addr.s_addr), metric1.metric, metric1.channel);

#endif
	       //Checking if consective hops have consecutive channels, or if the intermediate hop tx_prob is small in which case
	       //the cross channel interference will further reduce the it.
	       if(i < numEntries -1)
	       {
		    copyStructFromPkt(metric2,  packet + (i+1)*sizeof(HopEntry));
		    if(areConsecutiveChannels(metric1.channel,metric2.channel))
		    {
			 tx_prob = PACKET_TRANSMISSION_TIME / metric1.metric;
			 if(tx_prob < CONSECTV_CHANNEL_THRSHHOLD_TX_PROB)
			 {
// 			      new_txprob = (tx_prob * (1 + tx_prob - CONSECTV_CHANNEL_THRSHHOLD_TX_PROB));
			    
 			      new_txprob = (tx_prob * tx_prob);
			      if(new_txprob < 0)
			      {
				   new_txprob = 0.1 ;
			      }	
			      metric1.metric = PACKET_TRANSMISSION_TIME / new_txprob;
#ifdef DEBUG_ROUTE
			      err_msg("ROUTEDISC:debug_msg:%d Metric:Channel %u and next one %u are consecutive new metric for this hop %f\n", 
				   i, metric1.channel, metric2.channel, metric1.metric);
#endif
			      
			 }
		    }
		    else if(metric1.channel != metric2.channel)           
		    {
			 // Concurrent transmission on the two consecutive hops is possible.
			 // Cross channel interference from transmission on second hop may increase error rate on first hop.
			 tx_prob = PACKET_TRANSMISSION_TIME/metric1.metric ;
			 if(tx_prob < THRSHHOLD_TX_PROB)
			 {
			      //The tx_prob of this intermediate hop is less. The cross channel interference on other hop
			      //on whichever channel will still increase the error rate on this hop.
			      //new_txprob = (tx_prob * (1 + tx_prob - THRSHHOLD_TX_PROB));
			      new_txprob = (tx_prob * tx_prob);
			      metric1.metric = PACKET_TRANSMISSION_TIME / new_txprob;
#ifdef DEBUG_ROUTE
			      err_msg("ROUTEDISC:debug_msg:%d Metric:tx_prob less than threshold, lowering it, new metric is %f\n", 
				   i, metric1.metric);
#endif
			      
			 }
		    }	      
		    
	       }

	       
	       // compute max on channel of current node - this is sub-optimal computation
	       localCost = metric1.metric;
	       for( j = i+1, k = 2; (j < numEntries) && (k < CHAN_REUSE_HOPS); j++, k++) {
		    copyStructFromPkt(metric2,  packet + j*sizeof(HopEntry));
		    if(metric1.channel == metric2.channel && i != j)
		    {
			 if(k >= 4)
			 {
			      localCost += metric2.metric * serv->routingData->alpha  ;
			 }
			 else
			 {
			       localCost += metric2.metric ;
			       k = 1;
			 }
		    }
	       }
	       if(localCost > maxCost)
	       {
		    maxCost = localCost;
		    maxChannel = metric1.channel;
		    bottleNeckEntry = i;
	       }
	  }
 	  metricValue = (1.0 - serv->routingData->beta)*sumCost + serv->routingData->beta*maxCost ;
 
#ifdef DEBUG_ROUTE
	  err_msg("ROUTEDISC:debug_msg:Metric:total metric %f, max channel %u max cost %f bottlEntr:%u\n", 
		    metricValue, maxChannel, maxCost, bottleNeckEntry);
#endif

#ifdef ROUTE_EXPERIMENT
	  route_msg_ctr += sprintf(route_debug_msg + route_msg_ctr,"Metric: %f , max channel %u  max cost %f bottlneckEntry:%u\n", 
			  metricValue, maxChannel, maxCost, bottleNeckEntry);
#endif


	  break;
     }
     break;
     default:
	  err_msg("ComputeMetricCost: Invalid metric type\n");
     }

#ifdef DEBUG_ROUTE
     {
	  char localAddr[INET_ADDRSTRLEN];
	  err_msg("ROUTETABLE:(%s):debug_msg:Metric cost is %f\n",
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  metricValue);
     }
#endif

#ifdef ROUTE_EXPERIMENT

     route_debug_msg[route_msg_ctr] = '\0';

#endif

     
     return metricValue;

}


float
ComputeETT(ServerData *serv, const struct in_addr *prevNode)
{
// we are using a constant for packet transmission time for simplicity. It is possible to query
// atheros drivers for the data rate using standard wireless extension ioctl calls and that can
// be used to compute a more accurate value. Since the initial testbed implementation uses
// fixed data rates, we have not coded in the probe for data rate.

     float packetTime = PACKET_TRANSMISSION_TIME; 
     float etx = GetNeighborDeliveryRatio(serv, prevNode);
#ifdef DEBUG_ROUTE
     err_msg("mcr_route.c:debug_msg:Metric:etx from %s is %f, ETT %f\n", print_ip(prevNode->s_addr), etx, etx*packetTime); 
#endif
     
     if(etx <= 0)
	  return packetTime*1000; // large number to signify link break

     return packetTime * etx;

}

float 
ComputeSwitchingCost(ServerData *serv,  RREQHeader *hdr)
{
     int i;
     for(i = 0; i < hdr->numChannels; i++)
     {
	  if(hdr->cList[i].num == serv->helloData->fixedChannel)
	  {
#ifdef DEBUG_ROUTE
	       err_msg("mcr_route.c:debug_msg:Metric:fixed_chann %u , switch cost from prevNode = %u\n",hdr->cList[i].num, hdr->cList[i].cost); 
#endif
	       return hdr->cList[i].cost;
	  }
     }
     return 0;
}

float
GetLastHopCost(ServerData *serv, const struct in_addr *prevNode, RREQHeader *hdr)
{
     // Right now we will just support shortest hop metric
    
     if((serv->routingData->routeMetricType == MCR_MCETT) || (serv->routingData->routeMetricType == MCR_CHAN_REUSE_CETT))
	  return hdr->lastHopCost + ComputeETT(serv, prevNode) + ComputeSwitchingCost(serv, hdr);
     else if(serv->routingData->routeMetricType == MCR_WCETT)
	  return hdr->lastHopCost + ComputeETT(serv, prevNode);

     return hdr->lastHopCost;

}

BOOL areConsecutiveChannels(int chan1, int chan2)
{
     int diff = chan1 - chan2;
     if((diff == 12) || (diff == -12) || (diff == 16) || (diff == -16))
	  return TRUE;
     return FALSE;
}

/**************** Manage tables used to support route discovery **********************/

void
UpdateForwardedList(ServerData *serv)
{
     ForwardedRREQ *list = serv->routingData->forwardedList, *prev = NULL, *delEntry;
    
     while(list != NULL) {
	  if(list->ticks > 0) {
	       list->ticks --;
	       prev = list;
	       list = list->next;
	  } else {
	       if(prev == NULL) 
		    serv->routingData->forwardedList = list->next;
	       else 
		    prev->next = list->next;

	       delEntry = list;
	       list = list->next;
	       free(delEntry);	    
	  }
     }
}


void
UpdateForwardedRREPList(ServerData *serv)
{
     ForwardedRREP *list = serv->routingData->forwardedRREPList, *prev = NULL, *delEntry;
    
     while(list != NULL) {
	  if(list->ticks > 0) {
	       list->ticks --;
	       prev = list;
	       list = list->next;
	  } else {
	       if(prev == NULL) 
		    serv->routingData->forwardedRREPList = list->next;
	       else 
		    prev->next = list->next;

	       delEntry = list;
	       list = list->next;
	       free(delEntry);	    
	  }
     }
}



void
UpdateRepliedList(ServerData *serv)
{
     RepliedRREQ *list = serv->routingData->repliedList, *prev = NULL, *delEntry;
    
     while(list != NULL) {
	  if(list->ticks > 0) {
	       list->ticks --;
	       prev = list;
	       list = list->next;
	  } else {
	       if(prev == NULL) 
		    serv->routingData->repliedList = list->next;
	       else 
		    prev->next = list->next;

	       delEntry = list;
	       list = list->next;
	       free(delEntry);	    
	  }
     }
}

/************************** Routing table management *************************/

const RouteTableEntry *
GetRouteTableEntry(ServerData *serv, const struct in_addr *dst)
{
     RouteTableEntry *list = serv->routingData->table.routeList;
     while( list != NULL) {
	  if( !memcmp( &list->destination, dst, sizeof(struct in_addr)))
	  {
	       if(list->ticks > 0)
		    return list;
	       return NULL;
	  }
	  list = list->next;
     }
     return NULL;
}

/* Note: be careful when adding entry to kernel table as local table maintains entry based on
   src-dst combination while kernel entry is based on dst only
*/

BOOL
AddRouteEntry(ServerData *serv, const struct in_addr *dst, const struct in_addr *src, const struct in_addr *nextHop, const struct in_addr *prevHop, float metric, char *route, ROUTE_TYPE type)
{
     RouteTableEntry *list = serv->routingData->table.routeList, *prev = NULL;

#ifdef DEBUG_MSG_ON
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];
     char nextAddr[INET_ADDRSTRLEN];

     err_msg("ROUTETABLE:debug_msg:(%s):Metric: Adding route to %s with nexthop %s with metric %f, type = %d\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, dst, remoteAddr, sizeof(remoteAddr)),
	     Inet_ntop(AF_INET, nextHop, nextAddr, sizeof(nextAddr)), metric, type    
	  );
#endif
     
     if(!IsOneHopNeighbor(serv, nextHop))
     {
#ifdef DEBUG_MSG_ON
	  err_msg("mcr_route.c:debug_msg:Cannot add to dst through the nexthop?\n");
#endif
	  return FALSE;
     }

     // Current policy is to maintain only one route to a node
     // but we remember routes based on combination of source-destination 
     // identifiers
     // So, if we already have a route, we replace it
     // 
     // But what if we already have route to dst from some other source
     // Multiple sources for the same destination is yet to be properly handeled
	
     while( list != NULL) {

	  //This should probably be changed to have a linked list of sources, so 
	  //as to accomodate multiple sources to a destination. This would be useful
	  //in providing RERR msg feedback to all sources on a route failure.
	  
	  if( !memcmp( &list->destination, dst, sizeof(struct in_addr))) 
	  {
#ifdef DEBUG_MSG_ON
	       err_msg("mcr_route.c:debug_msg: The route to %s already exists",
			      print_ip(list->destination.s_addr));
	       err_msg("mcr_route.c:debug_msg: The new source is %s \n", print_ip(src->s_addr));
#endif
	       memcpy(&list->source, src, sizeof(struct in_addr));
	       
	       // Need to update kernel only if nextHop is different
	       if(memcmp( &list->nextHop, nextHop, sizeof(struct in_addr)))
	       {    
#ifdef DEBUG_MSG_ON 
		    err_msg("mcr_route.c:debug_msg: Nexthop diff is %s\n",print_ip(list->nextHop.s_addr));
#endif
		    DeleteRemoteEntry(serv, &list->destination, &list->nextHop);
		    Ioctl_DeleteRoute(serv, &list->destination, &list->nextHop);
		
		    memcpy(&list->nextHop, nextHop, sizeof(struct in_addr));


		    AddRemoteEntry(serv, &list->destination, &list->nextHop);
		    Ioctl_AddRoute(serv, &list->destination, &list->nextHop);
	       }

	       memcpy(&list->prevHop, prevHop, sizeof(struct in_addr));	    
	       list->metric = metric;

	       if(list->route)
		    free(list->route);
	       list->route = route;
	       list->type = type;
	       list->ticks = ROUTE_ENTRY_TICKS;
	       list->refreshTicks = ROUTE_REFRESH_TICKS;
	       return TRUE;
	  }

	  prev = list;
	  list = list->next;
     }
     list = (RouteTableEntry *) malloc( sizeof(RouteTableEntry));
     memcpy(&list->destination, dst, sizeof(struct in_addr));
     memcpy(&list->source, src, sizeof(struct in_addr));
     memcpy(&list->nextHop, nextHop, sizeof(struct in_addr));
     memcpy(&list->prevHop, prevHop, sizeof(struct in_addr));

     AddRemoteEntry(serv, &list->destination, &list->nextHop);
     Ioctl_AddRoute(serv, &list->destination, &list->nextHop);

     list->metric = metric;
     list->route = route;
     list->type = type;
     list->ticks = ROUTE_ENTRY_TICKS;
     list->refreshTicks = ROUTE_REFRESH_TICKS;
     list->next = NULL;
    
     if(prev == NULL)
	  serv->routingData->table.routeList = list;
     else
	  prev->next = list;
     return TRUE;
}

// Used to find routes that may be affected when a link to a neighbor is down
RouteTableEntry *
ExtractAffectedRoutes(ServerData *serv, const struct in_addr *addr) 
{
     RouteTableEntry *list = serv->routingData->table.routeList, *prev = NULL, *returnList = NULL, *tmp;

     while(list != NULL) {
	  if( !memcmp(&list->nextHop, addr, sizeof(struct in_addr)) )
	  {
	       if(prev == NULL)
		    serv->routingData->table.routeList = list->next;
	       else
		    prev->next = list->next;
	       tmp = list;
	       list = list->next;
	       tmp->next = returnList;
	       returnList = tmp;
	  } else {
	       prev = list;
	       list = list->next;
	  }
     }
     return returnList;
}


void
UpdateExpireTimeInRoutingTable(ServerData *serv, const struct in_addr *addr, long ticksLeft) 
{
     RouteTableEntry *list = serv->routingData->table.routeList;

     while(list != NULL) 
     {
	  if( !memcmp(&list->destination, addr, sizeof(struct in_addr)) )
	  {
	       list->ticks = ticksLeft;
#ifdef DEBUG_MSG_ON
	       err_msg("mcr_route.c:debug_msg:Got the list for %s, making tick=%d\n",print_ip(addr->s_addr), ticksLeft);
#endif
	       return;
	  }
	  list = list->next;
     }


}

void
UpdateRoutingTable(ServerData *serv)
{
     RouteTableEntry *list = serv->routingData->table.routeList, *prev = NULL, *delEntry;
    
     while(list != NULL) {
	  if(list->ticks > 0) {

/* Userspace code does not know if a route is active because we do not know at userspace if some application is sending  
   packets over the route. Therefore, when the userspace thinks that the route may have expired, it asks the kernel
   if the route has been in use. If the kernel says that the route was in use, then we update the route expiry time.
   Otherwise, the route is deleted. To allow time for messages to be exchanged with kernel, we query the kernel some
   TICKS_QUERY_KERNEL seconds before the route is actually set to expire
*/	   
	       // Ask kernel to call back if route is valid
	       if(list->ticks == TICKS_QUERY_KERNEL)
	       {
#ifdef DEBUG_MSG_ON
		    err_msg("mcr_route.c:debug_msg:Asking kernel if the route to %s is still valid\n",
			   print_ip(list->destination.s_addr));   
#endif
		    Netlink_create_and_send_msg(serv, KMCR_ISROUTEVALID, &list->destination, &list->destination, 0, 0);
	       }

	       list->ticks --;

	       if(list->refreshTicks > 0)
		    list->refreshTicks --;
	       else
	       {
		    // Only source node refreshes route. Route refresh is a procedure to discover a better
		    // route, if available
		    if(list->type == SRC)
		    {
#ifdef DEBUG_MSG_ON
			 err_msg("mcr_route.c:debug_msg:Sending RFRESH RREQ to %s\n", print_ip(list->destination.s_addr));
#endif
			 InitiateRREQ(serv, &list->destination, TRUE);
		    }
		    list->refreshTicks = ROUTE_REFRESH_TICKS;
	       }
	       prev = list;
	       list = list->next;

	  } else {
	       if(prev == NULL) 
		    serv->routingData->table.routeList = list->next;
	       else 
		    prev->next = list->next;

	       delEntry = list;
	       list = list->next;
	       if(delEntry->route)
		    free(delEntry->route);

#ifdef DEBUG_MSG_ON 
	       err_msg("mcr_route.c:debug_msg:RouteEntry to %s  expired, deleting it\n", 
			 print_ip(delEntry->destination.s_addr));
#endif
	       
	       DeleteRemoteEntry(serv, &delEntry->destination, &delEntry->nextHop);
	       Ioctl_DeleteRoute(serv, &delEntry->destination, &delEntry->nextHop);
	       free(delEntry);	    
	  }
     }
}


// Called when periodic timer expires
void
UpdateTables(ServerData *serv)
{
     UpdateRoutingTable(serv);
     UpdateForwardedList(serv);
     UpdateForwardedRREPList(serv);
     UpdateRepliedList(serv);

#ifdef DEBUG_ROUTE_DELAY
     debugDispCtr<<=4;
     if(!debugDispCtr)
     {
	  int i = 0;
	  for(; i < debugStrCtr ; i++)
	  {
	       err_msg(debugStr[i]);
	  }
          debugStrCtr = 0;
	  debugDispCtr=1;
     }
#endif
     
     SetRouteTimeout(serv);
}

/******************** Function to construct a routing protocol packet ******************/

void
BuildAndSendRoutingPacket(ServerData *serv, ROUTE_PACKET_TYPE type,
			  char *packet, int totalSize, const struct sockaddr_in *socket)
{
     McrPacketHeader mcrHdr;
     RoutingProtocolHeader rHdr;
     socklen_t len = sizeof(struct sockaddr_in);

     mcrHdr.len = totalSize - sizeof(mcrHdr);
     mcrHdr.type = ROUTE_PKT;
     memcpy(&mcrHdr.fromAddr, &serv->myIpAddr, sizeof(struct in_addr));
     rHdr.type = type;
    
     copyStructIntoPkt(mcrHdr, packet);
     copyStructIntoPkt(rHdr, packet + sizeof(mcrHdr));

     if(debug_mode == LOCAL_ONLY)
	  Sendto(serv->sendfd, packet, totalSize, 0, (SA *) &serv->sendBroadcastSocket, len);
     else	
	  Sendto(serv->sendfd, packet, totalSize, 0, (SA *) socket, len);
}

/****************** Managing route errors ************************/

// Route failed
void
RouteFailed(ServerData *serv, const RouteTableEntry *entry)
{

#ifdef DEBUG_ROUTE    
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];    
     err_msg("ROUTEERROR:(%s): Route to %s failed\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, &entry->destination, remoteAddr, sizeof(remoteAddr))  
	  );
#endif
    
// Route failed uses the DISCFAILED message to ask kmcr to remove route from its table and free any pending
// packets that are buffered
     Netlink_create_and_send_msg(serv, KMCR_DISCFAILED, &entry->destination, &entry->destination, 0, 0);
}

void
SendRERR(ServerData *serv, const RouteTableEntry *entry, BOOL isBroadcast)
{
     char packet[MAX_MESSAGE_LENGTH];
     RERRHeader hdr;
     int size = sizeof(McrPacketHeader)+sizeof(RoutingProtocolHeader);

#ifdef DEBUG_ROUTE
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];
     err_msg("ROUTEERROR:(%s):debug_msg: Sending RERR to source %s\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, &entry->source, remoteAddr, sizeof(remoteAddr))  
	  );
#endif
   
     memcpy(&hdr.originator, &entry->source, sizeof(struct in_addr));
     memcpy(&hdr.destination, &entry->destination, sizeof(struct in_addr));
     copyStructIntoPkt(hdr, packet+size);
     size += sizeof(RERRHeader);


     // Create and send out RERR

     if(!isBroadcast)
     {

/* TODO:  We have information about the route that failed, we can unicast the RERR packet
 *	  to the prev. hop on the route. The problem is we may not have the route to the prev.
 *	  hop in the kernel routing table (remember one hop neighbors do not have
 *	  default routes added in kernel routing table). If we unicast the RERR to prev Hop in 
 *	  the route then the Kmcr module would direct us to discover a route to prev. hop. We could
 *	  add a one hop route to the prev. Hop in the kernel table and then unicast the RERR, just
 *	  as we do for RREPs. Since RERR are not as frequent as RREPs currently we do not do this.
 *	  We could also send the RERR as a unicast to the source of the the route, but this 
 *	  would have two problems:
 *	  1) The source can not verify if the RERR it receives is from a node which is actually
 *	  on the route towards the destination. It might be that source has a different route to
 *	  the destination and this may be a stale RERR from a node which does not have recent
 *	  information.
 *	  2) Routing information at helloServer of all the intermediate nodes from source uptill
 *	  this node for the destination will not be deleted. This is not such a big issue, since
 *	  this information will timeout.
 *
 *	  Currently we take the simple and easy approach and just broadcast the RERR. Later on a
 *	  more clean and efficient solution could be worked out.
 */

     //	  memcpy(&serv->sendUnicastSocket.sin_addr, &entry->prevHop, sizeof(struct in_addr));
     //	  memcpy(&serv->sendUnicastSocket.sin_addr, &entry->source, sizeof(struct in_addr));
     //	  BuildAndSendRoutingPacket(serv, RERR_PKT, packet, size, &serv->sendUnicastSocket);
	  BuildAndSendRoutingPacket(serv, RERR_PKT, packet, size,&serv->sendBroadcastSocket); 
     }
     else
     {

	  BuildAndSendRoutingPacket(serv, RERR_PKT, packet, size,&serv->sendBroadcastSocket); 
     }

         
}

void
ForwardRERR(ServerData *serv, const RERRHeader *hdr, const struct in_addr *nextHop)
{
     char packet[MAX_MESSAGE_LENGTH];
     struct sockaddr_in socket;
     int size = sizeof(McrPacketHeader)+sizeof(RoutingProtocolHeader);

#ifdef DEBUG_MSG_ON
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];
     err_msg("ROUTEERROR:(%s):debug_msg: Forwarding RERR to %s\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, nextHop, remoteAddr, sizeof(remoteAddr))  
	  );
#endif
    
     // socket for forwarding RERR
     bzero( &socket, sizeof(struct sockaddr_in));
     socket.sin_family = AF_INET;
     socket.sin_port = htons(HELLO_PORT);

     memcpy(&socket.sin_addr, nextHop, sizeof(struct in_addr));
    
     memcpy(packet + size, hdr, sizeof(RERRHeader));
     size += sizeof(RERRHeader);
    
//   BuildAndSendRoutingPacket(serv, RERR_PKT, packet, size, &socket);

/* TODO: Unicasting of RERR must be implemented for efficiency reasons. Currently we
 *       broadcast the RERR
 */     
     BuildAndSendRoutingPacket(serv, RERR_PKT, packet, size, &serv->sendBroadcastSocket);
}


void
ProcessRERR(ServerData *serv, const struct in_addr *prevHop, const char *pkt, int size)
{
     RERRHeader hdr;
     // Changes by Rishi, deleting from the linked list correctly implemented.
     RouteTableEntry *list = serv->routingData->table.routeList, *prev = NULL;

#ifdef DEBUG_MSG_ON    
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];
     err_msg("ROUTEERROR:(%s):debug_msg: Received RERR from %s\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, prevHop, remoteAddr, sizeof(remoteAddr))  
	  );
#endif
    
     copyStructFromPkt(hdr, pkt);
     pkt += sizeof(hdr);

     while(list != NULL) {

	  // we have a match
	  // We check if next hop for the destination is the node which sent us the RERR 
	  if( !memcmp( &list->nextHop, prevHop, sizeof(struct in_addr)) &&
          !memcmp( &list->destination, &hdr.destination, sizeof(struct in_addr))) 
	  {
	       
	       if(prev == NULL)
	       {
		    serv->routingData->table.routeList = list->next;
	       }
	       else
	       {
		    prev->next = list->next;
	       }

	       // Not at source
	       if( memcmp( &list->source, &serv->myIpAddr, sizeof(struct in_addr))) 
		    ForwardRERR(serv, &hdr, &list->prevHop);
	       else // at source
		    RouteFailed(serv, list);
		
	       if(list->route)
		    free(list->route);
	       DeleteRemoteEntry(serv, &list->destination, &list->nextHop);
	       Ioctl_DeleteRoute(serv, &list->destination, &list->nextHop);
	       free(list);
	       return;  // we have only one entry, so this is ok	    
	  }
	  prev = list;
	  list = list->next;
     }
}


/****************** Route Reply code ***************************/

void
RREQSucceeded(ServerData *serv, const struct in_addr *addr)
{
     // This is just a placeholder function for debugging

#ifdef DEBUG_MSG_ON    
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];    
     err_msg("ROUTEDISC:debug_msg:(%s): Route discovery attempt succeeded to %s\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, addr, remoteAddr, sizeof(remoteAddr))  
	  );
#endif

}

void
RREPReceivedAtSource(ServerData *serv, const struct in_addr *prevHop, const char *pkt, const RREPHeader *hdr)
{
     PendingRREQ *list = serv->routingData->pendingList;
     PendingRREQ *prev = NULL;
     BOOL isSuccess;
   
#ifdef DEBUG_ROUTE_DELAY
     struct timeval tv;
     if(gettimeofday(&tv, NULL))
     {
	  err_msg("ROUTE_DELAY:debug_msg:Error gettimeofday while receiving RREP");
     }
#endif

     // Delete pending status of all RREQs to that destination
     while( list != NULL) {
	  if( !memcmp(&list->destination, &hdr->destination, sizeof(struct in_addr)) ) 
	  {

#ifdef DEBUG_ROUTE_DELAY
	       snprintf(debugStr[debugStrCtr],MAX_DEBUG_STR_SIZE,"ROUTE_DELAY:debug_msg:RREQ sent for %s  seq:%u:at sec:%lu:usec:%lu:\n",
			 print_ip(list->destination.s_addr), list->seq, list->tv.tv_sec, list->tv.tv_usec);
	       debugStrCtr=(debugStrCtr+1)% MAX_DEBUG_STR;
#endif

	       if(prev == NULL) 
		    serv->routingData->pendingList = list->next;
	       else
		    prev->next = list->next;
	    
	       free(list);
	    
	       if(prev == NULL)
		    list = serv->routingData->pendingList;
	       else
		    list = prev->next;
	  } else {
	       prev = list;
	       list = list->next;
	  }
     }  
    

#ifdef DEBUG_ROUTE_DELAY
	       snprintf(debugStr[debugStrCtr],MAX_DEBUG_STR_SIZE,"ROUTE_DELAY:debug_msg:RREP recvd frm %s, src seq:%u:, dst seq:%u:, at sec:%lu:usec:%lu:\n",
			 print_ip(hdr->destination.s_addr), hdr->seq, hdr->dstSeq, tv.tv_sec, tv.tv_usec);
       
	       debugStrCtr=(debugStrCtr+1)% MAX_DEBUG_STR;
#endif

	       
#ifdef DEBUG_ROUTE

     err_msg("ROUTEDISC:debug_msg:Metric:RREP Received at destination\n");
#endif
  
     // AddEntry in routing table
     isSuccess = AddRouteEntry(serv, &hdr->destination, &hdr->originator, prevHop, &hdr->originator, hdr->metric, NULL, SRC);
        
     // Hook to inform successful route discovery
     if(isSuccess){
	    RREQSucceeded(serv, &hdr->destination);
     }
}


BOOL ShouldProcessRREP(ServerData *serv, const RREPHeader *hdr)
{
     ForwardedRREP *list = serv->routingData->forwardedRREPList, *prev = NULL;
     
     
     // See if a newer RREP from the destination has already been seen before
     while( list != NULL) {

	  if( !memcmp(&list->destination, &hdr->destination, sizeof(struct in_addr))) {
	       
	       if(!memcmp(&list->source, &hdr->originator, sizeof(struct in_addr)))
	       {
		    if(list->dstSeq < hdr->dstSeq)
		    {
			 // We process the RREP, since its newer than the one we have seen before
			 list->dstSeq = hdr->dstSeq;
			 list->ticks = FORWARDED_RREP_TABLE_TICKS ; 
			 return TRUE;
		    }
		    else
		    {
#ifdef DEBUG_ROUTE
			 err_msg("ROUTEDISC:debug_msg:Metric:The RREP (seq %u) is stale, we have earlier seen seq %u\n", hdr->dstSeq, list->dstSeq);
#endif
			 return FALSE;
		    }
	       }
	  }

	  prev = list;
	  list = list->next;
     }
    
     // Insert new entry
     list = (ForwardedRREP *) malloc( sizeof(ForwardedRREP));
     list->next = NULL;
     memcpy(&list->source, &hdr->originator, sizeof(struct in_addr));
     memcpy(&list->destination, &hdr->destination, sizeof(struct in_addr));
     list->dstSeq = hdr->dstSeq;
     list->srcSeq = hdr->seq ;
     list->metric = hdr->metric;
     list->ticks = FORWARDED_RREP_TABLE_TICKS;

     if(prev == NULL)
	  serv->routingData->forwardedRREPList = list;
     else
	  prev->next = list;
     return TRUE;
}





/*
 * nextHop refers to the next hop from the perspective of RREQ
 * but is the previous hop with respect to RREP
 */
void
ProcessRREP(ServerData *serv, const struct in_addr *nextHop, const char *pkt, int size)
{
     RREPHeader hdr;
     HopEntry entry;
     char packet[MAX_MESSAGE_LENGTH];
     struct sockaddr_in socket;
     const struct in_addr* prevHop;
     BOOL isSuccess;

     int pktSize = sizeof(McrPacketHeader)+sizeof(RoutingProtocolHeader);

     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];

#ifdef DEBUG_ROUTE    
     err_msg("ROUTEDISC:debug_msg:(%s): Received RREP from %s\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, nextHop, remoteAddr, sizeof(remoteAddr))  
	  );
#endif
    
     copyStructFromPkt(hdr, pkt);
     pkt += sizeof(hdr);

     //////// Source node
#ifdef DEBUG_ROUTE
     err_msg("ROUTEDISC:(%s):debug_msg:Metric: Received RREP from %s seq %u\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, nextHop, remoteAddr, sizeof(remoteAddr)), hdr.dstSeq);  
#endif
     
     if(!ShouldProcessRREP(serv, &hdr))
     {
	  return;
     }
     
     if( !memcmp(&hdr.originator, &serv->myIpAddr, sizeof(struct in_addr)) ) 
	  return RREPReceivedAtSource(serv, nextHop, pkt, &hdr);
	
     //////// Intermediate node: should forward packet RREP

     // Sanity check: Last entry should be this node
     copyStructFromPkt(entry, pkt + sizeof(HopEntry)* (hdr.numEntries - 1));
     if( memcmp(&entry.addr, &serv->myIpAddr, sizeof(struct in_addr))  || hdr.numEntries < 1) {
	  err_msg("ROUTEDISC:debug_msg:(%s): I am not last hop on RREP - acutal address is %s\n",
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  Inet_ntop(AF_INET, &entry.addr, remoteAddr, sizeof(remoteAddr))  
	       );
	  return;
     }

     // socket for forwarding RREP
     bzero( &socket, sizeof(struct sockaddr_in));
     socket.sin_family = AF_INET;
     socket.sin_port = htons(HELLO_PORT);

     // Add route 
     if(hdr.numEntries > 1) {
	  copyStructFromPkt(entry, pkt + sizeof(HopEntry)* (hdr.numEntries - 2));
	  prevHop = &entry.addr;
     }  else {
	  prevHop = &hdr.originator;
     }
    
     // Forward route
     isSuccess = AddRouteEntry(serv, &hdr.destination, &hdr.originator, nextHop, prevHop, hdr.metric, NULL, FORWARD);
    
     if(!isSuccess)
     {
	  err_msg("ROUTEDISC:debug_msg:(%s): We could not add route to %s in kernel, and RREP is being dropped\n",
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  Inet_ntop(AF_INET, &hdr.destination, remoteAddr, sizeof(remoteAddr))  
	       );
	  return;
     }

#ifdef DEBUG_ROUTE
     err_msg("ROUTEDISC:debug_msg:forwarding RREP to %s for src %s\n", Inet_ntop(AF_INET, prevHop, localAddr, sizeof(localAddr)),Inet_ntop(AF_INET,  &hdr.originator, remoteAddr, sizeof(remoteAddr)));

#endif
     
     // Reverse route
     if(hdr.originator.s_addr != prevHop->s_addr)
     {
	     //We will send the RREP as a unicast to the prevHop, thus we need a unicast ROUTE entry for prevHop
	     isSuccess = AddRouteEntry(serv, prevHop, &hdr.destination, prevHop, nextHop, hdr.metric, NULL, OVERHEARD);
     }
    
     if(!isSuccess)
     {
	  err_msg("ROUTEDISC:(%s):debug_msg: We could not add route to %s in kernel, and RREP is being dropped\n",
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  Inet_ntop(AF_INET, prevHop, remoteAddr, sizeof(remoteAddr))  
	       );
	  return;
     }

     isSuccess = AddRouteEntry(serv, &hdr.originator, &hdr.destination, prevHop, nextHop, hdr.metric, NULL, OVERHEARD);

     if(!isSuccess)
     {
	  err_msg("ROUTEDISC:(%s):debug_msg: We could not add route to %s in kernel, and RREP is being dropped\n",
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  Inet_ntop(AF_INET, &hdr.originator, remoteAddr, sizeof(remoteAddr))  
	       );
	  return;
     }
	
     memcpy(&socket.sin_addr, prevHop, sizeof(struct in_addr));

     hdr.numEntries --;  // As we want to remove our own entry

     // Forward RREP
     copyStructIntoPkt(hdr, packet+pktSize);
     pktSize += sizeof(RREPHeader);
     
     // Copy old route into new packet
     if(hdr.numEntries > 0) {
	  memcpy(packet+pktSize, pkt, sizeof(HopEntry) * hdr.numEntries );
	  pktSize += sizeof(HopEntry) * hdr.numEntries;
     }

     BuildAndSendRoutingPacket(serv, RREP_PKT, packet, pktSize, &socket);
}

void
SendRREP(ServerData *serv, const RREQHeader *hdr, const char *pkt, float metric)
{
     HopEntry entry;
     char packet[MAX_MESSAGE_LENGTH];
     RREPHeader rHdr;
     int size = sizeof(McrPacketHeader)+sizeof(RoutingProtocolHeader);
     struct sockaddr_in socket;
     const struct in_addr *nextHop;
     BOOL isSuccess;

     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];

     // socket for forwarding RREP
     bzero( &socket, sizeof(struct sockaddr_in));
     socket.sin_family = AF_INET;
     socket.sin_port = htons(HELLO_PORT);

     // Add reverse route to routing table
     if(hdr->numEntries >= 1) {
	  copyStructFromPkt(entry, pkt + (hdr->numEntries - 1) * sizeof(HopEntry)); // prev hop
	  nextHop = &entry.addr;

	  //We would unicast RREP to the next hop neighbor, thus we need a unicast ROUTE entry for next hop neighbor too. 
     	  isSuccess = AddRouteEntry(serv, nextHop, &hdr->destination, nextHop, &hdr->destination, metric, NULL, OVERHEARD);

	  if(!isSuccess)
          {
	       err_msg("ROUTEDISC:debug_msg:(%s): We could not add route to %s in kernel, and RREP is being dropped\n",
	       Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	       Inet_ntop(AF_INET, nextHop, remoteAddr, sizeof(remoteAddr))  
	       );
	       return;
	  }

     } else { // 1 hop route
	  nextHop = &hdr->originator;
     }

     
     isSuccess = AddRouteEntry(serv, &hdr->originator, &hdr->destination, nextHop, &hdr->destination, metric, NULL, OVERHEARD);

     if(!isSuccess)
     {
	  err_msg("ROUTEDISC:debug_msg:(%s): We could not add route to %s in kernel, and RREP is being dropped\n",
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
		  Inet_ntop(AF_INET, &hdr->originator, remoteAddr, sizeof(remoteAddr))  
	       );
	  return;
     }


#ifdef DEBUG_ROUTE    
     err_msg("ROUTEDISC:debug_msg:Metric: Sending RREP to %s for source %s new metric %f\n",
	     Inet_ntop(AF_INET, nextHop, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, &hdr->originator, remoteAddr, sizeof(remoteAddr)), metric);
#endif

     
     memcpy(&socket.sin_addr, nextHop, sizeof(struct in_addr));

     // Create and send out RREP
     memcpy(&rHdr.originator, &hdr->originator, sizeof(struct in_addr));
     memcpy(&rHdr.destination, &hdr->destination, sizeof(struct in_addr));
     rHdr.seq = hdr->seq;
     rHdr.dstSeq = serv->routingData->seq++ ; 
     rHdr.numEntries = hdr->numEntries;
     rHdr.metric = metric;

     copyStructIntoPkt(rHdr, packet+size);
     size += sizeof(RREPHeader);
    
     // Copy old route into new packet
     if(rHdr.numEntries > 0) {
	  memcpy(packet+size, pkt, sizeof(HopEntry) * rHdr.numEntries );
	  size += sizeof(HopEntry) * rHdr.numEntries;
     }

     BuildAndSendRoutingPacket(serv, RREP_PKT, packet, size, &socket);
}



void
RREQReceivedAtDest(ServerData *serv, const RREQHeader *hdr, const char *pkt)
{
     RepliedRREQ *list = serv->routingData->repliedList, *prev = NULL;
     struct in_addr route[MAX_RREQ_TTL + 1];
     float metric = ComputeMetricCost(serv, pkt, hdr->numEntries, hdr->lastHopCost, serv->helloData->fixedChannel, route);
     int z;
#ifdef DEBUG_ROUTE   
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];    
     err_msg("ROUTEDISC: debug_msg:(%s): Received RREQ at destination from source %s, metric cost = %f\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, &hdr->originator, remoteAddr, sizeof(remoteAddr)), metric);
#endif

     // Check if reply has already been processed
     while(list != NULL) 
     {
	  if( !memcmp(&list->source, &hdr->originator, sizeof(struct in_addr)))
	  {
	       
	       if(list->seq > hdr->seq)
	       {
		    // We do nothing if we see a stale RREQ
		       
#ifdef DEBUG_ROUTE
		    err_msg("ROUTEDISC: debug_msg:(%s):Metric:RREQ > seq num exists dest %s before metric =%f new=%f\n",
			Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     		Inet_ntop(AF_INET, &hdr->originator, remoteAddr, sizeof(remoteAddr)), list->metric,metric);
#endif
		    return;  
	       }
	       
#ifdef DEBUG_ROUTE
	       err_msg("ROUTEDISC: debug_msg:(%s):Metric:RREQ existed with seq %d, metric %f new RREQ seq %d metric %f src %s\n",
	 		Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)), list->seq,list->metric, hdr->seq, 
	     		metric,Inet_ntop(AF_INET, &hdr->originator, remoteAddr, sizeof(remoteAddr)));
#endif


	       list->ticks = REPLIED_TABLE_TICKS;     // Refreshing the RREQ replied list
	       if(hdr->isRefresh == TRUE)
	       { 
		    // RREQ is actually a RREFRESH message.
		    if((list->metric * REFRESH_METRIC_FACTOR) > metric)
		    {

			 SendRREP(serv, hdr, pkt, metric);
#ifdef DEBUG_ROUTE
			 err_msg("ROUTEDISC:debug_msg:Metric:entry to %s existed, RFRESH RREQ, before metric %f, new %f, take the new route, hop = %d\n", 
			      print_ip(hdr->originator.s_addr), list->metric, metric, hdr->numEntries+1);
#endif


#ifdef ROUTE_EXPERIMENT
			 routeDebugMsg(routeDebugMsgFile, "----------Entry %s existed, RRFRESH seq %u , dst seq %u take new route--------\n%s earlier metric %f\n",
					 print_ip(hdr->originator.s_addr), hdr->seq, (serv->routingData->seq -1),route_debug_msg, list->metric);
			 
#endif		
	 
			 list->seq = hdr->seq;
			 list->metric = metric;
			 list->hops = hdr->numEntries + 1;
			 memcpy((void*)list->route, (void*)route, hdr->numEntries*sizeof(struct in_addr));
			 
#ifdef DEBUG_ROUTE
			 for(z = 0 ; z < list->hops - 1 ; z++)
			 {
				 err_msg("ROUTEDISC:debug_msg:%d %s\n", (z+1), print_ip(list->route[z].s_addr));
			 }
#endif

			      
		    }
		    else if(list->hops == (hdr->numEntries + 1))
		    {
#ifdef DEBUG_ROUTE
			 err_msg("ROUTEDISC:debug_msg:Metric:hops for chosen route same as numEntries = %d seeing if same route\n", list->hops);
#endif
			 if((list->hops == 1) ||(!memcmp((void*)list->route, (void*)route,hdr->numEntries*sizeof(struct in_addr))))
			 {
			      // Received RREQ having traversed the same route as the one being used currently.
			      
#ifdef DEBUG_ROUTE
			      err_msg("ROUTEDISC:debug_msg:Metric:entry to %s exists, RFRESH RREQ for same route, before metric %f, new %f\n", 
				   print_ip(hdr->originator.s_addr), list->metric, metric);
#endif

			      
			      if(((metric * UPDATE_CURR_ROUTE_METRIC_FACTOR) > list->metric) || (list->metric > metric))
			      {     
#ifdef DEBUG_ROUTE
				   err_msg("ROUTEDISC:debug_msg:Metric:Updating the old metric with new metric\n");
#endif
				   
				   list->metric = metric;

			      }
			 }
		    }
			      
			 
	       }
	       else if((list->metric > metric) || (list->seq < hdr->seq))
	       {
			 
		    SendRREP(serv, hdr, pkt, metric);
#ifdef DEBUG_ROUTE
		    err_msg("ROUTEDISC:debug_msg:Metric:entry fr %s existed, new RREQ, before metric %f, new %f, take the new route hop %d\n", 
			 print_ip(hdr->originator.s_addr), list->metric, metric, (hdr->numEntries + 1));
#endif

#ifdef ROUTE_EXPERIMENT
			 routeDebugMsg(routeDebugMsgFile,"----------Entry %s existed, new RREQ seq %u , dst seq %u take new route-----\n%s -----earlier metric:%f------\n",
					 print_ip(hdr->originator.s_addr), hdr->seq, (serv->routingData->seq -1), route_debug_msg, list->metric);
			 
#endif		

		         list->seq = hdr->seq;
		         list->metric = metric;
			 list->hops = hdr->numEntries + 1;
			 memcpy((void*)list->route, (void*)route, hdr->numEntries*sizeof(struct in_addr));
			 
			 
#ifdef DEBUG_ROUTE
			 for(z = 0 ; z < list->hops - 1 ; z++)
			 {
				 err_msg("ROUTEDISC:debug_msg:%d %s\n", (z+1), print_ip(list->route[z].s_addr));
			 }
#endif


#ifdef DEBUG_ROUTE_DELAY
			 snprintf(debugStr[debugStrCtr],MAX_DEBUG_STR_SIZE,"ROUTE_DELAY:debug_msg:RREP sent for %s seq:%u:, dst seq:%u:, hops:%u:\n",
			 print_ip(hdr->originator.s_addr), hdr->seq,(serv->routingData->seq -1), list->hops);
			 debugStrCtr=(debugStrCtr+1)% MAX_DEBUG_STR;
 
#endif


	       }
	       else
	       {
		       // Worse than the earlier discovered route. Not taking any action
#ifdef ROUTE_EXPERIMENT
			 routeDebugMsg(routeNotTakenMsgFile, "----------Entry %s existed, new RREQ seq %u , dst seq %u ,  NOT TAKING THIS ROUTE-----\n%s -----earlier metric:%f------\n",
					 print_ip(hdr->originator.s_addr), hdr->seq, (serv->routingData->seq -1), route_debug_msg, list->metric);
			 
#endif		


	       }
	       return;
	  }
	  prev = list;
	  list = list->next;
     }

   
     list = (RepliedRREQ *) malloc( sizeof(RepliedRREQ));

     list->next = NULL;
     list->seq = hdr->seq;
     list->metric = metric;
     list->hops = hdr->numEntries + 1;
     memcpy(&list->source, &hdr->originator, sizeof(struct in_addr));
     list->ticks = REPLIED_TABLE_TICKS;
     memcpy((void*)list->route, (void*)route, hdr->numEntries*sizeof(struct in_addr));
     if(prev == NULL)
	  serv->routingData->repliedList = list;
     else
	  prev->next = list;
     
     SendRREP(serv, hdr, pkt, metric);
#ifdef DEBUG_ROUTE
     err_msg("ROUTEDISC:debug_msg:Metric:entry fr %s doesnt exist, seq %d new metric %f, take the new route, hop %d\n", 
	       print_ip(hdr->originator.s_addr), list->seq, metric, list->hops);
     for(z = 0 ; z < list->hops - 1 ; z++)
     {
	  err_msg("ROUTEDISC:debug_msg:%d %s\n", (z+1), print_ip(list->route[z].s_addr));
     }

#endif


#ifdef ROUTE_EXPERIMENT
     routeDebugMsg(routeDebugMsgFile,"----------Entry %s created, RREQ seq %u, dst seq %u take new route--------\n%s\n",
	       print_ip(hdr->originator.s_addr), hdr->seq, (serv->routingData->seq -1), route_debug_msg);
			 
#endif		


		    
#ifdef DEBUG_ROUTE_DELAY
     snprintf(debugStr[debugStrCtr],MAX_DEBUG_STR_SIZE,"ROUTE_DELAY:debug_msg:RREP sent for %s seq:%u:, dst seq:%u:, hops:%u:\n",
	       print_ip(hdr->originator.s_addr), hdr->seq,(serv->routingData->seq -1), list->hops);
     debugStrCtr=(debugStrCtr+1)% MAX_DEBUG_STR;
 
#endif



}


/***************** Route Request code ******************/

void
InsertCostIntoRREQ(ServerData *serv, RREQHeader *hdr)
{
    
     // Query driver for channel statistics
     if(serv->helloData->totalInterfaces > 1)
     {
	  struct muic_stats_u_response stats;
	  int i, totalUsage=0, idle;
	  HelloServerData *helloData = serv->helloData;
	  float cost;
	
	  hdr->numChannels = helloData->totalChannels;

	  // Right now, we get stats for first non-fixed interface
	  for(i = 0; i < helloData->totalInterfaces; i++)
	       if(helloData->inter[i].isFixed == FALSE)
		    break;

	  Ioctl_GetSwitchingStats(serv, helloData->inter[i].name, &stats);
	  for( i = 0; i < stats.nchan; i++)
	       totalUsage += stats.usage[i];
	
	  idle = MAX_CHANNEL_USAGE - totalUsage;
	  if(idle < 0)
	       idle = 0;
	
#ifdef DEBUG_ROUTE
	  err_msg("ROUTEDISC:SWITCH_COST:debug_msg:Metric:idle = %d, total usage = %d, chan = %d\n", idle, totalUsage, stats.nchan);
#endif

	  for( i = 0; i < stats.nchan; i++)
	  {
	       hdr->cList[i].num = stats.channel[i];
	       if(stats.usage[i] > MAX_CHANNEL_USAGE)
	       {
#ifdef DEBUG_ROUTE
		    err_msg("ROUTEDISC:SWITCH_COST:debug_msg:Metric:stats:Chan = %d, usage = %u > MAX_CHAN_USAGE\n", 
			      stats.channel[i], stats.usage[i]);
			
#endif
		    stats.usage[i] = MAX_CHANNEL_USAGE - idle;
	       }
	       cost = ((MAX_CHANNEL_USAGE - stats.usage[i] - ((float)idle)) * SWITCHING_COST) / (float)MAX_CHANNEL_USAGE;
	       if(cost < 0)
		    cost  = 0;
	       hdr->cList[i].cost = cost; 
#ifdef DEBUG_ROUTE
	       err_msg("ROUTEDISC:SWITCH_COST:debug_msg:Metric:stats:Chan = %d, usage = %u, cost = %f\n", stats.channel[i], stats.usage[i], cost);
#endif
	  }
	
     } else
	  hdr->numChannels = 0;

     switch(serv->routingData->routeMetricType)
     {
     case MCR_CHAN_REUSE_CETT:	  
     case MCR_MCETT:
     case MCR_WCETT:
     {
	  hdr->lastHopCost = 0;
	  break;
     }
     default:
	  hdr->lastHopCost = 1;
     }
}


void
SendRREQ(ServerData *serv, struct in_addr *addr, BOOL isRefresh)
{
     RREQHeader hdr;
     char packet[MAX_MESSAGE_LENGTH];
     int preHdrSize = sizeof(McrPacketHeader)+sizeof(RoutingProtocolHeader);
     int totalSize;

#ifdef DEBUG_ROUTE   
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];    
     err_msg("ROUTEDISC:(%s): Sent a route discovery request to destination %s\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, addr, remoteAddr, sizeof(remoteAddr))  
	  );
#endif 

     memcpy(&hdr.originator, &serv->myIpAddr, sizeof(struct in_addr));
     memcpy(&hdr.destination, addr, sizeof(struct in_addr));
     hdr.numEntries = 0;
     hdr.ttl = MAX_RREQ_TTL;
     hdr.seq = serv->routingData->seq;
     hdr.isRefresh = isRefresh;

     InsertCostIntoRREQ(serv, &hdr);

     copyStructIntoPkt(hdr, packet+preHdrSize);
     totalSize = preHdrSize + sizeof(hdr);

     BuildAndSendRoutingPacket(serv, RREQ_PKT, packet, totalSize, &serv->sendBroadcastSocket);
}


void
InitiateRREQ(ServerData *serv, struct in_addr *addr, BOOL isRefresh)
{
     PendingRREQ *list = serv->routingData->pendingList;
     PendingRREQ *prev = NULL;
     struct timeval delay;
     unsigned int length = 0;
// Send a new RREQ only if a request is not already present
     while( list != NULL) {
	  if( !memcmp(&list->destination, addr, sizeof(struct in_addr)) )
	  {
	       err_msg("ROUTEDISC:debug_msg:RREQ for %s was ordered, but it is in  pendList\n"
			 ,print_ip(list->destination.s_addr));
	       return;
	  }

	  prev = list; 
	  list = list->next;
	  length++;
     }
    
     list = (PendingRREQ *) malloc( sizeof(PendingRREQ) );
     list->numAttempts = 1;
     list->isRefresh = isRefresh;
     list->seq = serv->routingData->seq;
     memcpy(&list->destination, addr, sizeof(struct in_addr) );
     list->next = NULL;

     if(prev == NULL)
	  serv->routingData->pendingList = list;
     else
	  prev->next = list;
#ifdef DEBUG_MSG_ON
     err_msg("ROUTEDISC:debug_msg: Addr %s is added in RREQ list at position:%d",
	   print_ip(addr->s_addr),length);
#endif

#ifdef DEBUG_ROUTE_DELAY
     err_msg("ROUTE_DELAY:debug_msg:sending RREQ for %s is refresh %u\n", print_ip(addr->s_addr), isRefresh);
#endif

     
     SendRREQ(serv, addr, isRefresh);

     
#ifdef DEBUG_ROUTE_DELAY
     if(gettimeofday(&(list->tv), NULL))
     {
	  err_msg("ROUTE_DELAY:debug_msg:gettimeofday returned error\n");
     }
//     debugDispCtr = 1;
#endif 
      
     delay.tv_usec = 0;
     delay.tv_sec = ROUTE_DISCOVERY_TIMEOUT;
     SetTimeout(serv, &delay, RREQ_TIMEOUT, serv->routingData->seq);
     ++serv->routingData->seq;
}

void
RREQFailed(ServerData *serv, const struct in_addr *addr)
{

#ifdef DEBUG_ROUTE       
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];    
     err_msg("ROUTEDISC:(%s): Route discovery attempt failed to %s\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, addr, remoteAddr, sizeof(remoteAddr))  
	  );
#endif

     Netlink_create_and_send_msg(serv, KMCR_DISCFAILED, addr, addr, 0, 0);
}

void
ReInitiateRREQ(ServerData *serv, int seq)
{
     PendingRREQ *list = serv->routingData->pendingList;
     PendingRREQ *prev = NULL;
     struct timeval delay;

     // First see if the entry is still in the list
     while( list != NULL) {
	  if(list->seq == seq )
	       break;
	  prev = list;
	  list = list->next;
     }  
     // Not found, so the rreq must have succeeded earlier
     if(list == NULL)
	  return; 
    
     if( (list->isRefresh == TRUE) || (++list->numAttempts > MAX_DISCOVERY_ATTEMPTS)) {
	  if(list->isRefresh != TRUE)
	       RREQFailed(serv, &list->destination);
	  if(prev == NULL)
	       serv->routingData->pendingList = list->next;
	  else
	       prev->next = list->next;
	  free(list);
	  return;
     }
    
     list->seq = serv->routingData->seq;

#ifdef DEBUG_ROUTE_DELAY
     err_msg("ROUTE_DELAY:debug_msg:reinitiating send RREQ for %s is refresh %u\n", print_ip(list->destination.s_addr), list->isRefresh);
#endif
 
     SendRREQ(serv, &list->destination, list->isRefresh);
  
     delay.tv_usec = 0;
     delay.tv_sec = ROUTE_DISCOVERY_TIMEOUT *(1 << (list->numAttempts - 1)); // exponential backoff
     SetTimeout(serv, &delay, RREQ_TIMEOUT, serv->routingData->seq);
     ++serv->routingData->seq;
}



/* Function forwards (broadcasts) the RREQ packet.
 */
void
ForwardRREQ(ServerData *serv, RREQHeader *hdr, const char *pkt)
{
     char packet[MAX_MESSAGE_LENGTH];
     HopEntry entry;
     int size = sizeof(McrPacketHeader)+sizeof(RoutingProtocolHeader);

#ifdef DEBUG_ROUTE   
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];    
     err_msg("ROUTEDISC:(%s): Forwarded RREQ to destination %s\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, &hdr->destination, remoteAddr, sizeof(remoteAddr))  
	  );
#endif

     // Create next entry before updating hdr
     memcpy(&entry.addr, &serv->myIpAddr, sizeof(struct in_addr));
     entry.metric = hdr->lastHopCost;
     entry.channel = serv->helloData->fixedChannel; // channel on which RREQ was received

     ++hdr->numEntries;
     --hdr->ttl;

     InsertCostIntoRREQ(serv, hdr);
        
     memcpy(packet+size, hdr, sizeof(RREQHeader));
     size += sizeof(RREQHeader);
    
     // Copy old entries into new packet
     if(hdr->numEntries > 1) {
	  memcpy(packet+size, pkt, sizeof(HopEntry)*(hdr->numEntries - 1));
	  size += sizeof(HopEntry)*(hdr->numEntries - 1);
     }
     copyStructIntoPkt(entry, packet+size);
     size += sizeof(HopEntry);
     
     BuildAndSendRoutingPacket(serv, RREQ_PKT, packet, size, &serv->sendBroadcastSocket);
    
}

BOOL
ShouldForwardPacket(ServerData *serv, const RREQHeader *hdr, const char *pkt, const struct in_addr *prevHop)
{
     struct in_addr route[MAX_RREQ_TTL];
     float cost = ComputeMetricCost(serv, pkt, hdr->numEntries, hdr->lastHopCost, serv->helloData->fixedChannel, route);
     ForwardedRREQ *list = serv->routingData->forwardedList, *prev = NULL;

     // See if the RREQ has already been forwarded before
     while( list != NULL) {

	  if( !memcmp(&list->source, &hdr->originator, sizeof(struct in_addr)) && list->seq == hdr->seq ) {
	       // We forward the RREQ if new metric value is lower than the lowest seen metric value so far
	       // Or if the new metric value is less than lowest metric value times forward factor
	       if(list->metric > cost)
	       {
		       list->metric = cost;
		       return TRUE;
	       } 
	       else if( (list->metric * RREQ_FORWARD_FACTOR) > cost)
	       {
		    // though cost is > earlier seen route metric (RREQ_FORWARD_FACTOR > 1)
		    // we still forward the RREQ, since the optimal path may not follow the greedy property 
		    return TRUE;
	       }
	       return FALSE;
	  }

	  prev = list;
	  list = list->next;
     }
    
     // Insert new entry
     list = (ForwardedRREQ *) malloc( sizeof(ForwardedRREQ));
     list->next = NULL;
     memcpy(&list->source, &hdr->originator, sizeof(struct in_addr));
     memcpy(&list->destination, &hdr->destination, sizeof(struct in_addr));
     list->seq = hdr->seq;
     list->metric = cost;
     list->ticks = FORWARDED_TABLE_TICKS;

     if(prev == NULL)
	  serv->routingData->forwardedList = list;
     else
	  prev->next = list;
     return TRUE;
}




BOOL
NodeAlreadyOnRoute(ServerData *serv, const RREQHeader *hdr, const char *pkt)
{
     HopEntry entry;
     int numHops = hdr->numEntries;

     while(numHops > 0) {
	  copyStructFromPkt(entry, pkt);
	  if( !memcmp(&entry.addr, &serv->myIpAddr, sizeof(struct in_addr)) )
	       return TRUE;
	  pkt += sizeof(entry);
	  numHops --;
     }
     return FALSE;
}

void
ProcessRREQ(ServerData *serv, const struct in_addr *prevHop, const char *pkt, int size)
{
     RREQHeader hdr;

#ifdef DEBUG_ROUTE       
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];    
     err_msg("ROUTEDISC:debug_msg:(%s):Metric: Received a RREQ with previous hop %s\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, prevHop, remoteAddr, sizeof(remoteAddr))  
	  );
#endif

     copyStructFromPkt(hdr, pkt);
     pkt += sizeof(hdr);

     // Source node
     if( !memcmp(&hdr.originator, &serv->myIpAddr, sizeof(struct in_addr)) )
	  return; 

     // Update last hop metric cost
     hdr.lastHopCost = GetLastHopCost(serv, prevHop, &hdr);
    
#ifdef DEBUG_ROUTE 
     err_msg("ROUTEDISC:debug_msg:Metric: lastHopCost from %s is %f\n", print_ip(prevHop->s_addr), hdr.lastHopCost);
#endif
     
     // Destination node
     if( !memcmp(&hdr.destination,  &serv->myIpAddr, sizeof(struct in_addr)) )
	  return RREQReceivedAtDest(serv, &hdr, pkt);
    
#ifdef DEBUG_ROUTE    
     err_msg("ROUTEDISC: debug_msg: (%s): Intermediate node RREQ with destination %s\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, &hdr.destination, remoteAddr, sizeof(remoteAddr))  
	  );
#endif
  
     // Intermediate node, but ttl will expire
     if(hdr.ttl <= 1)
	  return;

     // Intermediate node, but already on the route
     if( NodeAlreadyOnRoute(serv, &hdr, pkt) )
	  return;


     if(ShouldForwardPacket(serv, &hdr,pkt,prevHop))
     {

#ifdef DEBUG_ROUTE
	  err_msg("ROUTEDISC:mcr_route.c:debug_msg:Metric: broadcasting RREQ \n");  
#endif
	       // Intermediate node, and should forward packet to everyone
	  ForwardRREQ(serv, &hdr, pkt);
	  
     }
#ifdef DEBUG_ROUTE
     else
     {
	       err_msg("ROUTEDISC:mcr_route.c:debug_msg:Metric: Not forwarding of RREQ for %s\n", print_ip(hdr.destination.s_addr));  
     }
#endif
     
}

/****************** Timer Management ********************/

// Set any type of routing protocol timeouts
static void
SetTimeout(ServerData *serv, struct timeval *delay, ROUTE_TIMEOUT_TYPE type, int data)
{
     AddTimeout(serv, delay, ROUTING_TIMEOUT, type, data);
}


// This timeout is used to update all data structures
// such as routing table, forwardedList and receivedList
void
SetRouteTimeout(ServerData *serv)
{
     RoutingData *data = serv->routingData;
     struct timeval delay;
    
     delay.tv_sec = data->routeTimeoutInterval;
     delay.tv_usec = 0;
     SetTimeout(serv, &delay, TABLE_TIMEOUT, 0);
}


void
RoutingTimerExpired(ServerData *serv, ROUTE_TIMEOUT_TYPE type, int data)
{
     switch(type) {
     case RREQ_TIMEOUT:
	  ReInitiateRREQ(serv, data);
	  break;
     case TABLE_TIMEOUT:
	  UpdateTables(serv);
	  break;
     default:
	  err_msg("Invalid Routing timeout\n");
     }
}



/******************* Debugging command calls **************/
// This allows testing route discovery and route error code without the kernel module

void
ProcessCMD(ServerData *serv, const struct in_addr *prevHop, const char *pkt, int size)
{
     CommandHeader hdr;
     char localAddr[INET_ADDRSTRLEN];
     copyStructFromPkt(hdr, pkt);

     switch(hdr.type) {
     case INITIATE_RREQ:
	  if(!memcmp(&hdr.source, &serv->myIpAddr, sizeof(struct in_addr) ))
	       InitiateRREQ(serv, &hdr.destination, FALSE);
	  break;
     case INITIATE_RERR:
	  if(!memcmp(&hdr.source, &serv->myIpAddr, sizeof(struct in_addr) ))
	       NeighborNotReachable(serv, &hdr.destination);
	  break;
     default:
	  err_msg("(%s): Invalid Debug Command packet\n",
		  Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)));
     }
}


/**************** External hooks to identify route breakages ***************/

void 
RouteBrokenForDestination(ServerData *serv, struct in_addr *addr)
{
     // This is a placeholder for now. This can be used as a hook from
     // kernel multichannel routing module when an icmp destination not reachable
     // message is intercepted

}

// called from interface management protocol
void 
NeighborNotReachable(ServerData *serv, struct in_addr *addr)
{
     /* The strategy is to remove all those routes which
      * were using the missing neighbor as a downstream node,
      * i.e., all routes for which neighbor is the nextHop
      * (including nodes for which neighbor is the destination)
      * For each route that is removed, and RERR is sent to 
      * upstream nodes of the route, if the route was a "FORWARD"
      * route
      */
     RouteTableEntry *list = ExtractAffectedRoutes(serv, addr);
     RouteTableEntry *deleteEntry;

#ifdef DEBUG_ROUTE       
     char localAddr[INET_ADDRSTRLEN];
     char remoteAddr[INET_ADDRSTRLEN];    
     err_msg("ROUTEERROR:(%s):debug_msg:Neighbor %s not reachable\n",
	     Inet_ntop(AF_INET, &serv->myIpAddr, localAddr, sizeof(localAddr)),
	     Inet_ntop(AF_INET, addr, remoteAddr, sizeof(remoteAddr))  
	  );
#endif

     while(list != NULL) {
	  // I am not the source && entry is a forward entry

	  //RERR is sent to affected source when next hop of a route failed
	  //Now we can have multiple sources with respect to a destination. This is yet to be properly handeled.
#ifdef DEBUG_MSG_ON
   	  err_msg("mcr_route.c:debug_msg:Affected Node:%s\n", print_ip(list->source.s_addr));
#endif
	  if( memcmp( &list->source, &serv->myIpAddr, sizeof(struct in_addr)) && list->type == FORWARD)
	       SendRERR(serv, list, FALSE);

	  deleteEntry = list;
	  list = list->next;
	  if(deleteEntry->route)
	       free(deleteEntry->route);
	  DeleteRemoteEntry(serv, &deleteEntry->destination, &deleteEntry->nextHop);
	  Ioctl_DeleteRoute(serv, &deleteEntry->destination, &deleteEntry->nextHop);
	  free(deleteEntry);
     }

}

/****************** Receive messages from kernel multichannel module ************/

void 
ReadNetlinkSocket(ServerData *serv)
{
     int n;
     int addrlen = sizeof(struct sockaddr_nl);
     char buf[MAX_MESSAGE_LENGTH];
     struct in_addr dest_addr, src_addr;
     int type;
     char localAddr[INET_ADDRSTRLEN];
     const RouteTableEntry* rt_entry;
     RouteTableEntry temp;
     kmcr_rt_msg_t *m;
     n = recvfrom(serv->netlinkfd, buf, MAX_MESSAGE_LENGTH, 0, (struct sockaddr *) &serv->netlink_peer, &addrlen);  
     if(!n)
	  return;

     type = ((struct nlmsghdr *) buf)->nlmsg_type;
    
     switch (type) 
     {
	  case KMCR_ROUTEDISC:
	       m = NLMSG_DATA((struct nlmsghdr *) (buf));
	       dest_addr.s_addr = m->dest;
#ifdef DEBUG_MSG_ON	
	  err_msg("NETLINK: debug_msg: Got route discovery msg from kernel for %s\n",
		Inet_ntop(AF_INET, &dest_addr, localAddr, sizeof(localAddr)));
#endif

	       // We already may have found a route - check for it, otherwise
	       // initiate RREQ
	       if(GetRouteTableEntry(serv, &dest_addr) == NULL)
		    InitiateRREQ(serv, &dest_addr, FALSE);
	       break;
	  case KMCR_ROUTEVALID:
	       m = NLMSG_DATA((struct nlmsghdr *) (buf));
	       dest_addr.s_addr = m->dest;
	  
#ifdef DEBUG_MSG_ON	
	       err_msg("NETLINK: debug_msg: Got route valid msg from kernel for %s, timeleft = %d\n",
	       Inet_ntop(AF_INET, &dest_addr, localAddr, sizeof(localAddr)), m->route_expire);
#endif

	       UpdateExpireTimeInRoutingTable(serv, &dest_addr, m->route_expire);
	       break;
	  case KMCR_FWD_ERR:
	       // Received a unicast packet which can not be forwarded
	       // This must be due to stale route at the sender.
	       // Sending a FWD_RERR message to the source of the packet

#ifdef DEBUG_MSG_ON
	       err_msg("NETLINK:debug_msg: Got KMCR_FWD_ERR for dest:\n", print_ip(dest_addr.s_addr));
#endif
	  
	       m = NLMSG_DATA((struct nlmsghdr *) (buf));
	       dest_addr.s_addr = m->dest;
	       src_addr.s_addr = m->src;
	       temp.destination = dest_addr;
	       temp.source = src_addr;
	       //Check if the src of the packet is reachable
	       rt_entry = GetRouteTableEntry(serv, &src_addr);
	       if(rt_entry == NULL)
	       {
		    // We do not have a path to the sender of the packet
		    // Broadcast the RERR ?
		    SendRERR(serv, &temp, TRUE);
	  
	       }	  
	       else
	       {
		    // unicast the RERR to next hop to reach src
		    // Created a temp route entry as if we were an intermediate node for a 
		    // route: src --->prevHop -> this node ---> dest and this route failed, 
		    // thus sending RERR to prevHop of this imaginary(temp) route.
		    temp.prevHop = rt_entry->nextHop;
		    SendRERR(serv, &temp, FALSE);      //Sends RERR to prevHop of the RouteEntry temp
	       
	       }
	       break;
	  default:
	       break;
     }
}

/********************* Handle routing packets **********************************/

void 
ProcessRoutingPacket(ServerData *serv, const McrPacketHeader *hdr, const char *dataPkt, int dataSize)
{
     RoutingProtocolHeader routingHeader;
    
     copyStructFromPkt(routingHeader, dataPkt);
     dataPkt += sizeof(routingHeader);
     dataSize -= sizeof(routingHeader);

     switch(routingHeader.type) {
	  case RREQ_PKT:
	       if(!IsOneHopNeighbor(serv, &hdr->fromAddr))
	       {
#ifdef DEBUG_MSG_ON
		    err_msg("mcr_route.c:debug_msg:Received RREQ from %s not neighbor, dropping\n",
			 print_ip(hdr->fromAddr.s_addr));
#endif
	       }
	       else
	       {
		    ProcessRREQ(serv, &hdr->fromAddr, dataPkt, dataSize);
	       }
	       break;
	  case RREP_PKT:
	       if(!IsOneHopNeighbor(serv, &hdr->fromAddr))
	       {

#ifdef DEBUG_MSG_ON
		    err_msg("mcr_route.c:debug_msg:Received RREP from %s not neighbor, dropping\n",
			 print_ip(hdr->fromAddr.s_addr));
#endif

	       }
	       else
	       {
		    ProcessRREP(serv, &hdr->fromAddr, dataPkt, dataSize);
	       }
	       break;	
	  case RERR_PKT:
	       if(!IsOneHopNeighbor(serv, &hdr->fromAddr))
	       {

#ifdef DEBUG_MSG_ON
		    err_msg("mcr_route.c:debug_msg:Received RERR from %s not neighbor, dropping\n",
			 print_ip(hdr->fromAddr.s_addr));
#endif
	       }
	       else
	       {	  
		    ProcessRERR(serv, &hdr->fromAddr, dataPkt, dataSize);
	       }
	       break;
	  case CMD_PKT:
	       ProcessCMD(serv, &hdr->fromAddr, dataPkt, dataSize);
	       break;
     }
}



/********************************* Initialization ************** ********************/

void
RoutingInit(ServerData *serv)
{
     RoutingData *data;
     data = (RoutingData *) malloc( sizeof(RoutingData));
    
     if(data == NULL)
	  err_quit("Could not allocate memory for the routingData\n");


// Configure netlink socket for communication with kernel    
     serv->netlinkfd = Socket(PF_NETLINK, SOCK_RAW, NETLINK_KMCR);
     memset(&serv->netlink_local, 0, sizeof(struct sockaddr_nl));
     serv->netlink_local.nl_family = AF_NETLINK;
     serv->netlink_local.nl_pid = getpid();
     serv->netlink_local.nl_groups = MCR_GROUP_ID;

     memset(&serv->netlink_peer, 0, sizeof(struct sockaddr_nl));
     serv->netlink_peer.nl_family = AF_NETLINK;
     serv->netlink_peer.nl_pid = 0;
     serv->netlink_peer.nl_groups = 0;

     Bind(serv->netlinkfd, (struct sockaddr *) &serv->netlink_local, sizeof(struct sockaddr_nl));

     data->table.routeList = NULL;
     data->pendingList = NULL;
     data->forwardedList = NULL;
     data->forwardedRREPList = NULL;
     data->repliedList = NULL;
     data->seq = 1;
     data->routeTimeoutInterval = ROUTE_TIMEOUT_INTERVAL;
     data->routeMetricType = MCR_SHORTEST_HOP_AND_DIVERSITY;
     data->beta = DEFAULT_BETA;
     data->alpha = DEFAULT_ALPHA ;
     serv->routingData = data;
    
}


