/*****************************************************************************
 * mcr_route.h: Multichannel routing protocol
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 * Modifications by: Rishi Bhardwaj, <rbhardw2 at uiuc dot edu>
 *
 *****************************************************************************/

/* (Aug 1, 2007) Changes by Rishi Bhardwaj:
 * 1) forwarded_rrep table maintained at all the nodes to keep tab of route discovered
 *    from a source to destination and the latest RREP seq. no. seen from a destination.
 *    This helps to suppress a RREP from destination to source if already a more recent, thus better,
 *    RREP has been forwarded to the source.
 * 2) We use replied_rreq table at the destination to do book keeping of the last route discovered from a source.
 *    When a RREQ or RFRESH message is received from a source, the newly discovered route is compared to the
 *    old route. If the new route discovered is the same as the route in replied_rreq table, 
 *    we update the routing metric associated with the route. Otherwise if the routing metric for the new route 
 *    is better than the last route, the destination choses the new route and sends RREP to the source.
 * 3) Introduce a new routing metric: MCR_CHAN_REUSE_CETT and its associated routing parameters. 
 *    A Modification of MCR metric.
 * 4) Changed the type of all sequence numbers to be unsigned int. 
 */


#ifndef _MCR_ROUTE_H
#define _MCR_ROUTE_H

/*********************************** Default contants ******************************************/

#define MAX_RREQ_TTL   7                             // for route discovery messages

#define ROUTE_DISCOVERY_TIMEOUT 1    // in seconds

#define MAX_DISCOVERY_ATTEMPTS  3     // max RREQ attempts

#define  FORWARDED_TABLE_TICKS  5       // Number of seconds to keep forwarded RREQ info


#define FORWARDED_RREP_TABLE_TICKS  5   // Number of seconds to keep info for forwarded RREP   

#define  ROUTE_ENTRY_TICKS    10             // Number of seconds of inactivity before deleting a route

//Note: To avoid infinite route life ROUTE_REFRESH_TICKS should be larger than ROUTE_ENTRY_TICKS   
// Otherwise, routes to inactive destination get refreshed (and inactivity value reset) before it can be expired
#define ROUTE_REFRESH_TICKS    30         // Periodicity of route refresh, in seconds. 

/* REPLIED_TABLE_TICKS is the number of seconds to keep RREP info. It must be > ROUTE_REFRESH_TICKS.
 * Receiving of RRFRESH packet refreshes RREP info, RRFRESH packet is received every ROUTE_REFRESH_TICKS.
 */
#define  REPLIED_TABLE_TICKS  (ROUTE_REFRESH_TICKS + 5) 

#define RREQ_FORWARD_FACTOR   1.2      // How much lower the new RREQ could be and still be forwarded

#define REFRESH_METRIC_FACTOR 0.8      // How much better should a refreshed be for destination to reply with RREP 

#define KERNEL_ROUTE_TIMEOUT  10        // in seconds. Set this to the same as ROUTE_ENTRY_TICKS * ROUTE_TIMEOUT_INTERVAL

#define UPDATE_CURR_ROUTE_METRIC_FACTOR  0.8  // How much worse does the metric for current route become before we update it

#define ROUTE_TIMEOUT_INTERVAL 1    // This is number of ticks per second. The code does not handle this well, always set it to 1

#define TOTAL_CHANNEL 15            // this is also defined in mcr_hello.h - keep the definitions consistent. TODO: Use a single constant


/* Parameters for different routing metrics */

#define MCR_DIVERSITY_COUNT_WINDOW 3

#define PACKET_TRANSMISSION_TIME 2000   // time in microseconds, currently set for 1500 byte packet with 6Mbps data rate 

#define DEFAULT_BETA    0.5

#define DEFAULT_ALPHA   0.5

#define CHAN_REUSE_HOPS 5    // Number of hops after which a channel may be repeated in a route, used in MCR_CHAN_REUSE_CETT routing metric

#define THRSHHOLD_TX_PROB 0.75
#define CONSECTV_CHANNEL_THRSHHOLD_TX_PROB 0.9

/* Parameters done */


#ifdef DEBUG_ROUTE_DELAY

//Constants used for experiment
#define MAX_DEBUG_STR    30
#define MAX_DEBUG_STR_SIZE 200
//experiment constants done

#endif


#ifdef ROUTE_EXPERIMENT

#define MAX_ROUTE_DEBUG_MESSAGE_LENGTH 1000 

#endif



// Note: in the latest code it is required that ROUTE_TIMEOUT_INTERVAL is equal to 1 (once per second), and ROUTE_ENTRY_TICKS == KERNEL_ROUTE_TIMEOUT - this is to ensure route timeout interaction between kernel and user space works correctly


/******************* Types of metrics **********************/

// Not everything is implemented. The list is taken from metrics that were simulated
typedef enum 
{
    MCR_SHORTEST_HOP = 1,
    MCR_SHORTEST_HOP_AND_DIVERSITY = 2,
    MCR_SHORTEST_HOP_AND_MAX_DIVERSITY = 3,
    MCR_DIVERSITY_AND_SWITCHING = 4,
    MCR_MAX_DIVERSITY_AND_SWITCHING = 5,
    MCR_MCETT = 6,   // The new MCETT metric
    MCR_WCETT = 7,   // modified WCETT metric, based on work by Draves et al. in Mobicom 2005
    MCR_CHAN_REUSE_CETT = 8

} McrMetricType;

/*********** Timer types ************/

typedef enum {
    RREQ_TIMEOUT,
    TABLE_TIMEOUT
} ROUTE_TIMEOUT_TYPE;


/***********  Packet formats  **************/

typedef enum {
    RREQ_PKT,
    RREP_PKT,
    RERR_PKT,
    CMD_PKT  // Special packet for debugging
} ROUTE_PACKET_TYPE;

// support for debugging
typedef enum {
    INITIATE_RREQ,
    INITIATE_RERR
} DEBUG_COMMAND_TYPE;

typedef struct struct_command_header {
    DEBUG_COMMAND_TYPE type;
    struct in_addr source;
    struct in_addr destination;
} CommandHeader;


/************* Route packet structures ***************/

typedef struct struct_routing_protocol_header {
    ROUTE_PACKET_TYPE type;
} RoutingProtocolHeader;

typedef struct channel_entry {
    int num;
    float cost;
} ChannelEntry;

typedef struct struct_rreq_header {
    struct in_addr originator;
    struct in_addr destination;
    int numEntries;
    int ttl;
    unsigned int seq;
    float lastHopCost;
    int numChannels;
    ChannelEntry cList[TOTAL_CHANNEL];
    BOOL isRefresh;
} RREQHeader;
 

typedef struct struct_rrep_header {
    struct in_addr originator;
    struct in_addr destination;
    int numEntries;
    unsigned int seq;
    unsigned int dstSeq;
    float metric;
} RREPHeader;

typedef struct struct_rerr_header {
    struct in_addr originator;
    struct in_addr destination;
} RERRHeader;

typedef struct struct_hop_entry {
    struct in_addr addr;
    int channel;
    float metric;
} HopEntry;

/*****************   Routing table  structures **********************/

typedef enum {
    SRC, // explicitly discovered with a RREQ request
    FORWARD, // overheard route in FORWARD direction
    OVERHEARD // route is noted down from RREP
} ROUTE_TYPE;

typedef struct struct_routing_table_entry {
    struct in_addr destination;
    struct in_addr source;
    struct in_addr nextHop;
    struct in_addr prevHop;
    
    float metric;
    char *route;
    int ticks; // Used to remove entry
    int refreshTicks;
    ROUTE_TYPE type;
    struct struct_routing_table_entry * next;
} RouteTableEntry;


// We put this in a separate structure
// with a single entry as eventually more
// data may get into this, and the table
// could be more efficiently stored as
// a hash table, etc. if required

typedef struct struct_routing_table {
    RouteTableEntry *routeList;
} RoutingTable;

/*********  Auxillary tables used in route discovery *************/

typedef struct struct_pending_rreq {
    struct in_addr destination;
    unsigned int seq;
    int numAttempts;
    BOOL isRefresh;
#ifdef DEBUG_ROUTE_DELAY
    struct timeval tv;
#endif
    struct struct_pending_rreq *next;
} PendingRREQ;

typedef struct  struct_forwarded_rreq {
    struct in_addr source;
    struct in_addr destination;
    unsigned int seq;
    float metric;
    int ticks; // Used to clean up data structure
    struct struct_forwarded_rreq *next;
    
} ForwardedRREQ;


typedef struct  struct_replied_rreq {
    struct in_addr source;
    unsigned int seq;
    float metric;
    struct in_addr route[MAX_RREQ_TTL + 1];  // To remember the entire route
    unsigned short hops;
    int ticks; // Used to clean up data structure
    struct struct_replied_rreq *next;    
} RepliedRREQ;


/* We use this structure to keep tab of the
 * RREP received from a destination. There may be
 * multiple RREP sent from a destination for the 
 * the same sequence numbered RREQ from source.
 * We must not process the RREP which is older
 * than the the RREP last received from the destination.
 */
typedef struct struct_forwarded_rrep {
     struct in_addr source;
     struct in_addr destination;
     unsigned int srcSeq;
     unsigned int dstSeq;
     float metric;
     int ticks; // Used to clean up data structure
     struct struct_forwarded_rrep *next;
} ForwardedRREP;

/****************  Main routing protocol data structure  ****************/

typedef struct struct_routing_data {
    RoutingTable table;
    PendingRREQ *pendingList;
    ForwardedRREQ *forwardedList;
    RepliedRREQ *repliedList;
    ForwardedRREP *forwardedRREPList;
    unsigned int seq;
    int routeTimeoutInterval;
    McrMetricType routeMetricType;

    // Metric paramters
    float beta;
    float alpha;
} RoutingData;


/***********  Function declarations for external usage  *****************/

void RoutingInit(ServerData *serv);

void InitiateRREQ(ServerData *serv, struct in_addr *addr, BOOL isRefresh);

void RouteBrokenForDestination(ServerData *serv, struct in_addr *addr);

void NeighborNotReachable(ServerData *serv, struct in_addr *addr);

void ProcessRoutingPacket(ServerData *serv, const McrPacketHeader *hdr, const char *dataPkt, int dataSize);

void SetRouteTimeout(ServerData *serv);

void RoutingTimerExpired(ServerData *serv, ROUTE_TIMEOUT_TYPE type, int data);

void ReadNetlinkSocket(ServerData *serv);

BOOL areConsecutiveChannels(int chan1, int chan2);
#endif /* _MCR_ROUTE_H */
