/*****************************************************************************
 * timer.c: Support for timers
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 *
 *****************************************************************************/

/*************************** NOTES *******************************************

 This file has code for adding/deleting timers and managing timeouts

*********************************************************************************/

#include "utilities.h"
#include "integrated.h"
#include "timer.h"
#include "mcr_hello.h"
#include "mcr_route.h"
#include "mcr_gateway.h"

void 
AddTimeout(ServerData *serv, struct timeval *delay, TIMER_TYPE type, int subType, int timerData)
{
    Timer *newTimer, *cur = serv->timerQueue, *prev = NULL;
    struct timeval curTime;

    newTimer = (Timer *) malloc(sizeof(Timer));
    if(newTimer == NULL) 
	err_quit("Could not allocate memory for a timer\n");

    Gettimeofday(&curTime, 0);
    newTimer->time.tv_usec = curTime.tv_usec + delay->tv_usec;
    newTimer->time.tv_sec = curTime.tv_sec + delay->tv_sec;
    newTimer->type = type;
    newTimer->subType = subType;
    newTimer->timerData = timerData;
    newTimer->next = NULL;
    // Insert into queue at the appropriate place
    
    if(cur == NULL) {
	serv->timerQueue = newTimer;
	return;
    }
	
    while(cur != NULL && timercmp(&cur->time, &newTimer->time, <=) ) {
	prev = cur;
	cur = cur->next;
    }
    if(prev != NULL)
	prev->next = newTimer;
    else
	serv->timerQueue = newTimer;
    newTimer->next = cur;
}


// Extracts out the expired timeouts and updates
// servers timeout list
Timer *
ExtractExpiredTimeouts(ServerData *serv)
{
    Timer *cur = serv->timerQueue, *prev = NULL;
    struct timeval curTime;
    
    if(cur == NULL)
	return NULL;
    
    Gettimeofday(&curTime, 0);
    if((curTime.tv_usec + TIMER_GRANULARITY) > MILLION) {
	++curTime.tv_sec;
	curTime.tv_usec += TIMER_GRANULARITY - MILLION;
    }
    else
	curTime.tv_usec += TIMER_GRANULARITY;
    
    // Add some jitter to current time to account for timer
    // variability

    // Extract the list of timers that have expired & update timerQueue
    // timercmp is a predefined macro
    while(cur != NULL && timercmp(&cur->time, &curTime, <=) ) {
	prev = cur;
	cur = cur->next;
    }
    // nothing matched
    if(prev == NULL) 
	return NULL;

    prev->next = NULL;
    prev = serv->timerQueue;
    serv->timerQueue = cur;
    cur = prev;
    return cur;
}


// This code is not very modular. The best way would have been to register
// function pointers for different timeouts and invoke them. Its a TODO item to fix this
void
InvokeTimeouts(ServerData *serv, Timer *cur)
{
    Timer *prev;
     while( cur != NULL) {
	switch(cur->type) {
	case HELLO_TIMEOUT:
	    HelloTimeout(serv);
	    break;

	case ROUTING_TIMEOUT:
	    RoutingTimerExpired(serv, (ROUTE_TIMEOUT_TYPE)cur->subType, cur->timerData);
	    break;
	    
	case STATUS_TIMEOUT:
	    StatusTimeout(serv);
	    break;

	case GATEWAY_TIMEOUT:
	    GatewayTimeout(serv);
	    break;
	    
	case SCAN_TIMEOUT:
	    ScanTimeout(serv);
	    break;
	}
	prev = cur;
	cur = cur->next;
	free(prev);
    } // end of while

}

struct timeval * 
GetNextTimeout(ServerData *serv)
{
    static struct timeval timeout;
    struct timeval curTime;
    Timer *expiredTimers;
    if(serv->timerQueue == NULL)
	return NULL;

    // Run all expired timeouts
    while( (expiredTimers = ExtractExpiredTimeouts(serv)) != NULL) 
	InvokeTimeouts(serv, expiredTimers);
    

    Gettimeofday(&curTime, 0);
    if(timercmp(&serv->timerQueue->time, &curTime, <=) ) 
    {
	err_msg("PROBLEM: We are disabling timers at this point\n");
	return NULL;
    }

    Gettimediff(&timeout, &serv->timerQueue->time, &curTime);
//    err_msg("Returned non-zero timeout value (%d:%d),(%d:%d)\n",
//	curTime.tv_sec, curTime.tv_usec,timeout.tv_sec,timeout.tv_usec);
    return &timeout;
}

void 
ExecuteTimeout(ServerData *serv)
{
 // We have this function solely to first extract out timeouts 
     // and then call timeouts; This is because some of the timeout 
    // functions may in turn add a new timer to the timerQueue.
    // To avoid race conditions, we ensure that we have finished
    // accessing timerQueue before any of the timeout functions are
    // called

     Timer *expiredTimers =  ExtractExpiredTimeouts(serv);
     InvokeTimeouts( serv, expiredTimers);
}
