/*****************************************************************************
 * utilities.c: General purpose networking utilities
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 * Modifications by: Rishi Bhardwaj, <rbhardw2 at uiuc dot edu>
 *
 *****************************************************************************/

/*************************** NOTES *******************************************

 This file contains wrapper and utility functions for socket programming.
 The code here is based on the code in the book
  Unix Network Programming: Volume 1, 2nd edition by Richard Stevens

*********************************************************************************/

#include "utilities.h"
extern int daemon_proc;

extern FILE *errMsgFile;
extern BOOL suppress_debug_msg;


/******************* Utilities for handling time functions *********************/

// compute t1 = t2 - t3, with assumption t2 > t3
void
Gettimediff(struct timeval *t1, struct timeval *t2, struct timeval *t3) {

     if(t2->tv_usec < t3->tv_usec) {
	  if(t3->tv_sec >= t2->tv_sec) {
	       t1->tv_sec = 0;
	       t1->tv_usec = 0;
	       return;
	  }
	  t1->tv_usec = 1000000 + t2->tv_usec -  t3->tv_usec;
	  t1->tv_sec = t2->tv_sec - t3->tv_sec - 1;
     } else {
	  if(t3->tv_sec > t2->tv_sec) {
	       t1->tv_sec = 0;
	       t1->tv_usec = 0;
	       return;
	  }
	  t1->tv_usec = t2->tv_usec -  t3->tv_usec;
	  t1->tv_sec = t2->tv_sec - t3->tv_sec;
     }
}

void
Gettimeofday(struct timeval *tv, struct timezone *tz)
{
     if(gettimeofday(tv, tz) < 0)
	  err_sys("Gettimeofday error\n");
}

/*************************** Socket wrappers *************************************/

void
Bind(int fd, const struct sockaddr *sa, socklen_t salen)
{
     if (bind(fd, sa, salen) < 0)
	  err_sys("bind error");
}

void
Sendto(int fd, const void *ptr, size_t nbytes, int flags,
       const struct sockaddr *sa, socklen_t salen)
{
     if (sendto(fd, ptr, nbytes, flags, sa, salen) != nbytes)
	  err_sys("sendto error");
}

ssize_t
Recvfrom(int fd, void *ptr, size_t nbytes, int flags,
	 struct sockaddr *sa, socklen_t *salenptr)
{
     ssize_t		n;

     if ( (n = recvfrom(fd, ptr, nbytes, flags, sa, salenptr)) < 0)
	  err_sys("recvfrom error");
     return(n);
}

void
Setsockopt(int fd, int level, int optname, const void *optval, socklen_t optlen)
{
     if (setsockopt(fd, level, optname, optval, optlen) < 0)
	  err_sys("setsockopt error");
}

int
Socket(int family, int type, int protocol)
{
     int n;

     if ( (n = socket(family, type, protocol)) < 0)
	  err_sys("socket error");
     return(n);
}

/**************************** Address translation **********************************/

const char *
Inet_ntop(int family, const void *addrptr, char *strptr, size_t len)
{
     const char	*ptr;

     if (strptr == NULL)		/* check for old code */
	  err_quit("NULL 3rd argument to inet_ntop");
     if ( (ptr = inet_ntop(family, addrptr, strptr, len)) == NULL)
	  err_sys("inet_ntop error");		/* sets errno */
     return(ptr);
}

void
Inet_pton(int family, const char *strptr, void *addrptr)
{
     int		n;

     if ( (n = inet_pton(family, strptr, addrptr)) < 0)
	  err_sys("inet_pton error for %s", strptr);	/* errno set */
     else if (n == 0)
	  err_quit("inet_pton error for %s", strptr);	/* errno not set */

     /* nothing to return */
}


/******************************* Error handling and logging  *********************************/

static void
err_doit(int errnoflag, int level, const char *fmt, va_list ap, FILE *file)
{
     int	errno_save, n;
     char	buf[MAXLINE];

     errno_save = errno;		/* value caller might want printed */

     vsnprintf(buf, sizeof(buf), fmt, ap);	/* this is safe */
     n = strlen(buf);
     if (errnoflag)
	  snprintf(buf+n, sizeof(buf)-n, ": %s", strerror(errno_save));
     strcat(buf, "\n");


     
     if (daemon_proc) {
   	  if(file == NULL)
	  {
	       syslog(level|LOG_LOCAL2, buf);
	  }
	  else
	  {
	       fflush(file);
	       fputs(buf, file);
	       fflush(file);
	  }
     } else {
	  fflush(stdout);		/* in case stdout and stderr are the same */
	  fputs(buf, stderr);
	  fflush(stderr);
     }
     return;
}


void
err_msg(const char *fmt, ...)
{
    
     if (suppress_debug_msg)
     	return;
     
     va_list ap;

     va_start(ap, fmt);
     err_doit(0, LOG_INFO, fmt, ap, errMsgFile);
     va_end(ap);
     
     return;
}

#ifdef ROUTE_EXPERIMENT
void routeDebugMsg(FILE* file, const char *fmt, ...)
{
     va_list		ap;
     va_start(ap, fmt);
     err_doit(0, LOG_INFO, fmt, ap, file);
     va_end(ap);
     return;
}

#endif



void
err_sys(const char *fmt, ...)
{
     va_list	ap;
     va_start(ap, fmt);
     err_doit(1, LOG_ERR, fmt, ap, errMsgFile);
     va_end(ap);
     exit(1);
}

void
err_quit(const char *fmt, ...)
{
     va_list	ap;

     va_start(ap, fmt);
     err_doit(0, LOG_ERR, fmt, ap, errMsgFile);
     va_end(ap);
     exit(1);
}

