/*****************************************************************************
 * kmcr_main.c: The main file of the userspace daemon implementation. 
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 * Modifications by: Rishi Bhardwaj, < rbhardw2 at uiuc dot edu >
 *
 *
 *****************************************************************************/


/*************************** NOTES *******************************************

This file is the main file of the kernel multichannel routing module.
 Code in this module is based on architecture used by "kaodv" module by 
Erik Nordstrm from Uppsala university.

The code in this file interfaces with kernel netfilter hooks and intercepts
packets. This file also glues together other components of the kernel
module implementation.


Changes by Rishi Bhardwaj: 
Added support for detection of route failure (packet forwarding error).
When the Kmcr module receives a packet on the PF_IP_PRE_ROUTING hook to be
forwarded to some other destination, it checks if a route to the destination
exists in the routing table. If a route to the destination does not exist,
the KMCR module sends a netlink message to user space with information about the 
source and destination of the packet being dropped. The user space is expected 
to handle the situation appropriately.

*********************************************************************************/

#include <asm/byteorder.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/in.h>
#include <linux/icmp.h>
#include <net/route.h>

#include "kmcr_netlink.h"
#include "kmcr_table.h"
#include "kmcr_queue.h"
#include "kmcr_ipenc.h"

MODULE_DESCRIPTION("Kernel support for multichannel routing");
MODULE_AUTHOR("Pradeep Kyasanur, Rishi Bhardwaj");


#define MAX_INTERFACES 5  // maximum number of real interfaces that
                          // are used only for sending data

static char *interfaceName = "bond0"; // this is the interface exposed by channel abstraction layer

static char *sendInterfaceName[MAX_INTERFACES] = { "ath1" }; // real name of the switchable interface

MODULE_PARM( interfaceName, "s");
MODULE_PARM_DESC( interfaceName, "Exposed interface used by mesh subnet");

MODULE_PARM(sendInterfaceName, "1-" __MODULE_STRING(MAX_INTERFACES) "s");
MODULE_PARM_DESC( sendInterfaceName, "Real interfaces used only to send data");

// Local data structure for storing interface information
struct netdev_info {
     __u32 ip_addr;
     __u32 bc_addr;
     __u32 mask;
     int ifindex;
};

static struct netdev_info netdevs;

// Utility functions to display ip address in dotted decimal format
static char inline *print_ip(__u32 addr)
{
     static char buf[16];

     sprintf(buf, "%d.%d.%d.%d",
	     0x0ff & addr,
	     0x0ff & (addr >> 8),
	     0x0ff & (addr >> 16), 0x0ff & (addr >> 24));

     return buf;
}

static char inline *print_ip_new(__u32 addr, char* buf)
{
	 sprintf(buf, "%d.%d.%d.%d",
	     0x0ff & addr,
	     0x0ff & (addr >> 8),
	     0x0ff & (addr >> 16), 0x0ff & (addr >> 24));
	return buf;
}

/*************************** Netfilter hook ************************************/

static unsigned int
kmcr_hook(unsigned int hooknum,
	  struct sk_buff **skb,
	  const struct net_device *in,
	  const struct net_device *out, int (*okfn) (struct sk_buff *))
{
     struct iphdr *iph = (*skb)->nh.iph;
     struct route_entry e;
     struct rtable *rt;
     __u32 clientAddr = htonl(0xc0a81f01); // 192.168.31.1
     __u32 clientNetmask = htonl(0xffffff00); // 255.255.255.0

// Step 1: Ignore packets on any interface except the bonded interface, and ignore non-IP packets

     if(in != NULL)
     {
	  if(strcmp(in->name, interfaceName)) {
	       return NF_ACCEPT;
	  }
     } 
     if(out != NULL)
     {
	  if(strcmp(out->name, interfaceName)) {
	       return NF_ACCEPT;
	  }
     }


     if(iph == NULL)
	  return NF_ACCEPT;

    
     // Step 2: Filter out packets coming in on switchable "send" interfaces in PRE_ROUTING hook
     if(hooknum == NF_IP_PRE_ROUTING && (*skb)->real_dev != NULL)
     {
	  int i;
	  for(i = 0; i < MAX_INTERFACES; i++)
	  {
	       // Drop packets on interface matching name
	       if(sendInterfaceName[i] && 
		  !strcmp((*skb)->real_dev->name, sendInterfaceName[i]) )
	       {
		    return NF_DROP;
	       }
	  }
     }

     // Step 3: Ignore broadcast and multicast packets 
     if (iph->daddr == INADDR_BROADCAST ||
	 IN_MULTICAST(ntohl(iph->daddr)) || iph->daddr == netdevs.bc_addr)
     {
	  return NF_ACCEPT;
     }

     // Step 4:  Ignore packets coming from the client address space - to allow unmodified clients to 
     //                interoperate with our implementation
     if( (iph->daddr & clientNetmask) == (clientAddr & clientNetmask))
     {
	  return NF_ACCEPT;
     }
    
     // Step 5: We have reached packets we care about, so process them
     switch(hooknum) {
     case NF_IP_PRE_ROUTING: 

	  // We update reverse route in PRE_ROUTING hook
	  kmcr_table_access_update(iph->saddr);  
          
	  // We also see if the dest. addr is in routing table or not
	  // if no route to dest exists we send a forwarding route error to the src 
	  if(netdevs.ip_addr != iph->daddr)
	  {
	       if(kmcr_table_access_update(iph->daddr) != 0)
	       {
		    // packet can not be forwarded, FWD_RERR must be sent to the source of the packet.
		    // Send FWDing error event to the user process, which will decide what action to take.
		    kmcr_netlink_send_rt_msg(KMCR_FWD_ERR,iph->saddr,iph->daddr, 0);

	       }
	       
	  } 
	  else if (iph->protocol == IPPROTO_MIPE){  // && iph->daddr == netdevs.ip_addr) {
	  // Packet has to be decapsulated at gateway

	       ip_pkt_decapsulate(*skb);
	       iph = (*skb)->nh.iph;
	  }
	  break;

     case  NF_IP_LOCAL_OUT:
	
	  // Should we encapsulate the packet?
	  rt = (struct rtable *)(*skb)->dst;
	  if(rt == NULL)
	       printk("KMCR: Surprise... skb->dst is null\n");
	  else
	  {
	       // We encapsulate if dest is in different subnet
	       // AND gateway specified is current gateway
	       if( ((iph->daddr ^ netdevs.ip_addr)& netdevs.mask) &&
		   rt->rt_gateway == kmcr_get_gateway_addr())
	       {
		    if(ip_pkt_encapsulate(skb, rt->rt_gateway) != 0)
			 return NF_STOLEN;
		    ip_route_me_harder(skb);
		    iph = (*skb)->nh.iph;
	       } 
	    
	  }

	  if(!kmcr_table_get(iph->daddr, &e))
	  {
	       // No entry - initiate route discovery
	       kmcr_netlink_send_rt_msg(KMCR_ROUTEDISC,
					iph->saddr,
					iph->daddr, 0);
	       kmcr_table_add(iph->daddr, 0, 0, ROUTE_DISC_PENDING);

	       // buffer packet
	       kmcr_enqueue_packet(iph->daddr, *skb, okfn);
	       return NF_STOLEN;
	  } else
	  {
	       if(e.flags & ROUTE_DISC_PENDING)
	       {
		    // buffer packet
		    kmcr_enqueue_packet(iph->daddr, *skb, okfn);
		    return NF_STOLEN;
	       } else
		    return NF_ACCEPT;
	   
	  }
	  break;

     case  NF_IP_POST_ROUTING:

	  // We update forward route in POST_ROUTING hook
	  if(netdevs.ip_addr == iph->saddr)
	  {
	       // We only upadate the entry for local out packets, 
	       // since for packets being forwarded the entry has already been updated in PRE_ROUTING hook.
	       kmcr_table_access_update(iph->daddr); 
	  }
	  
	  kmcr_table_stats_update(iph->daddr);

	  break;
     default:
	  printk("KMCR: Unexpected packet received in kmcr_hook\n");
     }
     return NF_ACCEPT;
}


/***************** Structures used to register with net filter *******************************/
static struct nf_hook_ops kmcr_ops[] = {
     {
	  .hook = kmcr_hook,
#ifdef KERNEL26
	  .owner = THIS_MODULE,
#endif
	  .pf = PF_INET,
	  .hooknum = NF_IP_PRE_ROUTING,
	  .priority = NF_IP_PRI_FIRST,
     },
     {
	  .hook = kmcr_hook,
#ifdef KERNEL26
	  .owner = THIS_MODULE,
#endif
	  .pf = PF_INET,
	  .hooknum = NF_IP_LOCAL_OUT,
	  .priority = NF_IP_PRI_FILTER,
     },
     {
	  .hook = kmcr_hook,
#ifdef KERNEL26
	  .owner = THIS_MODULE,
#endif
	  .pf = PF_INET,
	  .hooknum = NF_IP_POST_ROUTING,
	  .priority = NF_IP_PRI_FILTER,
     },
};

/**************************** Initialization and cleanup ****************************/

static int __init kmcr_init(void)
{
     struct net_device *dev = NULL;
     struct in_device *indev = NULL;
     struct in_ifaddr **ifap = NULL;
     struct in_ifaddr *ifa = NULL;
     int ret=-1;

     kmcr_table_init();
     kmcr_queue_init();

     ret = kmcr_netlink_init();
    
     if(ret < 0)
	  goto cleanup_initial;
    
     ret = nf_register_hook(&kmcr_ops[0]);

     if (ret < 0)
	  goto cleanup_netlink;
    
     ret = nf_register_hook(&kmcr_ops[1]);
    
     if (ret < 0)
	  goto cleanup_pre_routing;
    
     ret = nf_register_hook(&kmcr_ops[2]);
    
     if (ret < 0)
	  goto cleanup_local_out;


     // Get exposed interface information
     dev = dev_get_by_name(interfaceName);
     if (!dev) {
	  printk("KMCR: No device %s available!\n",
		 interfaceName);
	  ret = -1;
	  goto cleanup_post_routing;
     }
     netdevs.ifindex = dev->ifindex;
    
     indev = in_dev_get(dev);
    
     if (indev) {
	  for (ifap = &indev->ifa_list; (ifa = *ifap) != NULL;
	       ifap = &ifa->ifa_next)
	       if (!strcmp(dev->name, ifa->ifa_label))
		    break;
	
	  if (ifa) {
	       netdevs.ip_addr = ifa->ifa_address;
	       netdevs.bc_addr = ifa->ifa_broadcast;
	       netdevs.mask = ifa->ifa_mask;
	  }
	  in_dev_put(indev);
     }
     dev_put(dev);
    
     printk("KMCR module loaded successfully (%x) (%x) (%x)\n", netdevs.ip_addr, netdevs.bc_addr, netdevs.mask);
     return ret;

cleanup_post_routing:
     nf_unregister_hook(&kmcr_ops[2]);
cleanup_local_out:
     nf_unregister_hook(&kmcr_ops[1]);
cleanup_pre_routing:
     nf_unregister_hook(&kmcr_ops[0]);
cleanup_netlink:
     kmcr_netlink_fini();
cleanup_initial:  
     kmcr_queue_fini();
     kmcr_table_fini();
     printk("KMCR module load failed\n");
     return ret;

}

static void __exit kmcr_exit(void)
{
     nf_unregister_hook(&kmcr_ops[2]);
     nf_unregister_hook(&kmcr_ops[1]);
     nf_unregister_hook(&kmcr_ops[0]);
     kmcr_netlink_fini();
     kmcr_queue_fini();
     kmcr_table_fini();
     printk("KMCR module unloaded\n");

}

module_init(kmcr_init);
module_exit(kmcr_exit);
