/*****************************************************************************
 * kmcr_netlink.c: Support for communication with userspace
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 *
 *****************************************************************************/


/*************************** NOTES *******************************************

* This file provides support for netlink communication with user space

*********************************************************************************/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/netlink.h>
#include <net/sock.h>

#include "kmcr_netlink.h"
#include "kmcr_table.h"
#include "kmcr_queue.h"

static int peer_pid;
static struct sock *kmcrnl;
static DECLARE_MUTEX(kmcrnl_sem);

static __u32 gatewayAddr;




// Utility functions to display ip address in dotted decimal format
static char inline *print_ip(__u32 addr)
{
     static char buf[16];

     sprintf(buf, "%d.%d.%d.%d",
	     0x0ff & addr,
	     0x0ff & (addr >> 8),
	     0x0ff & (addr >> 16), 0x0ff & (addr >> 24));

     return buf;
}



/****************************** Interaction with other modules *********************************/

__u32 
kmcr_get_gateway_addr(void)
{
     return gatewayAddr;
}


/**************************  Steps taken when user space program finishes ********************/

static int
kmcr_netlink_rcv_nl_event(struct notifier_block *this,
			  unsigned long event, void *ptr)
{
     struct netlink_notify *n = ptr;

     if (event == NETLINK_URELEASE && n->protocol == NETLINK_KMCR && n->pid) {
	  if (n->pid == peer_pid) {
	       peer_pid = 0;
			
	       kmcr_table_flush();
	       kmcr_queue_flush();
	       gatewayAddr = 0;
	       printk("KMCR: Peer has left\n");
	       // PEER HAS LEFT - TAKE SUITABLE ACTION

	  } else
	       printk("KMCR: Unknown peer has left\n");
     }
     return NOTIFY_DONE;
}

/******************** Send netlink message  to user space ***********************/

static struct sk_buff *kmcr_netlink_build_msg(int type, void *m, int len)
{
     unsigned char *old_tail;
     size_t size = 0;
     struct sk_buff *skb;
     struct nlmsghdr *nlh;
     struct kmcr_rt_msg *rtmsg;

     size = NLMSG_SPACE(len);

     skb = alloc_skb(size, GFP_ATOMIC);
     if (!skb)
	  return NULL;

     old_tail = skb->tail;
     nlh = NLMSG_PUT(skb, 0, 0, type, size - sizeof(*nlh));

     rtmsg = NLMSG_DATA(nlh);

     memcpy(rtmsg, m, len);

     nlh->nlmsg_len = skb->tail - old_tail;
     return skb;
	
nlmsg_failure:
     if (skb)
	  kfree_skb(skb);
	
     printk(KERN_ERR "KMCR: error creating routing message\n");
     return NULL;
	
}


void kmcr_netlink_send_rt_msg(int type, __u32 src, __u32 dest, unsigned long timeLeft)
{
     struct sk_buff *skb = NULL;
     struct kmcr_rt_msg m;
    
     m.src = src;
     m.dest = dest;
     m.nhop = 0;
     m.route_expire = timeLeft;
     m.flags = 0;
    
     skb = kmcr_netlink_build_msg(type, &m, sizeof(struct kmcr_rt_msg));
    
     if (skb == NULL) {
	  printk("KMCR: kmcr_netlink - skb=NULL\n");
	  return;
     }

     netlink_broadcast(kmcrnl, skb, 0, MCR_GROUP_ID, GFP_USER);
}


/*********************  Receive netlink message from user space ***************************/

static int
kmcr_netlink_receive_peer(unsigned char type, void *msg, unsigned int len)
{
     int status = 0;
     struct kmcr_rt_msg *rtm;
     struct route_entry e;
     unsigned long timeLeft;

     if(len < sizeof(struct kmcr_rt_msg))
     {
	  printk("KMCR: Add route message is too small %d instead of %d\n", len, sizeof(struct kmcr_rt_msg));
	  return -EINVAL;
     }
     rtm = (struct kmcr_rt_msg *)msg;

     switch (type) {
     case KMCR_ADDROUTE:
	  if(rtm->dest == 0) // ignore default routes
	       break;

	  if (kmcr_table_get(rtm->dest, &e))
	  {
	       kmcr_table_update(rtm->dest, rtm->nhop, rtm->route_expire, ROUTE_ACTIVE);

	       // We are being conservative and
	       // trying to send out buffered packets
	       // even if route state was not routes pending
	       kmcr_queue_set_verdict(KMCR_QUEUE_SEND, rtm->dest);
	  }
	  else
	  {
	       kmcr_table_add(rtm->dest, rtm->nhop, rtm->route_expire, ROUTE_ACTIVE);
	  }
//	    kaodv_queue_set_verdict(KAODV_QUEUE_SEND, rtm->dest);
	  break;

     case KMCR_DISCFAILED:
	  kmcr_table_del(rtm->dest);
	  kmcr_queue_set_verdict(KMCR_QUEUE_DROP, rtm->dest);
	  break;

     case KMCR_DELROUTE:
	  if(rtm->dest == 0) // ignore default routes
	       break;

	  kmcr_table_del(rtm->dest);

	  // Just to be safe
	  kmcr_queue_set_verdict(KMCR_QUEUE_DROP, rtm->dest);
	  break;

     case KMCR_ISROUTEVALID:
	  timeLeft = kmcr_table_time_remaining(rtm->dest);
	  if(timeLeft > TICKS_QUERY_KERNEL)
	       kmcr_netlink_send_rt_msg(KMCR_ROUTEVALID,
					0, rtm->dest, timeLeft);
	  break;

     case KMCR_GATEWAY:
	  gatewayAddr = rtm->dest;
	  break;
	    
     default:
	  status = -EINVAL;
     }
     return status;
}



#define RCV_SKB_FAIL(err) do { netlink_ack(skb, nlh, (err)); return; } while (0)

static inline void kmcr_netlink_rcv_skb(struct sk_buff *skb)
{
     int status, type, pid, flags, nlmsglen, skblen;
     struct nlmsghdr *nlh;

     skblen = skb->len;
     if (skblen < sizeof(*nlh)) {
	  printk("skblen too small\n");
	  return;
     }
     nlh = (struct nlmsghdr *)skb->data;
     nlmsglen = nlh->nlmsg_len;
     if (nlmsglen < sizeof(*nlh) || skblen < nlmsglen) {
	  printk("nlsmsg=%d skblen=%d to small\n", nlmsglen, skblen);
	  return;
     }
     pid = nlh->nlmsg_pid;
     flags = nlh->nlmsg_flags;

     if (pid <= 0 || !(flags & NLM_F_REQUEST) || flags & NLM_F_MULTI)
     {
	  netlink_ack(skb, nlh, -EINVAL);
	  return;
     }

     if (flags & MSG_TRUNC)
     {
	  netlink_ack(skb, nlh, -ECOMM);
	  return;
     }

     type = nlh->nlmsg_type;


     if (peer_pid) {
	  if (peer_pid != pid) {
	       netlink_ack(skb, nlh, -EBUSY);
	       return;
	  }
     } else
	  peer_pid = pid;


     status = kmcr_netlink_receive_peer(type, NLMSG_DATA(nlh),
					skblen - NLMSG_LENGTH(0));
     if (status < 0) {
	  netlink_ack(skb, nlh, status);
	  return;
     }

     if (flags & NLM_F_ACK)
	  netlink_ack(skb, nlh, 0);
     return;
}


static void kmcr_netlink_rcv_sk(struct sock *sk, int len)
{
     do {
	  struct sk_buff *skb;

	  if (down_trylock(&kmcrnl_sem))
	       return;

	  while ((skb = skb_dequeue(&sk->receive_queue)) != NULL) {
	       kmcr_netlink_rcv_skb(skb);
	       kfree_skb(skb);
	  }

	  up(&kmcrnl_sem);

     } while (kmcrnl && kmcrnl->receive_queue.qlen);

     return;
}

// Structure used for registering notifications when userspace code exits
static struct notifier_block kmcr_nl_notifier = {
     .notifier_call = kmcr_netlink_rcv_nl_event,
     NULL,
     0
};


/**************************** Initialization and cleanup ****************************/

int kmcr_netlink_init(void)
{
     kmcrnl = netlink_kernel_create(NETLINK_KMCR, kmcr_netlink_rcv_sk);
     if (kmcrnl == NULL) {
	  printk(KERN_ERR "KMCR: Failed to create netlink socket\n");
	  return -1;
     }
     if(kmcrnl->socket == NULL)
	  printk("KMCR: Bug found: Netlink initialization not correct\n");
     else if(kmcrnl->socket->sk != kmcrnl)
	  printk("KMCR: Bug found socket-->sock link not correct\n");
     netlink_register_notifier(&kmcr_nl_notifier);
     return 1;
}

void kmcr_netlink_fini(void)
{
     netlink_unregister_notifier(&kmcr_nl_notifier);
     if(kmcrnl->socket == NULL)
	  printk("Cannot release kcmrnl->socket - it is NULL\n");
     else
	  sock_release(kmcrnl->socket);
     down(&kmcrnl_sem);
     up(&kmcrnl_sem);
    
     printk("KMCR: Netlink finished\n");
}
 
