/*****************************************************************************
 * kmcr_queue.c: Support for packet buffering
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 *
 *****************************************************************************/


/*************************** NOTES *******************************************

This file provides support for buffering packets while route discovery 
is in progress. The buffered packets can be later forwarded or dropped

*********************************************************************************/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/netdevice.h>
#include <linux/netfilter_ipv4.h>
#include <linux/spinlock.h>
#include <net/icmp.h>

#include "kmcr_table.h"
#include "kmcr_queue.h"

#define QUEUE_MAX_SIZE 1024


struct queue_entry {
	struct list_head list;
	struct sk_buff *skb;
	int (*okfn) (struct sk_buff *);
	__u32 daddr;
};

typedef int (*kmcr_queue_cmpfn) (unsigned long, unsigned long);

static rwlock_t queue_lock = RW_LOCK_UNLOCKED;
static unsigned int queue_size;
static LIST_HEAD(queue_list);



/******************  Queuing functions that are local to file  *************************/

static inline struct queue_entry
*__kmcr_find_entry(kmcr_queue_cmpfn cmpfn, unsigned long data)
{
	struct list_head *p;

	list_for_each_prev(p, &queue_list) {
		struct queue_entry *entry = (struct queue_entry *)p;

		if (!cmpfn || cmpfn(entry->daddr, data))
			return entry;
	}
	return NULL;
}


static inline struct queue_entry
*__kmcr_dequeue_entry(kmcr_queue_cmpfn cmpfn, unsigned long data)
{
	struct queue_entry *entry;

	entry = __kmcr_find_entry(cmpfn, data);
	if (entry == NULL)
		return NULL;

	list_del(&entry->list);
	queue_size--;

	return entry;
}

static inline void __kmcr_queue_flush(void)
{
	struct queue_entry *entry;

	while ((entry = __kmcr_dequeue_entry(NULL, 0))) {
		kfree_skb(entry->skb);
		kfree(entry);
	}
}

static inline int __kmcr_enqueue_entry(struct queue_entry *entry)
{
	if (queue_size >= QUEUE_MAX_SIZE) {
	    printk(KERN_WARNING "KMCR: Queue full at %d entries, "
			       "dropping packet(s).\n", queue_size);
		return -ENOSPC;
	}
	list_add(&entry->list, &queue_list);
	queue_size++;
	return 0;
}

static inline int dest_cmp(unsigned long addr1, unsigned long addr2)
{
	return (addr1 == addr2);
}

static struct queue_entry
*kmcr_dequeue_entry(kmcr_queue_cmpfn cmpfn, unsigned long data)
{
    struct queue_entry *entry;
    
    write_lock_bh(&queue_lock);
    entry = __kmcr_dequeue_entry(cmpfn, data);
    write_unlock_bh(&queue_lock);
    return entry;
}


/********************* Queuing functions exposed to other files ****************************/

void kmcr_queue_flush(void)
{
	write_lock_bh(&queue_lock);
	__kmcr_queue_flush();
	write_unlock_bh(&queue_lock);
}



int kmcr_enqueue_packet(unsigned long daddr, struct sk_buff *skb, int (*okfn) (struct sk_buff *))
{
	int status = -EINVAL;
	struct queue_entry *entry;

	entry = kmalloc(sizeof(*entry), GFP_ATOMIC);

	if (entry == NULL) {
		printk(KERN_ERR
		       "KMCR: OOM in kmcr_enqueue_packet()\n");
		kfree_skb(skb);  // Free skb as we are returning NF_STOLEN anyway
		return -ENOMEM;
	}

	entry->okfn = okfn;
	entry->skb = skb;
	entry->daddr = daddr;

	write_lock_bh(&queue_lock);

	status = __kmcr_enqueue_entry(entry);

	if (status < 0)
		goto err_out_unlock;

	write_unlock_bh(&queue_lock);
	return status;

      err_out_unlock:
	write_unlock_bh(&queue_lock);
	kfree_skb(entry->skb);
	kfree(entry);

	return status;
}


int kmcr_queue_set_verdict(int verdict, __u32 daddr)
{
	struct queue_entry *entry;
	int pkts = 0;
	if (verdict == KMCR_QUEUE_DROP) 
	{
	    
	    while (1) {
		entry = kmcr_dequeue_entry(dest_cmp, daddr);
		
		if (entry == NULL)
		    return pkts;
		
		/* Send an ICMP message informing the application that the
		 * destination was unreachable. */
		if (pkts == 0)
		    icmp_send(entry->skb, ICMP_DEST_UNREACH,
			      ICMP_HOST_UNREACH, 0);
		
		kfree_skb(entry->skb);
		kfree(entry);
		pkts++;
	    }
	} else if (verdict == KMCR_QUEUE_SEND) 
	{
		while (1) {
		    entry = kmcr_dequeue_entry(dest_cmp, daddr);
		    
		    if (entry == NULL)
		    {
			return pkts;
		    }
		    
		    ip_route_me_harder(&entry->skb);
		    pkts++;
		    /* Inject packet */
		    entry->okfn(entry->skb);
		    kfree(entry);
		}
	}
	return 0;
}

/**************************** Initialization and cleanup ****************************/

void kmcr_queue_init(void)
{
    // Nothing to do for now
    return;
}

void kmcr_queue_fini(void)
{
    kmcr_queue_flush();

}
