/*****************************************************************************
 * kmcr_table.c: Routing table support
 *
 * Copyright (C) 2006 University of Illinois
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This code was written in the wireless networking group 
 * (http://www.crhc.uiuc.edu/wireless) at University of Illinois at Urbana-Champaign
 * Author: Pradeep Kyasanur, <kyasanur at crhc dot uiuc dot edu>
 * Modifications by: Rishi Bhardwaj, <rbhardw2 at uiuc dot edu >
 *
 *****************************************************************************/


/*************************** NOTES *******************************************

This file maintains a routing data structure that mirrors the userspace
routing table. The data structure is used to detect when packets have 
to be buffered and when on-demand route discovery has to be initiated .

*********************************************************************************/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/spinlock.h>
#include <linux/proc_fs.h>

#include "kmcr_table.h"

#define TABLE_MAX_LEN 100

static unsigned int table_len;

static rwlock_t table_lock = RW_LOCK_UNLOCKED;
static LIST_HEAD(table_head);



/******************** Table management functions local to file  ***********************/

static inline void __kmcr_table_flush(void)
{
     struct list_head *pos, *tmp;

     list_for_each_safe(pos, tmp, &table_head) {
	  struct route_entry *e = (struct route_entry *)pos;
	  list_del(&e->ptr);
	  table_len--;
	  kfree(e);
     }
}

static inline int __kmcr_table_add(struct route_entry *e)
{

     if (table_len >= TABLE_MAX_LEN) {
	  printk(KERN_WARNING "KMCR: kmcr_table_add - Max list len reached\n");
	  return -ENOSPC;
     }
     list_add(&e->ptr, &table_head);
     return 1;
}

static inline struct route_entry *__kmcr_table_find(__u32 daddr)
{
     struct list_head *pos;

     list_for_each(pos, &table_head) {
	  struct route_entry *e = (struct route_entry *)pos;

	  if (e->daddr == daddr)
	       return e;
     }
     return NULL;
}

static inline int __kmcr_table_del(struct route_entry *e)
{
     // sanity check
     if (e == NULL) 
	  return 0;

     list_del(&e->ptr);
     return 1;
}


/******************** Table manipulation functions available to external files **************************/

int kmcr_table_get(__u32 daddr, struct route_entry *ret)
{
     struct route_entry *e;
     int res = 0;

     read_lock_bh(&table_lock);
     e = __kmcr_table_find(daddr);

     if (e) {
	  res = 1;
	  if (ret)
	       memcpy(ret, e, sizeof(struct route_entry));
     }

     read_unlock_bh(&table_lock);
     return res;
}

int kmcr_table_add(__u32 daddr, __u32 nhop, unsigned long route_life,
		   unsigned short flags)
{
     struct route_entry *e;
     int status = 0;

     if (kmcr_table_get(daddr, NULL))
	  return 0;

     e = kmalloc(sizeof(struct route_entry), GFP_ATOMIC);

     if (e == NULL) {
	  printk(KERN_ERR "KMCR: Route table OOM in route_add\n");
	  return -ENOMEM;
     }

     e->daddr = daddr;
     e->nhop = nhop;
     e->flags = flags;
     e->last_access = jiffies;
     e->route_life = route_life * HZ; // route_life is in seconds, convert to jiffies
     e->pktCount = 0;

     write_lock_bh(&table_lock);

     status = __kmcr_table_add(e);

    
     // Shouldn't this be if(status > 0)  --- Comment by Rishi 
     if (status > 0)
	  table_len++;

     write_unlock_bh(&table_lock);

     if (status < 0)
	  kfree(e);

     return status;
}

void kmcr_table_update(__u32 daddr, __u32 nhop, unsigned long route_life,
		       unsigned short flags)
{
     struct route_entry *e;

     write_lock_bh(&table_lock);

     e = __kmcr_table_find(daddr);

     if (e == NULL) {
	  printk("KMCR: table_update - No entry to update!\n"); 
	  goto unlock;
     }
     e->nhop = nhop;
     e->flags = flags;
     /* Update expire time */
     e->route_life = route_life * HZ;
     e->last_access = jiffies;

unlock:
     write_unlock_bh(&table_lock);
}


long kmcr_table_time_remaining(__u32 daddr)
{
     struct route_entry *e;
     long timeLeft;

     read_lock_bh(&table_lock);

     e = __kmcr_table_find(daddr);    
     if(e == NULL)
	  timeLeft = 0;
     else
     {
	  if(e->route_life + e->last_access < jiffies)
	  {	
		timeLeft = 0;
	  }
	  else
	  {
	  	timeLeft = (e->route_life + e->last_access - jiffies)/ HZ;
	  }
     }
	
     read_unlock_bh(&table_lock);
     return timeLeft;

}


// Updates the entry of daddr for last access
// Used by netfilter hook to update the daddr entry whenever a packet is
// forwarded to daddr.
// Returns 0 if an entry for daddr was found and (-1) otherwise 
int kmcr_table_access_update(__u32 daddr)
{
     struct route_entry *e;
     int toRet = 0;

     write_lock_bh(&table_lock);

     e = __kmcr_table_find(daddr);

     if (e == NULL)
     {
	  // Routing entry to daddr not found in the routing table.
	  // If we received a packet to forward to daddr, 
	  // a RERR may be sent to the source of the packet.
	  toRet = -1;
	  goto unlock;
     }

     e->last_access = jiffies;

unlock:
     write_unlock_bh(&table_lock);
     return toRet;
}

void kmcr_table_stats_update(__u32 daddr)
{
     struct route_entry *e;

     write_lock_bh(&table_lock);

     e = __kmcr_table_find(daddr);

     if (e == NULL) 
	  goto unlock;

     e->pktCount++;

unlock:
     write_unlock_bh(&table_lock);
}


int kmcr_table_del(__u32 daddr)
{
     int res;
     struct route_entry *e;

     write_lock_bh(&table_lock);

     e = __kmcr_table_find(daddr);

     if (e == NULL) {
	  res = 0;
	  goto unlock;
     }

     res = __kmcr_table_del(e);

     if (res) {
	  kfree(e);
	  table_len--;
     }
unlock:
     write_unlock_bh(&table_lock);

     return res;
}

void kmcr_table_flush(void)
{
     write_lock_bh(&table_lock);
    
     __kmcr_table_flush();
    
     write_unlock_bh(&table_lock);
}


/**************  Functions to write data into kernel /proc structure ************************/

static int kmcr_table_print(char *buf)
{
     struct list_head *pos;
     int len = 0;

     read_lock_bh(&table_lock);

     len += sprintf(buf, "# %-20s %-15s %-15s %-15s %-15s\n", "Addr", "Nhop",
		    "Last Access", "Time left", "Pkt count");

     list_for_each(pos, &table_head) {
	  char addr[16], nhop[16];

	  struct route_entry *e = (struct route_entry *)pos;

	  sprintf(addr, "%d.%d.%d.%d",
		  0x0ff & e->daddr,
		  0x0ff & (e->daddr >> 8),
		  0x0ff & (e->daddr >> 16), 0x0ff & (e->daddr >> 24));

	  sprintf(nhop, "%d.%d.%d.%d",
		  0x0ff & e->nhop,
		  0x0ff & (e->nhop >> 8),
		  0x0ff & (e->nhop >> 16), 0x0ff & (e->nhop >> 24));

	  len +=
	       sprintf(buf + len, "  %-15s %-15s %12lu %14lu %14lu\n", addr, nhop,
		       e->last_access, 
		       (e->route_life + e->last_access < jiffies)? 0: (e->route_life + e->last_access - jiffies), 
		       e->pktCount);
     }

     read_unlock_bh(&table_lock);
     return len;
}


static int
kmcr_table_proc_info(char *buffer, char **start, off_t offset, int length)
{
     int len;

     len = kmcr_table_print(buffer);

     *start = buffer + offset;
     len -= offset;
     if (len > length)
	  len = length;
     else if (len < 0)
	  len = 0;
     return len;
}


/**************************** Initialization and cleanup ****************************/

void kmcr_table_init(void)
{
     proc_net_create("kmcr_table", 0, kmcr_table_proc_info);
     // Right now nothing to do
}

void kmcr_table_fini(void)
{
     kmcr_table_flush();
     proc_net_remove("kmcr_table");
}
